#include <stdio.h>
#include <stdlib.h>
#include "bitset.h"
#include "tc_types.h"
#include "tc_utils.h"
#include "sample.h"

/* 
   There is a nice page about counting bits:
   http://www.caam.rice.edu/~dougm/twiddle/BitCount.html
   I use the one for sparse bits
*/

typedef guint32 bs_segment;	/* bitset segment (4 bytes, 32 bits) */

#define BS_SEGMENT_BITS (sizeof(bs_segment) * 8)	/* bits per segment */
#define BS_SEGMENT(x) (((int)x)/((int)BS_SEGMENT_BITS))	/* number of segment for index x */
#define BS_POSITION(x) (x % BS_SEGMENT_BITS)	/* position of index x in its associated segment */

/* -------------------------------------------------------------------------- */
/* Creates a new bitset which contains nbits */
BitSet
bitset_new (bs_element max_id)
{
  BitSet bs = g_new0(struct bitset_struct, 1);

  bs->array = (GArray *) g_array_new (FALSE, TRUE, sizeof (bs_segment));
  bs->array = g_array_set_size (bs->array, BS_SEGMENT (max_id) + 1);
  bs->max_id = max_id;
  bs->it = 0;
  bs->total = 0;

  return bs;
}
/* -------------------------------------------------------------------------- */
/* Fills the set with all possible elements (all bits to 1) */
void
bitset_fullfill (BitSet *bs) {
  guint i;
  bs_segment *seg;

  for (i=0; i<=BS_SEGMENT((*bs)->max_id); i++) {
    seg = &g_array_index((*bs)->array, bs_segment, i);
    *seg = 0;
    *seg = ~*seg;
  }
  (*bs)->total = (*bs)->max_id+1;
}
/* -------------------------------------------------------------------------- */
BitSet
bitset_new_intersection (BitSet bs_a, BitSet bs_b)
{
  BitSet bs;
  guint i;

  if (!(bs_a && bs_b))
    tc_error ("bitset_new_intersection: invalid sets\n");
  bs = bitset_new (bs_a->max_id <
		   bs_b->max_id ? bs_a->max_id : bs_b->max_id);
	
  for (i = 0; i <= BS_SEGMENT (bs->max_id); i++)
    {
      bs_segment seg = g_array_index (bs->array, bs_segment, i) =
	g_array_index (bs_a->array, bs_segment, i) 
	& 
	g_array_index (bs_b->array, bs_segment, i);
      while (seg)
	{
	  bs->total++;
	  seg &= seg - 1;
	}
    }
  return bs;
}

/* -------------------------------------------------------------------------- */
BitSet
bitset_new_union (BitSet bs_a, BitSet bs_b)
{
  BitSet bs;
  guint i;
  bs_segment seg_a, seg_b, seg;

  if (!(bs_a && bs_b))
    tc_error ("bitset_new_union: invalid sets\n");
  bs =
    bitset_new (bs_a->max_id >
		bs_b->max_id ? bs_a->max_id : bs_b->max_id);

  for (i = 0; i <= BS_SEGMENT (bs->max_id); i++)
    {
      seg_a = i >
	BS_SEGMENT (bs_a->max_id) ? 0 : g_array_index (bs_a->
						       array,
						       bs_segment,
						       i);
      seg_b = i >
	BS_SEGMENT (bs_b->max_id) ? 0 : g_array_index (bs_b->
						       array,
						       bs_segment,
						       i);
      seg = g_array_index (bs->array, bs_segment, i) = seg_a | seg_b;
      while (seg)
	{
	  bs->total++;
	  seg &= seg - 1;
	}
    }
  return bs;
}

/* -------------------------------------------------------------------------- */
/* Creates a new bitset as copy of another */
BitSet bitset_new_copy(BitSet bs_a) {
  BitSet bs;
  guint i;

  if (!bs_a)
    tc_error ("bitset_new_copy: invalid set\n");
  bs = bitset_new (bs_a->max_id);
	
  for (i = 0; i <= BS_SEGMENT (bs->max_id); i++)
    g_array_index (bs->array, bs_segment, i) = g_array_index (bs_a->array, bs_segment, i);
  bs->total = bs_a->total;

  return bs;
}
/* -------------------------------------------------------------------------- */
/* Creates a new bitset as the difference between two */
BitSet bitset_new_diff(BitSet bs_a, BitSet bs_b) {
  BitSet bs;
  guint i;
  bs_segment seg_a, seg_b, seg;

  if (!(bs_a && bs_b))
    tc_error ("bitset_new_diff: invalid sets\n");
  bs = bitset_new (bs_a->max_id);

  for (i = 0; i <= BS_SEGMENT (bs->max_id); i++)
    {
      seg_a = i >
	BS_SEGMENT (bs_a->max_id) ? 0 : g_array_index (bs_a->
						       array,
						       bs_segment,
						       i);
      seg_b = i >
	BS_SEGMENT (bs_b->max_id) ? 0 : g_array_index (bs_b->
						       array,
						       bs_segment,
						       i);
      seg = g_array_index (bs->array, bs_segment, i) = seg_a & ~seg_b;
      while (seg)
	{
	  bs->total++;
	  seg &= seg - 1;
	}
    }
  return bs;
}
/* -------------------------------------------------------------------------- */
void
bitset_add_set (BitSet * bs, BitSet bs_a)
{
  guint i;
  bs_segment seg_a, seg;
	
  if (bs_a->max_id > (*bs)->max_id)
    {
      (*bs)->array = g_array_set_size ((*bs)->array, BS_SEGMENT (bs_a->max_id) + 1);
      (*bs)->max_id = bs_a->max_id;
    }
  (*bs)->total = 0;
  for (i = 0; i <= BS_SEGMENT ((*bs)->max_id); i++)
    {
      seg_a = i >	BS_SEGMENT (bs_a->max_id) ? 
	0 : 
	g_array_index (bs_a->array, bs_segment, i);
		
      seg =  g_array_index ((*bs)->array, bs_segment, i) |= seg_a;

      while (seg)
	{
	  (*bs)->total++;
	  seg &= seg - 1;
	}		
    }
}

/* -------------------------------------------------------------------------- */
void bitset_free (BitSet * bs)
{
  if (*bs)
    {
      g_array_free ((*bs)->array, TRUE);
      g_free (*bs);
    }
  else
    tc_error("bitset_free: set not initialized");
}

/* -------------------------------------------------------------------------- */
gboolean bitset_in (BitSet bs, bs_element id)
{
  bs_segment seg;
  if (!bs)
    tc_error ("bitset_in: invalid bitset\n");
  if (id > bs->max_id)
    return FALSE;
  seg = g_array_index (bs->array, bs_segment, BS_SEGMENT (id));
  return (seg & (1 << BS_POSITION (id)));
}

/* -------------------------------------------------------------------------- */
void bitset_delete(BitSet *bs, bs_element id)
{
  bs_segment *seg, mask;

  if (!(*bs))
    tc_error ("bitset_delete: invalid bitset\n");
  if (id > (*bs)->max_id)
    {
      fprintf (stderr,
	       "bitset_insert: index (%d) out of range (%d)\n", id,
	       (*bs)->max_id);
      exit (EXIT_FAILURE);
    }
  seg =
    &g_array_index ((*bs)->array, bs_segment, BS_SEGMENT (id));
  mask = 1 << BS_POSITION (id);
  if ((*seg & mask) != 0)
    (*bs)->total--;
  *seg = *seg & ~mask;
}

/* -------------------------------------------------------------------------- */
void bitset_insert (BitSet * bs, bs_element id)
{
  bs_segment *seg, mask;

  if (!(*bs))
    tc_error ("bitset_insert: invalid bitset\n");
  if (id > (*bs)->max_id)
    {
      fprintf (stderr,
	       "bitset_insert: index (%d) out of range (%d)\n", id,
	       (*bs)->max_id);
      exit (EXIT_FAILURE);
    }
  seg =
    &g_array_index ((*bs)->array, bs_segment, BS_SEGMENT (id));
  mask = 1 << BS_POSITION (id);
  if ((*seg & mask) == 0)
    (*bs)->total++;
  *seg = *seg | mask;
}

/* -------------------------------------------------------------------------- */
void bitset_it_reset (BitSet * bs)
{
  if (!(*bs))
    tc_error ("bitset_it_reset: invalid bitset\n");
  (*bs)->it = 0;
  while (!bitset_in (*bs, (*bs)->it) && (*bs)->it <= (*bs)->max_id)
    (*bs)->it++;
  if ((*bs)->it > (*bs)->max_id)
    (*bs)->it = -1;
}

/* -------------------------------------------------------------------------- */
bs_element bitset_it_next (BitSet * bs)
{
  bs_element cur;

  if (!(*bs))
    tc_error ("bitset_it_next: invalid bitset\n");
  cur = (*bs)->it;
  while (!bitset_in (*bs, ++(*bs)->it) && (*bs)->it <= (*bs)->max_id);
  if ((*bs)->it > (*bs)->max_id)
    (*bs)->it = -1;
  return cur;
}

/* -------------------------------------------------------------------------- */
gboolean bitset_it_end (BitSet bs)
{
  if (!bs)
    tc_error ("bitset_it_end: invalid bitset\n");
  return (bs->it == -1);
}

/* -------------------------------------------------------------------------- */
/* Prints to standard output elements in the set */
void bitset_dump (BitSet bs)
{
  bs_element i;
  printf("[ ");
  for (i = 0; i <= bs->max_id; i++)
    if (bitset_in (bs, i))
      printf ("%d ", i);
  printf ("] (%d elements)\n", bs->total);
}

/* -------------------------------------------------------------------------- */
/* Stores the bitset in a file at the given position */
void bitset_serialize (BitSet bs, FILE * f)
{
  fwrite (&bs->array->len, sizeof (guint), 1, f);
  fwrite (&bs->max_id, sizeof (bs_element), 1, f);
  fwrite (&bs->total, sizeof (bs_element), 1, f);
  fwrite (bs->array->data, sizeof (bs_segment), bs->array->len, f);
}

/* -------------------------------------------------------------------------- */
/* Reads the bitset from a file at the given position */
BitSet bitset_new_deserialize (FILE * f, long offset)
{
  guint size;	
  BitSet bs;

  fseek (f, offset, 0);
  fread (&size, sizeof (guint), 1, f);
  bs = bitset_new (size * BS_SEGMENT_BITS - 1);
  fread (&bs->max_id, sizeof (bs_element), 1, f);
  fread (&bs->total, sizeof (bs_element), 1, f);
  fread (bs->array->data, sizeof (bs_segment), size, f);
  return bs;
}

/* FOR TESTING  

gboolean verbose = TRUE;

#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>

int main(void) {
int i, j, total=25;
time_t t1;
(void) time (&t1);
srand ((long) t1);
BitSet bs = bitset_new(total), bs2 = bitset_new(total), intersec, unionn;
for (i=0; i<=total; i++) {
j=1+(int) (((float)total)*rand()/(RAND_MAX+1.0));
bitset_insert(&bs, j);
}
for (i=0; i<=total; i++) {
j=1+(int) (((float)total)*rand()/(RAND_MAX+1.0));
bitset_insert(&bs2, j);
}
intersec = bitset_new_intersection(bs, bs2);
unionn = bitset_new_union(bs, bs2);
bitset_dump(bs);
bitset_dump(bs2);
bitset_dump(intersec);
bitset_dump(unionn);
bitset_add_set(&bs, bs2);
bitset_dump(bs);

bitset_free(&bs);
bitset_free(&bs2);
bitset_free(&intersec);
bitset_free(&unionn);
return 0;
}

*/
