/* ---------------------------------------------------------------------------
 classifiers.h

 To add a new classifier:

 1. Add a new identifier
 2. Add a new string identifier
 3. Add the rest of functions as defined below
 4. Include the header here
 5. Implement them
 6. Add help options at main.c

 Arturo.Montejo.Raez@cern.ch - 2004

--------------------------------------------------------------------------- */
#ifndef __CLASSIFIERS_H_
#define __CLASSIFIERS_H_

#include <config.h>
#include <glib.h>
#include "plaum.h"
#include "rocchio.h"
#include "widrow_hoff.h"
#include "eg.h"
#include "lvq.h"
#include "intstr_table.h"

#ifdef HAVE_LIBGA
#include "genetic.h"
#endif

/* 
   Classifiers identifiers (algorithms) 
*/
enum {  
  CLASSIFIER_COMMAND,
  CLASSIFIER_PLAUM,
  CLASSIFIER_ROCCHIO,
  CLASSIFIER_NAIVE_BAYES,
  CLASSIFIER_WIDROW_HOFF,
  CLASSIFIER_KIVINEN_WARMUTH,
#ifdef HAVE_LIBGA
  CLASSIFIER_GENETIC,
#endif
  CLASSIFIER_LVQ,
  CLASSIFIER_NULL
};

/* 
   Classifiers string names 
*/
static char * classifier_str_ids[] = {  
  [CLASSIFIER_PLAUM] = "plaum",
  [CLASSIFIER_ROCCHIO] = "rocchio",
  [CLASSIFIER_NAIVE_BAYES] = "nb",
  [CLASSIFIER_WIDROW_HOFF] = "wh",
  [CLASSIFIER_KIVINEN_WARMUTH] = "eg",
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = "genetic",
#endif
  [CLASSIFIER_LVQ] = "lvq"
};


/* 
   Classifiers NEW functions
*/
static void* (*classifier_new[])(gchar *) = {  
  [CLASSIFIER_PLAUM] = (*plaum_new),
  [CLASSIFIER_ROCCHIO] = (*rocchio_new),
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = (*widrow_hoff_new),
  [CLASSIFIER_KIVINEN_WARMUTH] = (*eg_new),
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = (*genetic_new),
#endif
  [CLASSIFIER_LVQ] = (*lvq_new)
};

/* 
   Classifiers FREE functions
*/
static void (*classifier_free[])(void**) = {  
  [CLASSIFIER_PLAUM] = (*plaum_free),
  [CLASSIFIER_ROCCHIO] = (*rocchio_free),
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = (*widrow_hoff_free),
  [CLASSIFIER_KIVINEN_WARMUTH] = (*eg_free),
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = (*genetic_free),
#endif
  [CLASSIFIER_LVQ] = (*lvq_free)  
};

/* 
   Classifiers SAVE functions
*/
static void (*classifier_save[])(void*, gchar*) = {  
  [CLASSIFIER_PLAUM] = (*plaum_save),
  [CLASSIFIER_ROCCHIO] = (*rocchio_save),
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = (*widrow_hoff_save),
  [CLASSIFIER_KIVINEN_WARMUTH] = (*eg_save),
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = (*genetic_save),
#endif
  [CLASSIFIER_LVQ] = (*lvq_save)  
};

/* 
   Classifiers APPEND SAVE functions
*/
static void (*classifier_append_save[])(void*, gchar*) = {  
  [CLASSIFIER_PLAUM] = (*plaum_append_save),
  [CLASSIFIER_ROCCHIO] = NULL,
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = NULL,
  [CLASSIFIER_KIVINEN_WARMUTH] = NULL,
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = NULL,
#endif
  [CLASSIFIER_LVQ] = NULL  
};


/* 
   Classifiers SAVE TO ASCII functions (to export models)
*/
static void (*classifier_save_ascii[])(void*, gchar*, IntStrTable item_str, GArray *idid) = {
  [CLASSIFIER_PLAUM] = (*plaum_save_ascii),
  [CLASSIFIER_ROCCHIO] = NULL,
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = NULL,
  [CLASSIFIER_KIVINEN_WARMUTH] = NULL,
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = NULL,
#endif
  [CLASSIFIER_LVQ] = NULL
};

/* 
   Classifiers READ functions
*/
static void* (*classifier_new_read[])(gchar*) = {  
  [CLASSIFIER_PLAUM] = (*plaum_new_read),
  [CLASSIFIER_ROCCHIO] = (*rocchio_new_read),
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = (*widrow_hoff_new_read),
  [CLASSIFIER_KIVINEN_WARMUTH] = (*eg_new_read),
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = (*genetic_new_read),
#endif
  [CLASSIFIER_LVQ] = (*lvq_new_read)
};

/* 
   Classifiers READ AT functions
*/
static void* (*classifier_new_read_at[])(FILE *f) = {  
  [CLASSIFIER_PLAUM] = (*plaum_new_read_at),
  [CLASSIFIER_ROCCHIO] = NULL,
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = NULL,
  [CLASSIFIER_KIVINEN_WARMUTH] = NULL,
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = NULL,
#endif
  [CLASSIFIER_LVQ] = NULL
};


/* 
   Classifiers LEARN functions
*/
static void (*classifier_learn[])(void**, Samples) = {  
  [CLASSIFIER_PLAUM] = (*plaum_learn),
  [CLASSIFIER_ROCCHIO] = (*rocchio_learn),
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = (*widrow_hoff_learn),
  [CLASSIFIER_KIVINEN_WARMUTH] = (*eg_learn),
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = (*genetic_learn),
#endif
  [CLASSIFIER_LVQ] = (*lvq_learn)
};

/* 
   Classifiers CLASSIFY functions
*/
static double (*classifier_classify[])(void*, Sample) = {  
  [CLASSIFIER_PLAUM] = (*plaum_classify),
  [CLASSIFIER_ROCCHIO] = (*rocchio_classify),
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = (*widrow_hoff_classify),
  [CLASSIFIER_KIVINEN_WARMUTH] = (*eg_classify),
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = (*genetic_classify),
#endif
  [CLASSIFIER_LVQ] = (*lvq_classify)
};

/* 
   Classifiers DUMP functions
*/
static void (*classifier_dump[])(void*) = {  
  [CLASSIFIER_PLAUM] = (*plaum_dump),
  [CLASSIFIER_ROCCHIO] = (*rocchio_dump),
  [CLASSIFIER_NAIVE_BAYES] = NULL,
  [CLASSIFIER_WIDROW_HOFF] = (*widrow_hoff_dump),
  [CLASSIFIER_KIVINEN_WARMUTH] = (*eg_dump),
#ifdef HAVE_LIBGA
  [CLASSIFIER_GENETIC] = (*genetic_dump),
#endif
  [CLASSIFIER_LVQ] = (*lvq_dump)
};

#endif /*__CLASSIFIERS_H_*/
