/* --------------------------------------------------------------------------
   classrank.c 
   Vector of ranked classes
----------------------------------------------------------------------------- */

#include "classrank.h"

struct weight_struct {
  gdouble w;
  gint count;
  guint id;
};

/* -------------------------------------------------------------------------- */
ClassRank classrank_new(guint size) {
  guint i;
  struct weight_struct *ws;

  ClassRank cr = (ClassRank) g_array_new(FALSE, TRUE, sizeof(struct weight_struct));
  cr = g_array_set_size(cr, size);
  for (i=0; i<cr->len; i++) {
    ws = &g_array_index(cr, struct weight_struct, i);
    ws->id = i;
    ws->count = -1; /* this will allow us to know if the class was initialized */
    ws->w = 0.0;
  }
  return cr;
}
/* -------------------------------------------------------------------------- */
void classrank_free(ClassRank *cr) {
  if (*cr)
      g_array_free(*cr, TRUE);
}
/* -------------------------------------------------------------------------- */
void classrank_add(ClassRank *cr, guint id, gdouble weight) {
  struct weight_struct *ws;

  if (id >= (*cr)->len)
    tc_error("classrank_add: index out of range");
  ws = &g_array_index(*cr, struct weight_struct, id);
  ws->w += weight;
  ws->count = ws->count == -1 ? 1 : ws->count+1;
}
/* -------------------------------------------------------------------------- */
void classrank_avg(ClassRank *cr) {
  struct weight_struct *ws;
  guint i;
  for (i=0; i< (*cr)->len; i++) {
    ws = &g_array_index(*cr, struct weight_struct, i);
    if (ws->count > 0)
      ws->w /= (gdouble)ws->count;
  }
}
 /* -------------------------------------------------------------------------- */
 /* Return the 'ntop' positives classes. If ntop is zero all positives are returned.
 NOTE: the bitset returned must be dealloacted by the caller. */

gint __gcomp_weights(gconstpointer a, gconstpointer b) {
  gdouble diff =((struct weight_struct *)b)->w -
		((struct weight_struct *)a)->w; 
  if (diff == 0.0)
    return 0;
  if (diff > 0.0)
    return 1;
  return -1;
}

Sample classrank_get_top(ClassRank cr, guint ntop) {
  guint i, count;
  struct weight_struct *ws;
  Sample s;

  s = sample_new();
  if (cr->len == 0)
	return s;

  /* we sort the array of weights */
  g_array_sort(cr, __gcomp_weights);
  
  /* we take ntop */
  ntop = ntop >= cr->len || ntop == 0 ? cr->len : ntop;
  for (i=count=0; count<ntop && i<cr->len; i++) {
    ws = &g_array_index(cr, struct weight_struct, i);
	//printf("%lf %d\n", ws->w, ws->id);
	if (ws->count != -1) {
	  sample_add_feature(&s, feature_new(ws->id, ws->w));
	  count++;
	}
  }
  return s;
}

/* -------------------------------------------------------------------------- */
gint __gcomp_inv_weights(gconstpointer a, gconstpointer b) {
  gdouble diff =((struct weight_struct *)b)->w -
		((struct weight_struct *)a)->w; 
  if (diff == 0.0)
    return 0;
  if (diff > 0.0)
    return -1;
  return 1;
}

Sample classrank_get_bottom(ClassRank cr, guint nbottom) {
  guint i, count;
  struct weight_struct *ws;
  Sample s;

  s = sample_new();
  if (cr->len == 0)
	return s;

  /* we sort the array of weights */
  g_array_sort(cr, __gcomp_inv_weights);
  
  /* we take ntop */
  nbottom = nbottom >= cr->len || nbottom == 0 ? cr->len : nbottom;
  for (i=count=0; i<cr->len && count < nbottom; i++) {
    ws = &g_array_index(cr, struct weight_struct, i);
	//printf("%lf %d\n", ws->w, ws->id);
	if (ws->count != -1) {
	  sample_add_feature(&s, feature_new(ws->id, ws->w));
	  count++;
	}
  }
  return s;
}
/* -------------------------------------------------------------------------- */
void classrank_dump(ClassRank cr) {
  guint i, count=0;
  struct weight_struct *ws;

  printf("[");
  for (i=0; i<cr->len; i++) {
    ws = &g_array_index(cr, struct weight_struct, i);
    if (ws->w > 0.0) {
      printf("(%d, %.4lf, %d)", ws->id, ws->w, ws->count);
      count++;
    }
  }
  printf("] (%d positives)\n", count);
}
