#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <errno.h>
#include <error.h>
#include <math.h>

#include "eg.h"
#include "feature.h"
#include "sample.h"
#include "samples.h"
#include "weights.h"

extern gboolean verbose;

/* -------------------------------------------------------------------- */
void *eg_new(gchar *args)
{
  EG c = (EG) g_new0(struct eg_struct, 1);
  gint i = 1, len, pos=0;
  gchar **arg_val = NULL, **argv;

  if (!c)
    tc_error("eg_new: No memory for eg classifier\n");

  /* default values */
  c->eta = -1.0; /* will be translated when learning to default value */
  c->t = 0.0;

  /* no arguments defined, we return it with default values */
  if (!args)
    return ((void*)c);
  
  len = strlen(args);
  argv = g_strsplit(args, ":", 0);
  pos = strlen(argv[0]);
  if (len == pos)       /* no arguments */
    return ((void*)c);

  /* user defined values */  
  while (pos < len) {
    arg_val = g_strsplit(argv[i], "=", 2);
    if (!strncmp(arg_val[0], "eta", 3)) {
      c->eta = atof(arg_val[1]);
      if (c->eta <= 0.0)
	tc_error("Invalid value for eta in EG algorithm: must be positive");
    }
    else
      tc_error("Invalid arguments for EG");
    pos += strlen(argv[i++])+1; /* to control not to exceed the number of arguments */
    g_strfreev(arg_val);
  }
  g_strfreev(argv);
  return ((void*)c);
}

/* -------------------------------------------------------------------- */
void eg_free(void **cp) {
  EG c = (EG) *cp;
  if (c) {
    weights_free(&c->w);
    g_free(c);
  }
}
/* -------------------------------------------------------------------- */
void eg_save(void* cp, gchar *file) {
  EG c = (EG) cp;
  FILE *f;
  if (!c)
    tc_error("eg_save: EG structure not initialized");  
  if (!(f = fopen(file, "w")))
    tc_error_str("eg_save: unable to create file '%s'", file);
  fprintf(f, "%lf\n", c->t);
  weights_save(c->w, f);
  fclose(f);
}
/* -------------------------------------------------------------------- */
void *eg_new_read(gchar *file) {
  EG c;
  FILE *f = fopen(file, "r");
  if (!f)
    tc_error_str("eg_new_read: unable to open file '%s'", file);
  c = (EG) eg_new(NULL);
  fscanf (f, "%lf", &c->t);
  c->w = weights_new_read(f);
  fclose(f);
  return c;
}
/* -------------------------------------------------------------------- */
double eg_classify(void *cp, Sample s) {
  EG c = (EG) cp;
  double value = 0.0;
  Feature t;

  return weights_product_by_sample(c->w, s);
}
/* -------------------------------------------------------------------- */
void eg_dump (void *cp)
{
  EG c = (EG) cp;
  printf ("eta: %f\n", c->eta);
}
/* EG learning ----------------------------------------------------- */

/* -------------------------------------------------------------------- */
void eg_learn(void **cp, Samples ss)
{
  EG c = (EG) *cp;
  Sample s;
  Feature t;
  Weights prev, next, tmp;
  double dot, R, neww, yi, accumw;

  //eg_dump(c);

  if (c->eta == -1.0) {
    R = ss->max_w-ss->min_w;
    c->eta = 2.0/(3.0*R*R); /* default */
  }

  /* Initial weights: (1/d ... 1/d) */
  prev = weights_new (ss->max_feature_id);
  weights_set_all(&prev, 1.0/(double)(ss->max_feature_id+1.0));
  next = weights_new (ss->max_feature_id);

  samples_it_reset (&ss);  
  while (!samples_it_end (ss)) {
    s = samples_it_next (&ss);
    dot = weights_product_by_sample(prev, s);
	accumw = 0.0;
	yi = s->y < 0.0 ? 0.0 : s->y; /* because this algorithm considers lables {1,0} */
    sample_it_reset(&s);
    while (!sample_it_end(s)) {
      t = sample_it_next(&s);
      neww = weights_get(prev, t->id) * exp(-2.0 * c->eta * (dot - yi) * t->w);
	  weights_set(&next, t->id, neww);
	  accumw += neww;
    }
	/* normalization */
    sample_it_reset(&s);
    while (!sample_it_end(s) && accumw != 0.0) {
      t = sample_it_next(&s);
	  weights_set(&next, t->id, weights_get(next, t->id)/accumw);
    }

    tmp = next;
    next = prev;
    prev = tmp;
  }

  c->w = prev;    
  weights_free(&next);

  /* threshold */
  c->t = 0.0; //weights_threshold(c->w, ss);
}
