#include <config.h>

#ifdef HAVE_LIBGA

#include <stdio.h>
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <errno.h>
#include <error.h>
#include <math.h>

#include <ga/GASimpleGA.h>	// we're going to use the simple GA
#include <ga/GA1DBinStrGenome.h> // and the 1D binary string genome

#include "genetic.h"

#ifdef __cplusplus
extern "C" {
#endif

#include "feature.h"
#include "sample.h"
#include "samples.h"
#include "weights.h"
#include "tc_utils.h"

#ifdef __cplusplus
}
#endif

extern gboolean verbose;

/* -------------------------------------------------------------------- */
extern "C" void *genetic_new(gchar *args)
{
  Genetic c = (Genetic) g_new0(struct genetic_struct, 1);
  gint i = 1, len, pos=0;
  gchar **arg_val = NULL, **argv;

  if (!c)
    tc_error("genetic_new: No memory for eg classifier\n");

  /* default values */
  c->popsize = 50;    // size of population
  c->ngen = 100;      // number of generations (stop criteria)
  c->pmut = 0.01;      // probability of mutation
  c->pcross = 0.4;    // probability of crossover

  /* no arguments defined, we return it with default values */
  if (!args)
    return ((void*)c);

  len = strlen(args);
  argv = g_strsplit(args, ":", 0);
  pos = strlen(argv[0]);
  if (len == pos)       /* no arguments */
    return ((void*)c);

  /* user defined values */  
  while (pos < len) {
    arg_val = g_strsplit(argv[i], "=", 2);
    if (!strncmp(arg_val[0], "popsize", 7)) {
      c->popsize = atoi(arg_val[1]);
      if (c->popsize <= 0)
		tc_error("Invalid value for popsize in Genetic algorithm: must be integer positive");
    }
    else if (!strncmp(arg_val[0], "ngen", 4)) {
      c->ngen = atoi(arg_val[1]);
      if (c->ngen <= 0)
		tc_error("Invalid value for ngen in Genetic algorithm: must be integer positive");
    }
    else if (!strncmp(arg_val[0], "pmut", 4)) {
      c->pmut = atof(arg_val[1]);
      if (c->pmut < 0.0 || c->pmut > 1.0)
		tc_error("Invalid value for pmut in Genetic algorithm: must be within interval [0,1]");
    }
    else if (!strncmp(arg_val[0], "pcross", 6)) {
      c->pcross = atof(arg_val[1]);
      if (c->pcross < 0.0 || c->pcross > 1.0)
		tc_error("Invalid value for pcross in Genetic algorithm: must be within interval [0,1]");
    }
    else
      tc_error("Invalid arguments for Genetic");
    pos += strlen(argv[i++])+1; /* to control not to exceed the number of arguments */
    g_strfreev(arg_val);
  }
  g_strfreev(argv);
  return ((void*)c);
}

/* -------------------------------------------------------------------- */
extern "C" void genetic_free(void **cp) {
  Genetic c = (Genetic) *cp;
  if (c) {
	delete (GA1DBinaryStringGenome*)c->genome;
    g_free(c);
  }
}
/* -------------------------------------------------------------------- */
extern "C" void genetic_save(void* cp, gchar *file) {
  Genetic c = (Genetic) cp;
  GA1DBinaryStringGenome *genome = (GA1DBinaryStringGenome*)c->genome;
  FILE *f;
  guint i;
  if (!c)
    tc_error("genetic_save: Genetic structure not initialized");  
  if (!(f = fopen(file, "w")))
    tc_error_str("genetic_save: unable to create file '%s'", file);
  fprintf(f, "%d\n", genome->length());
  for (i=0; i<genome->length(); i++) 
	fprintf(f, " %d",genome->gene(i));
  fclose(f);
}
/* -------------------------------------------------------------------- */
extern "C" void *genetic_new_read(gchar *file) {
  Genetic c;
  FILE *f = fopen(file, "r");
  guint nbits, i=0;
  short bit;  

  if (!f)
    tc_error_str("genetic_new_read: unable to open file '%s'", file);
  c = (Genetic) genetic_new(NULL);
  fscanf (f, "%d\n", &nbits);
  c->genome = (void*) new GA1DBinaryStringGenome(nbits);
  ((GA1DBinaryStringGenome*)c->genome)->unset(0,nbits-1);

  while (!feof(f)) {
	fscanf(f, " %d", &bit);
	if (bit)
	  ((GA1DBinaryStringGenome*)c->genome)->gene(i,1);
	i++;
  }
  fclose(f);
  return c;
}
/* -------------------------------------------------------------------- */
double __genetic_distance(GA1DBinaryStringGenome *genome, Sample s) {
  double dist = 0.0;
  Feature f;
  guint i, nf=0, nz=0;

  sample_it_reset(&s);
  while (!sample_it_end(s)) {
	f = sample_it_next(&s);
	if (f->id < genome->length()){
	  dist += /*f->w**/genome->gene(f->id);
	}
	nf++;
  }
  //for (i=0; i<genome->length(); i++)
  //	nz += genome->gene(i);

  //printf("%lf\n", dist/((double)(nf+nz)));
  return dist;
  return dist/((double)(nf+nz));
}
/* -------------------------------------------------------------------- */
extern "C" double genetic_classify(void *cp, Sample s) {
  Genetic c = (Genetic) cp;
  guint nz = 0;
  GA1DBinaryStringGenome *genome = (GA1DBinaryStringGenome*)c->genome;
  guint i;

  //  for (i=0; i<genome->length(); i++)
  //	nz += genome->gene(i);
  return __genetic_distance(genome, s);
  /*
  if (__genetic_distance(genome, s) / (double)nz > 0.008)
	return 1;
  else
  return -1;*/
}
/* -------------------------------------------------------------------- */
extern "C" void genetic_dump (void *cp)
{
  Genetic c = (Genetic) cp;
  GA1DBinaryStringGenome *genome = (GA1DBinaryStringGenome*)c->genome;
  guint i;

  for (i=0; i<genome->length(); i++)
	printf("%d", genome->gene(i));
  printf("\npopulation size:    %d\n", c->popsize);
  printf("no. generations:    %d\n", c->ngen);
  printf("prob. of crossover: %f\n", c->pcross);
  printf("prob. of mutation:  %f\n", c->pmut);
  printf ("\n");
}
/* Genetic learning ----------------------------------------------------- */

float
__genetic_fitness(GAGenome& g) {
  
  GA1DBinaryStringGenome & genome = (GA1DBinaryStringGenome &)g;
  Genetic c = (Genetic) genome.userData();
  Sample s;
  guint npos=0, nneg=0;
  double tpos=0.0, tneg=0.0;

  float score=0.0;

  samples_it_reset(&c->ss);
  while (!samples_it_end(c->ss)) {
	s = samples_it_next(&c->ss);
	if (s->y > 0.0) {
	  // score += (guint)__genetic_distance(&genome, s);
	  tpos += __genetic_distance(&genome, s);
	  npos++;
	} else {
	  tneg += __genetic_distance(&genome, s);
	  nneg++;
	}
  } 
  //return score;
  printf("->%f %f %f\n", tpos, tneg, ((float)tpos)/((float)tneg));
  return tpos;
  return ((float)tpos)/((float)tneg);
  return ((float)(tpos*nneg))/((float)(tneg*npos + 1.0));
}

/* -------------------------------------------------------------------- */
extern "C" void genetic_learn(void **cp, Samples ss)
{
  Genetic c = (Genetic) *cp;

  c->ss = ss;
  c->genome = (void*) new GA1DBinaryStringGenome(ss->max_feature_id+1, 
						 __genetic_fitness, 
						 (void*)c);

  // evolution process
  GASimpleGA ga(*(GA1DBinaryStringGenome*)c->genome);
  ga.populationSize(c->popsize);
  ga.nGenerations(c->ngen);
  ga.pMutation(c->pmut);
  ga.pCrossover(c->pcross);
  ga.evolve();

  // get best individual
  delete (GA1DBinaryStringGenome*) c->genome;
  c->genome = (void*) new GA1DBinaryStringGenome((GA1DBinaryStringGenome&)ga.statistics().bestIndividual(0));
  cout << ga.statistics();
  genetic_dump(c);
}

#endif
