#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <errno.h>
#include <error.h>
#include <math.h>

#include "lvq.h"

extern gboolean verbose;

#define LVQ_INIT_MEANS 0
#define LVQ_INIT_KNN   1

#define _LVQ_ITER_MULTIPLICITY 30 // number of times the maximum number of cbv to set number of iterations

/* -------------------------------------------------------------------- */
void *lvq_new(gchar *args)
{
  LVQ c = (LVQ) g_new0(struct lvq_struct, 1);
  gint i = 1, len, pos=0;
  gchar **arg_val = NULL, **argv;

  if (!c)
    tc_error("lvq_new: No memory for lvq classifier\n");

  /* default values */  
  c->iterations = 0;            /* to be set to automatically */
  c->alpha_init = 0.05;
  c->alpha_threshold = 0.001;   /* no used so far */
  c->npos_cbv = c->nneg_cbv = 1;
  c->pos_cbv = c->neg_cbv = NULL;
  c->init = LVQ_INIT_MEANS;
  c->normcbv = 0;

  /* no arguments defined, we return it with default values */
  if (!args)
    return ((void*)c);
  
  len = strlen(args);
  argv = g_strsplit(args, ":", 0);
  pos = strlen(argv[0]);
  if (len == pos)       /* no arguments */
    return ((void*)c);

  /* user defined values */  
  while (pos < len) {
    arg_val = g_strsplit(argv[i], "=", 2);
    if (!strncmp(arg_val[0], "iter", 4))
      c->iterations = atoi(arg_val[1]);
    else if (!strncmp(arg_val[0], "means_init", 10))
      c->init = LVQ_INIT_MEANS;
    else if (!strncmp(arg_val[0], "knn_init", 8))
      c->init = LVQ_INIT_KNN;
    else if (!strncmp(arg_val[0], "alpha", 5))
      c->alpha_init = atof(arg_val[1]);
    else if (!strncmp(arg_val[0], "u", 1))
      c->alpha_threshold = atof(arg_val[1]);
    else if (!strncmp(arg_val[0], "ncbv", 4))
      c->npos_cbv = c->nneg_cbv = atoi(arg_val[1]);
    else if (!strncmp(arg_val[0], "ncbv_pos", 8))
      c->npos_cbv = atoi(arg_val[1]);
    else if (!strncmp(arg_val[0], "ncbv_neg", 8))
      c->nneg_cbv = atoi(arg_val[1]);
    else if (!strncmp(arg_val[0], "normcbv", 8))
      c->normcbv = atoi(arg_val[1]);
    else
      tc_error("Invalid arguments for LVQ");
    pos += strlen(argv[i++])+1; /* to control not to exceed the number of arguments */
    g_strfreev(arg_val);
  }

  g_strfreev(argv);
  return ((void*)c);
}

/* -------------------------------------------------------------------- */
void lvq_free(void **cp) {
  LVQ c = (LVQ) *cp;
  guint i;
  if (c) {
    for (i=0; i<c->npos_cbv && c->pos_cbv[i]; i++)
      weights_free(&c->pos_cbv[i]);
    g_free(c->pos_cbv);
    for (i=0; i<c->nneg_cbv && c->neg_cbv[i]; i++)
      weights_free(&c->neg_cbv[i]);
    g_free(c->neg_cbv);
    
    g_free(c);
  }
}
/* -------------------------------------------------------------------- */
void lvq_save(void* cp, gchar *file) {
  LVQ c = (LVQ) cp;
  FILE *f;
  guint i;
  if (!c)
    tc_error("lvq_save: LVQ structure not initialized");  
  if (!(f = fopen(file, "w")))
    tc_error_str("lvq_save: unable to create file '%s'", file);
  fprintf(f, "%d\n", c->npos_cbv);
  for (i=0; i<c->npos_cbv && c->pos_cbv[i]; i++)
    weights_save(c->pos_cbv[i], f);
  fprintf(f, "%d\n", c->nneg_cbv);
  for (i=0; i<c->nneg_cbv && c->neg_cbv[i]; i++)
    weights_save(c->neg_cbv[i], f);
  fclose(f);
}
/* -------------------------------------------------------------------- */
void *lvq_new_read(gchar *file) {
  LVQ c;
  FILE *f = fopen(file, "r");
  guint i;

  if (!f)
    tc_error_str("lvq_new_read: unable to open file '%s'", file);
  c = (LVQ) lvq_new(NULL);
  
  fscanf (f, "%d", &c->npos_cbv);
  c->pos_cbv = g_new0(Weights, c->npos_cbv);
  for (i=0; i<c->npos_cbv; i++)
    c->pos_cbv[i] = weights_new_read(f);
  fscanf (f, "%d", &c->nneg_cbv);
  c->neg_cbv = g_new0(Weights, c->nneg_cbv);
  for (i=0; i<c->nneg_cbv; i++)
    c->neg_cbv[i] = weights_new_read(f);
  
  fclose(f);
  
  return c;
}
/* -------------------------------------------------------------------- */
Weights __lvq_closest_cbv(LVQ c, Sample s, double *mindist, double *sign) {
  double dist;
  guint i;
  Weights cbv;
  
  *sign = 1.0;
  cbv = c->pos_cbv[0];
  *mindist = weights_distance_to_sample(cbv, s);
  for (i=1; i<c->npos_cbv && c->pos_cbv[i]; i++) {
    dist = weights_distance_to_sample(c->pos_cbv[i], s);
    if (dist < *mindist) {
      *mindist = dist;      
      cbv = c->pos_cbv[i];
    }
  }
  for (i=0; i<c->nneg_cbv && c->neg_cbv[i]; i++) {
    dist = weights_distance_to_sample(c->neg_cbv[i], s);
    if (dist < *mindist) {
      *mindist = dist;
      *sign = -1.0;
      cbv = c->neg_cbv[i];
    }
  }
  return cbv;  
}
/* -------------------------------------------------------------------- */
double lvq_classify(void *cp, Sample s) {
  LVQ c = (LVQ) cp;
  double dist, sign;
  Weights cbv = __lvq_closest_cbv(c, s, &dist, &sign);  
  return sign*dist;
}
/* -------------------------------------------------------------------- */
void lvq_dump (void *cp)
{
  LVQ c = (LVQ) cp;
  printf("iterations: %d\n", c->iterations);
  printf("alpha_init: %lf\n", c->alpha_init);
  printf("npos_cbv/nneg_cbv: %d/%d\n", c->npos_cbv, c->nneg_cbv);
}

/* Get npos and nneg ------------------------------------------------ */
void __lvq_get_posneg(Samples ss,  guint *npos, guint *nneg) {
  Sample s;
 
  *npos = *nneg = 0;
  samples_it_reset(&ss);
  while (!samples_it_end(ss)) {
    s = samples_it_next(&ss);
    if (s->y >= 0.0)
      (*npos)++;
    else
      (*nneg)++;
  }
}
/* LVQ learning ----------------------------------------------------- */
void __lvq_init_means(LVQ *cp, Samples ss) {
  LVQ c = (LVQ) *cp;
  Sample s;
  Weights pos_cbv, neg_cbv, cbv;
  guint pos=0, neg=0, i;
  Feature t;

  pos_cbv = weights_new(0);
  neg_cbv = weights_new(0);
  samples_it_reset(&ss);
  while (!samples_it_end(ss)) {
    s = samples_it_next(&ss);
    if (s->y > 0.0) {
      cbv = pos_cbv;
      pos++;
    }
    else {      
      cbv = neg_cbv;
      neg++;
    }

    sample_it_reset(&s);
    while (!sample_it_end(s)) {
      t = sample_it_next(&s);
      weights_add_weight(&cbv, t->id, t->w);
    }
  }
  weights_product_by_scalar(&pos_cbv, 1.0/(double)pos);
  weights_product_by_scalar(&neg_cbv, 1.0/(double)neg);

  c->pos_cbv = g_new0(Weights, c->npos_cbv);
  c->neg_cbv = g_new0(Weights, c->nneg_cbv);
  if (!c->pos_cbv || !c->neg_cbv) 
    tc_error("__lvq_init_means: no memory");

  for (i=0; i<c->npos_cbv; i++)
    c->pos_cbv[i] = weights_clone(pos_cbv);
  for (i=0; i<c->nneg_cbv; i++)
    c->neg_cbv[i] = weights_clone(neg_cbv);

  weights_free(&pos_cbv);
  weights_free(&neg_cbv);

  //c->iterations = ss->total/5;
}
/*--------------------------------------------------------------------*/
/* This function initializes codebook vectors using those closest to
   each other (KNN approach) */
void __lvq_init_knn(LVQ *cp, Samples ss) {
  LVQ c = (LVQ) *cp;
  Sample s, ssim;
  Weights *pos_cbv, *neg_cbv;
  Samples sss;
  guint posi=0, negi=0, i;
  gboolean posdone=FALSE, negdone=FALSE;

  pos_cbv = g_new0(Weights, c->npos_cbv);
  neg_cbv = g_new0(Weights, c->nneg_cbv);
  sss = samples_shallow_clone(ss);
  samples_it_reset(&ss);
  while (!samples_it_end(ss) && (!posdone || !negdone)) {
    s = samples_it_next(&ss);
    samples_sort_by_similarity(&sss, s);
    samples_it_reset(&sss);
    while (!samples_it_end(sss) && (!posdone || !negdone)) {
      ssim = samples_it_next(&sss);
      if (ssim->y == s->y) {
		if (s->y > 0.0 && !posdone) {
		  if (posi < c->npos_cbv)
			pos_cbv[posi++] = weights_new_from_sample(s);
		  else
			posdone = TRUE;	    
		} else if (!negdone) {
		  if (negi < c->nneg_cbv)
			neg_cbv[negi++] = weights_new_from_sample(s);
		  else
			negdone = TRUE;
		}
      }
    }
  }

  /* we resize vectors of docbook vectors and free some memory */
  samples_shallow_free(&sss);
  c->pos_cbv = g_new0(Weights, posi);
  c->neg_cbv = g_new0(Weights, negi);  
  for (i=0; i<posi; i++)
    c->pos_cbv[i] = pos_cbv[i];
  g_free(pos_cbv);
  for (i=0; i<negi; i++)
    c->neg_cbv[i] = neg_cbv[i];
  g_free(neg_cbv);
  c->npos_cbv = posi;
  c->nneg_cbv = negi;
  
}
/* -------------------------------------------------------------------- */
void lvq_learn(void **cp, Samples ss)
{
  LVQ c = (LVQ) *cp;
  Sample s;
  Weights cbv, diff;
  guint i, j, npos, nneg;
  double alpha, sign, dist, val, factor, ii;
  Feature t;

  /*
    First, we determine the number of codebook vectors if using balancing
  */
  if (c->normcbv) {
    __lvq_get_posneg(ss, &npos, &nneg);

    /* We adjust the number of codebook vectors to the balance of the
       class. The inner imbalance degree is:
       i_i = |1-2n_i/n|
    */
    // ii = (1.0 - 2.0*(double)npos/(double)(npos+nneg));
    c->npos_cbv = npos < nneg ? c->normcbv : (int) log((float)npos/nneg) / log(2) + c->normcbv;
    c->nneg_cbv = nneg < npos ? c->normcbv : (int) log((float)nneg/npos) / log(2) + c->normcbv;    
  }
  lvq_dump(c);
  printf("ndocs pos/neg: %d/%d\n", npos, nneg);

  /*
    Second, initialize codebook vectors
  */
  fprintf(stderr, "initializing codebook vectors\n");
  if (c->init == LVQ_INIT_KNN)
    __lvq_init_knn(&c, ss);
  else
    __lvq_init_means(&c, ss);

  /* 
     LVQ1 algorithm 
  */

  alpha = c->alpha_init;

  if (c->iterations == 0)
    c->iterations = (c->nneg_cbv + c->npos_cbv)*_LVQ_ITER_MULTIPLICITY;

  fprintf(stderr, "Running LVQ algorithm [%d iterations]", c->iterations);
  for (i=0; i<c->iterations && alpha > 0.0; i++) {
    fprintf(stderr, ".%d-%f", i, alpha);
    samples_it_reset(&ss);
    while (!samples_it_end(ss)) {
      s = samples_it_next(&ss);
      
      /* we look for the closest codebook vector */
      cbv = __lvq_closest_cbv(c, s, &dist, &sign);

      /* we update its weights */	
      factor = s->y*sign > 0.0 ? alpha : -alpha;
      weights_resize(&cbv, s->max_id);
      sample_it_reset(&s);
      j=0;
      while (!sample_it_end(s)) {
	t = sample_it_next(&s);
	for (;j<t->id;j++) {
	  val = weights_get(cbv, j);
	  weights_set(&cbv, j, val-factor*val);
	}
	val = weights_get(cbv,t->id);	
	weights_set(&cbv, t->id, val+factor*(t->w-val));
	j++;
      }
      for (;j<cbv->len;j++) {
	val = weights_get(cbv, j);
	weights_set(&cbv, j, val-factor*val);
      }
    }
    /* we decrease learning rate linearly */
    alpha -= alpha/((double)c->iterations);
  }
}
