/*---------------------------------------------------------------------
  ML-Tree

  A classification tree for multi-label data.

---------------------------------------------------------------------*/

#include "tm.h"
#include "mltree.h"
#include "fold.h"
#include "classrank.h"

struct class_score {
  tc_id id;
  gdouble score;
};

extern gboolean verbose;

/* -------------------------------------------------------------------------- */
gint __gcomp_scores(gconstpointer a, gconstpointer b) {
  /* the lowest the best! (low imbalance degree) */
  gdouble diff = ((struct class_score *)a)->score - ((struct class_score *)b)->score;
  if (diff == 0.0) 
    return 0;
  if (diff < 0.0)
    return -1;
  return 1;
}
/* -------------------------------------------------------------------------- */
gdouble __mltree_imbalance_degree(tc_id class_id, BitSet train_docs, BitSet *mclass_docs) {
  BitSet class_docs, postrain_docs, negtrain_docs;
  gdouble id;

  class_docs = mclass_docs[class_id];
  postrain_docs = bitset_new_intersection(train_docs, class_docs);
  negtrain_docs = bitset_new_diff(train_docs, class_docs);
  id = tm_imbalance_degree(postrain_docs->total, postrain_docs->total+negtrain_docs->total);
  bitset_free(&postrain_docs);
  bitset_free(&negtrain_docs);
  
  return id;
}
/* -------------------------------------------------------------------------- */
gint __comp_imbdegree(gconstpointer a, gconstpointer b) {
  gdouble diff = *(gdouble*)b-*(gdouble*)a;
  if (diff == 0.0)
	return 0;
  if (diff < 0.0)
	return -1;
  return 1;
}
/* -------------------------------------------------------------------------- */
gdouble __mltree_children_imbalance_degree(tc_id class_id, BitSet classes, BitSet train_docs, BitSet *mclass_docs, FoldTrainParams *params) {
  BitSet class_docs, postrain_docs, negtrain_docs;
  bs_element it;
  tc_id c;
  gdouble accum_id=0.0;
  GArray *balances;

  it = classes->it;
  bitset_it_reset(&classes);
  class_docs = mclass_docs[class_id];
  postrain_docs = bitset_new_intersection(train_docs, class_docs);
  negtrain_docs = bitset_new_diff(train_docs, class_docs);
  balances = g_array_new(FALSE, TRUE, sizeof(gdouble));
  balances = g_array_set_size(balances, classes->max_id+1);
  while (!bitset_it_end(classes)) {
    c = bitset_it_next(&classes);
    if (c != class_id) {
      g_array_index(balances, gdouble, c) =  __mltree_imbalance_degree(c, postrain_docs, mclass_docs) +
		__mltree_imbalance_degree(c, negtrain_docs, mclass_docs);
    }
  }
  g_array_sort(balances, __comp_imbdegree);
  for (c=0; c<params->tree_max_children && c<balances->len; c++)
	accum_id += g_array_index(balances, gdouble, c);
  g_array_free(balances, TRUE);
  bitset_free(&postrain_docs);
  bitset_free(&negtrain_docs);
  classes->it = it;
  return accum_id;
}
/* -------------------------------------------------------------------------- */
gdouble __mltree_balance(tc_id class_id, BitSet classes, BitSet train_docs, BitSet *mclass_docs, FoldTrainParams *params) {
  BitSet class_docs, postrain_docs, negtrain_docs;
  gdouble id, ig;
  guint dummy;

  class_docs = mclass_docs[class_id];
  switch (params->tree_balance) {
  case 0: 
	return __mltree_imbalance_degree(class_id, train_docs, mclass_docs);
	break;
  case 1:
	return 1.0/tm_infgain(mclass_docs, classes, class_docs, train_docs->total, &dummy); 
	break;
  case 2:
	return __mltree_imbalance_degree(class_id, train_docs, mclass_docs) /
	  tm_infgain(mclass_docs, classes, class_docs, train_docs->total, &dummy); 
	break;
  case 3: 
	return __mltree_children_imbalance_degree(class_id, classes, train_docs, mclass_docs, params);
	break;
  case 4:
	return __mltree_children_imbalance_degree(class_id, classes, train_docs, mclass_docs, params) /
	  tm_infgain(mclass_docs, classes, class_docs, train_docs->total, &dummy); 
	break;
  default:
	tc_error("__mltree_balance: invalid method");
  }
}
/* -------------------------------------------------------------------------- */
GArray *__get_balances(BitSet classes, BitSet *mclass_docs, BitSet train_docs, FoldTrainParams *params) {
  GArray *scores;
  tc_id count=0, c;
  struct class_score *score;
  bs_element it;

  scores = g_array_new (FALSE, TRUE, sizeof (struct class_score));
  scores = g_array_set_size (scores, classes->total);
  bitset_it_reset(&classes);
  while (!bitset_it_end(classes)) {    
    c = bitset_it_next(&classes);
    score = &g_array_index(scores, struct class_score, count);
    score->id = c;
    it = classes->it; /* to keep iterator */
    score->score = __mltree_balance(c, classes, train_docs, mclass_docs, params);
    classes->it = it;
    //printf("->%f\n", score->score);
    count++;
  }
  /* the lowest the best! (low imbalance degree) */
  g_array_sort(scores, __gcomp_scores);
  return scores;
}
/* -------------------------------------------------------------------------- */
gboolean __is_leaf(MLNode *node) {
  if (!node)
    return TRUE;
  return g_node_n_children((GNode *)node) == 0;
}
void __pos_neg_classes(tc_id class_id, BitSet classes, BitSet postrain_docs, BitSet *mclass_docs, BitSet *posclasses, BitSet *negclasses) {

  BitSet common;
  tc_id c;

  *posclasses = bitset_new(classes->max_id);
  *negclasses = bitset_new(classes->max_id);
  bitset_it_reset(&classes);
  while (!bitset_it_end(classes)) {    
	c = bitset_it_next(&classes);
	if (c!=class_id) {
	  common = bitset_new_intersection(postrain_docs, mclass_docs[c]);
	  if (common->total > 0)
		bitset_insert(posclasses, c);
	  else
		bitset_insert(negclasses, c);
	  bitset_free(&common);
	}
  }
}
/* -------------------------------------------------------------------------- */
MLNode *
mltree_new(Fold *f, BitSet classes, 
	   BitSet train_docs, BitSet eval_docs, 
	   Samples train_samples, Samples eval_samples,
	   BitSet *mclass_docs, 
	   FoldTrainParams *params, guint depth) {

  tc_id i, count=0;
  BitSet postrain_docs, poseval_docs, negeval_docs, negtrain_docs, posclasses, 
	negclasses, class_docs, trained, filtered_classes;
  MLNode *node, *child;
  GArray *scores;
  Method method=NULL;
  struct class_score *score;
  gboolean oldverbose;
  guint max_children;

  node = (MLNode*) g_node_new(NULL);
  node->data = NULL;

  /* no more classes to train */
  if (classes->total == 0 || (depth >= params->tree_max_depth && params->tree_max_depth > 0))
    return node;

  if (verbose) {
	printf("[%d/%d]", depth, classes->total);
	fflush(stdout);
  }

  /* 
     We compute the balance for each class and sort with the lowest
     one first
  */
  scores = __get_balances(classes, mclass_docs, train_docs, params);

  /* 
     We train for node class (if we can find a valid method)
  */

  max_children = params->tree_max_children > 0 && params->tree_max_children < scores->len ? 
    params->tree_max_children : scores->len;

  trained = bitset_new(classes->max_id);

  for (i=0; i<max_children && max_children < scores->len; i++) {
    score = &g_array_index(scores, struct class_score, i);
    class_docs = mclass_docs[score->id];
    postrain_docs = bitset_new_intersection(train_docs, class_docs);
    negtrain_docs = bitset_new_diff(train_docs, class_docs);
    poseval_docs =  bitset_new_intersection(eval_docs, class_docs);
    negeval_docs = bitset_new_diff(eval_docs, class_docs);

    /* at least, two documents per set */
    method = NULL;
    if (postrain_docs->total * negtrain_docs->total  *
    	poseval_docs->total * negeval_docs->total) {
      oldverbose = verbose;
      verbose = 0;
      method = fold_train_class(f, 
								score->id, 
								postrain_docs, negtrain_docs, 
								poseval_docs, negeval_docs, 
								params, 
								train_samples, eval_samples);
      verbose = oldverbose;
    }
	
    if (method != NULL) { /* valid method, we add a node */
      
	  tc_message_int("(%d)", score->id);

      /* Find positive and negative classes */	  
      child = (MLNode*) g_node_new(NULL);
      child->data = (gpointer) method_new_clone(method);
      g_node_append(node, (GNode *)child);
	
	  if (params->tree_duplicates) {
		__pos_neg_classes(score->id, classes, postrain_docs, mclass_docs, &posclasses, &negclasses);
		g_node_append(child, mltree_new(f, 
										posclasses, 
										postrain_docs, eval_docs, 
										train_samples, eval_samples,
										mclass_docs, 
										params, depth+1));
		
		g_node_append(child, mltree_new(f, 
										negclasses, 
										negtrain_docs, eval_docs, 
										train_samples, eval_samples,
										mclass_docs, 
										params, depth+1));
		bitset_free(&posclasses);
		bitset_free(&negclasses);
	  } else
		bitset_insert(&trained, score->id);
    } else 
	  max_children++;
	
    bitset_free(&postrain_docs);
    bitset_free(&negtrain_docs);
    bitset_free(&poseval_docs);
    bitset_free(&negeval_docs);
  }

  /* if we do not allow duplicates, we start the growing of child trees now, after removing
	 already trained classes */

  if (!params->tree_duplicates) {
	filtered_classes = bitset_new_diff(classes, trained);
	for (i=0; i<g_node_n_children(node); i++) {
	  child = g_node_nth_child(node, i);
	  class_docs = mclass_docs[((Method)child->data)->class_id];
	  postrain_docs = bitset_new_intersection(train_docs, class_docs);
	  negtrain_docs = bitset_new_diff(train_docs, class_docs);
	  __pos_neg_classes(score->id, filtered_classes, postrain_docs, mclass_docs, &posclasses, &negclasses);
	  g_node_append(child, mltree_new(f, 
									  posclasses, 
									  postrain_docs, eval_docs, 
									  train_samples, eval_samples,
									  mclass_docs, 
									  params, depth+1));
		
	  g_node_append(child, mltree_new(f, 
									  negclasses, 
									  negtrain_docs, eval_docs, 
									  train_samples, eval_samples,
									  mclass_docs, 
									  params, depth+1));
	  bitset_free(&posclasses);
	  bitset_free(&negclasses);
	  bitset_free(&postrain_docs);
	  bitset_free(&negtrain_docs);
	}
	bitset_free(&filtered_classes);
  }


  bitset_free(&trained);
  g_array_free(scores, TRUE);
  
  return node;
}
/* -------------------------------------------------------------------------- */
MLNode *
mltree_new_eval_partitioning(Fold *f, BitSet classes, 
	   BitSet train_docs, BitSet eval_docs, 
	   Samples train_samples, Samples eval_samples,
	   BitSet *mclass_docs, 
	   FoldTrainParams *params, guint depth) {

  tc_id i, count=0;
  BitSet postrain_docs, poseval_docs, negtrain_docs, negeval_docs, 
    posclasses, negclasses, class_docs, trained, filtered_classes;
  MLNode *node, *child;
  GArray *scores;
  Method method=NULL;
  struct class_score *score;
  gboolean oldverbose;
  guint max_children;

  node = (MLNode*) g_node_new(NULL);
  node->data = NULL;

  /* no more classes to train */
  if (classes->total == 0 || (depth >= params->tree_max_depth && params->tree_max_depth > 0))
    return node;

  if (verbose) {
	printf("[%d/%d]", depth, classes->total);
	fflush(stdout);
  }

  /* 
     We compute the balance for each class and sort with the lowest
     one first
  */
  scores = __get_balances(classes, mclass_docs, train_docs, params);

  /* 
     We train for node class (if we can find a valid method)
  */

  max_children = params->tree_max_children > 0 && params->tree_max_children < scores->len ? 
    params->tree_max_children : scores->len;

  trained = bitset_new(classes->max_id);

  for (i=0; i<max_children && max_children < scores->len; i++) {
    score = &g_array_index(scores, struct class_score, i);
    class_docs = mclass_docs[score->id];
    postrain_docs = bitset_new_intersection(train_docs, class_docs);
    negtrain_docs = bitset_new_diff(train_docs, class_docs);
    poseval_docs =  bitset_new_intersection(eval_docs, class_docs);
    negeval_docs = bitset_new_diff(eval_docs, class_docs);

    /* at least, two documents per set */
    method = NULL;
    if (postrain_docs->total * negtrain_docs->total  *
    	poseval_docs->total * negeval_docs->total) {
      oldverbose = verbose;
      verbose = 0;
      method = fold_train_class(f, 
								score->id, 
								postrain_docs, negtrain_docs, 
								poseval_docs, negeval_docs, 
								params, 
								train_samples, eval_samples);
      verbose = oldverbose;
    }
	
    if (method != NULL) { /* valid method, we add a node */
      
	  tc_message_int("(%d)", score->id);

      /* Find positive and negative classes */	  
      child = (MLNode*) g_node_new(NULL);
      child->data = (gpointer) method_new_clone(method);
      g_node_append(node, (GNode *)child);
	
	  if (params->tree_duplicates) {
		__pos_neg_classes(score->id, classes, postrain_docs, mclass_docs, &posclasses, &negclasses);
		g_node_append(child, mltree_new(f, 
										posclasses, 
										postrain_docs, poseval_docs, 
										train_samples, eval_samples,
										mclass_docs, 
										params, depth+1));
		
		g_node_append(child, mltree_new(f, 
										negclasses, 
										negtrain_docs, negeval_docs, 
										train_samples, eval_samples,
										mclass_docs, 
										params, depth+1));
		bitset_free(&posclasses);
		bitset_free(&negclasses);
	  } else
		bitset_insert(&trained, score->id);
    } else 
	  max_children++;
	
    bitset_free(&postrain_docs);
    bitset_free(&negtrain_docs);
    bitset_free(&poseval_docs);
    bitset_free(&negeval_docs);
  }

  /* if we do not allow duplicates, we start the growing of child trees now, after removing
	 already trained classes */

  if (!params->tree_duplicates) {
	filtered_classes = bitset_new_diff(classes, trained);
	for (i=0; i<g_node_n_children(node); i++) {
	  child = g_node_nth_child(node, i);
	  class_docs = mclass_docs[((Method)child->data)->class_id];
	  postrain_docs = bitset_new_intersection(train_docs, class_docs);
	  negtrain_docs = bitset_new_diff(train_docs, class_docs);
	  poseval_docs =  bitset_new_intersection(eval_docs, class_docs);
	  negeval_docs = bitset_new_diff(eval_docs, class_docs);	
	  __pos_neg_classes(score->id, filtered_classes, postrain_docs, mclass_docs, &posclasses, &negclasses);
	  g_node_append(child, mltree_new(f, 
									  posclasses, 
									  postrain_docs, poseval_docs, 
									  train_samples, eval_samples,
									  mclass_docs, 
									  params, depth+1));
		
	  g_node_append(child, mltree_new(f, 
									  negclasses, 
									  negtrain_docs, negeval_docs, 
									  train_samples, eval_samples,
									  mclass_docs, 
									  params, depth+1));
	  bitset_free(&posclasses);
	  bitset_free(&negclasses);
	  bitset_free(&postrain_docs);
	  bitset_free(&negtrain_docs);
	  bitset_free(&poseval_docs);
	  bitset_free(&negeval_docs);
	}
	bitset_free(&filtered_classes);
  }


  bitset_free(&trained);
  g_array_free(scores, TRUE);
  
  return node;
}
/* -------------------------------------------------------------------------- */
/* 
   This is the recursive function for classification in a ML-Tree.
   The vector of weights for classes 'cr' will be actualized with either positive of negative values.
*/
void __node_classify(MLNode *node, Sample s, Fold f, ClassRank *cr, gdouble accum_weight, FoldTrainParams *params) {
  Method m;
  FILE *file;
  double val, posweight, negweight;
  void *c;
  guint i;

  if (!node || !node->data && __is_leaf(node))
    return;

  tc_message(".");

  if (!node->data) {
    /* Father node */
    for (i=0; i<g_node_n_children(node); i++)
      __node_classify((MLNode *)g_node_nth_child(node, i), s, f, cr, accum_weight, params);
  } else { 
    /* Method node */    
    m = (Method)node->data;
    val = 1.0;
    if (m->classifier == CLASSIFIER_COMMAND) {
      method_exec_cmd(m, f->filepath_sample_tmp, f->filepath_predictions);
      file = fopen(f->filepath_predictions, "r");
      fscanf(file, "%lf", &val);
      fclose(file);
      remove(f->filepath_predictions);
    } else {
      c = classifier_new_read[m->classifier](m->model_file);
      val = classifier_classify[m->classifier](c, s);
      classifier_free[m->classifier](&c);
    }

	/* probability of being positive */
	posweight = accum_weight * m->recall; 
	/* probability of being negative */
	negweight = accum_weight * ((double)m->tn)/((double)(m->tn+m->fp)); 
    if (val > 0.0)
      classrank_add(cr, m->class_id, posweight);    
    else
      classrank_add(cr, m->class_id, -negweight);
    
    if (val > 0.0 || params->tree_traverse_all)
      __node_classify((MLNode *)g_node_nth_child(node, 0), s, f, cr,
		      posweight,
		      params);
    if (val < 0.0  || params->tree_traverse_all)
      __node_classify((MLNode *)g_node_nth_child(node, 1), s, f, cr,
		      negweight,
		      params);
  }
}
/* -------------------------------------------------------------------------- */
/* This function returns a vector of weights (one for each class).
   The returned weights must be deallocated after use. */

BitSet mltree_classify(MLNode *node, Sample s, Fold f, guint ntop, FoldTrainParams *params) {
  FILE *fout;
  ClassRank cr;
  BitSet classes;
  Sample rank;
  
  cr = classrank_new(f->class_info->len);

  /* we create sample file */
  fout = fopen (f->filepath_sample_tmp, "w");
  if (!fout)
    tc_error("mltree_classify: unable to create temporal file");
  fprintf(fout, "+0 "); /* no assignment */
  sample_serialize(s, fout, FALSE);
  fclose(fout);

  /* classify */
  __node_classify(node, s, f, &cr, 1.0, params);
  remove (f->filepath_sample_tmp);

  classrank_avg(&cr);  
  rank = classrank_get_top(cr, ntop);
  classrank_free(&cr);
  classes = sample_to_bitset(s);
  sample_free(&rank);

  return classes;
}
/* -------------------------------------------------------------------------- */
gboolean __free_method (GNode *node, gpointer data) {
  if (node && node->data)
    method_free((Method*)&node->data);  
  return FALSE;
}
/* -------------------------------------------------------------------------- */
void mltree_free(MLNode *node) {

  /* first, we remove methods */
  if (node) {
    g_node_traverse (node,
		     G_PRE_ORDER,
		     G_TRAVERSE_ALL,
		     -1,
		     __free_method,
		     NULL);

    /* now, we remove the tree */ 
    g_node_destroy(node);
  }
}
/* -------------------------------------------------------------------------- */
void __save_node(GNode *node, FILE *f) {
  static guint node_count=0;
  guint i;
  MLNode *child;

  if (node->data == NULL && __is_leaf(node))
    fprintf(f, "-1\n");
  else {
    fprintf(f, "%d\n", node_count++);
    for (i=0; i<g_node_n_children(node); i++) {    
      child = g_node_nth_child(node, i);
      fprintf(f, "%d\n", node_count++);
      method_serialize((Method) child->data, f);
      __save_node(g_node_nth_child(child, 0), f);
      __save_node(g_node_nth_child(child, 1), f);
    }
    fprintf(f, "-2\n");
  }
}
/* -------------------------------------------------------------------------- */
void mltree_save(MLNode *node, gchar *filename) {
  FILE *f = fopen(filename, "w");
  if (!f)
    tc_error("mltree_save: unable to create file");
  __save_node(node, f);
  fclose(f);
}
/* -------------------------------------------------------------------------- */
MLNode *__read_node(FILE *f) {
  gint id;
  MLNode *node, *child, *child1, *child2;
  gboolean children_left=TRUE;

  node = (MLNode*) g_node_new(NULL);
  node->data = NULL;

 if (feof(f))
    return node;
  fscanf(f, "%d", &id);
  if (id == -1)
    return node;
  
  while(children_left) {
    fscanf(f, "%d", &id);
    if (id == -2)
      children_left = FALSE;
    else {
      child = (MLNode*) g_node_new(NULL);
      child->data = (gpointer)method_new_deserialize(f);
      g_node_append(node, child);
      g_node_append(child, __read_node(f));
      g_node_append(child, __read_node(f));
    }
  }
 
  return node;
}
/* -------------------------------------------------------------------------- */
MLNode *mltree_new_read(gchar *filename) {  
  MLNode *node;

  FILE *f = fopen(filename, "r");
  if (!f)
    tc_error("mltree_new_read: unable to open file");
  node = __read_node(f);
  fclose(f);
  return node;
}
/* -------------------------------------------------------------------------- */
void __dump_node(MLNode *node, gchar *prefix, GArray *classes) {
  gchar *prefix_sig;
  MLNode *node0, *node1, *child;
  guint i, val=8;

  if (!node || !node->data && __is_leaf(node))
    return;

  if (node->data == NULL) {

    for (i=0; i<g_node_n_children(node); i++) {    
      printf("%s`-- ", prefix);
      child = (MLNode *)g_node_nth_child(node, i);      
      if ((GNode*)child != g_node_last_child(node))
		prefix_sig = g_strconcat(prefix, "|    ", NULL);
      else
		prefix_sig = g_strconcat(prefix, "     ", NULL);
      __dump_node(child, prefix_sig, classes);
    } 
  }
  else {
	if (((Method)child->data)->class_id >= classes->len)
	  classes = g_array_set_size(classes, ((Method)child->data)->class_id+1);
	g_array_index(classes, guint, ((Method)child->data)->class_id) =  val;
    printf("%d (%.4f/%.4f)\n", 
	   ((Method)child->data)->class_id, 
	   ((Method)child->data)->precision, 
	   ((Method)child->data)->recall);
    
    node0 = g_node_nth_child(node, 0);
    node1 = g_node_nth_child(node, 1);

    if (node1 && !__is_leaf(node1))
      prefix_sig = g_strconcat(prefix, "|    ", NULL);
    else
      prefix_sig = g_strconcat(prefix, "     ", NULL);

    if (node0 &&  !__is_leaf(node0)) {
      printf("%s`-- (+)\n", prefix);
      __dump_node(node0, prefix_sig, classes);
    }
 
    g_free(prefix_sig);
    prefix_sig = g_strconcat(prefix, "     ", NULL);

    if (node1 &&  !__is_leaf(node1)) {
      printf("%s`-- (-)\n", prefix);
      __dump_node(node1, prefix_sig, classes);
    }
  } 
  g_free(prefix_sig);  
}
/* -------------------------------------------------------------------------- */
void mltree_dump(MLNode *node) {
  guint i, count=0;
  GArray *classes = g_array_new(FALSE, TRUE, sizeof(guint));

  printf("ROOT\n");
  __dump_node(node, "", classes);

  for (i=0; i<classes->len; i++) {
	if (g_array_index(classes, guint, i) != 0)
	  count++;   
  }
  g_array_free(classes, TRUE);
  printf("%d classes\n", count);
}
