#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <errno.h>
#include <error.h>
#include <math.h>

#include "plaum.h"
#include "feature.h"
#include "sample.h"
#include "samples.h"
#include "weights.h"

/* --------------------------------------------------------------------
  This is an implementation of the Perceptron Learning Algorithm with 
  Uneven Margins. For information about this algorithm read:
 
 -  Li. Y. and Zaragoza H., Herbrich R., Shawe-Taylor J. and Kandola J.
    "The Perceptron Algorithm with Uneven Margins", Proceedings of the
    International Conference of Machine Learning (ICML'2002), 2002.

 ------------------------------------------------------------------- */

extern gboolean verbose;

/* -------------------------------------------------------------------- */
void *plaum_new(gchar *args)
{
  Plaum c = (Plaum) g_new0(struct plaum_struct, 1);
  gint i = 1, len, pos=0;
  gchar **arg_val = NULL, **argv;

  if (!c)
    tc_error("plaum_new: No memory for plaum classifier\n");

  /* default values */
  c->T = 100;
  c->neg_tau = -1.0;
  c->pos_tau = +1.0;
  c->eta = 1.0;
  c->lambda = 0.0;

  /* no arguments defined, we return it with default values */
  if (!args)
    return ((void*)c);
  
  len = strlen(args);
  argv = g_strsplit(args, ":", 0);
  pos = strlen(argv[0]);
  if (len == pos)       /* no arguments */
    return ((void*)c);

  /* user defined values */  
  while (pos < len) {
    arg_val = g_strsplit(argv[i], "=", 2);
    if (!strncmp(arg_val[0], "iter", 4)) /* max iterations */
      c->T = atoi(arg_val[1]);
    else if (strncmp(arg_val[0], "neg", 3)) /* negative tau */
      c->neg_tau = atof(arg_val[1]);
    else if (strncmp(arg_val[0], "pos", 3)) /* positive tau */
      c->pos_tau = atof(arg_val[1]);
    else if (strncmp(arg_val[0], "eta", 3)) /* eta */
      c->eta = atof(arg_val[1]);
    else if (strncmp(arg_val[0], "lambda", 6)) /* eta */
      c->lambda = atof(arg_val[1]);
    else
      tc_error("Invalid arguments for plaum");
    pos += strlen(argv[i++])+1; /* to control not to exceed the number of arguments */
    g_strfreev(arg_val);
  }
  g_strfreev(argv);
  return ((void*)c);
}

/* -------------------------------------------------------------------- */
void plaum_free(void **cp) {
  Plaum c = (Plaum) *cp;
  if (c) {
    weights_free(&c->w);
    g_free(c);
  }
}
/* -------------------------------------------------------------------- */
void plaum_save(void* cp, gchar *file) {
  Plaum c = (Plaum)cp;
  FILE *f;
  if (!c)
    tc_error("plaum_save: plaum structure not initialized");  
  if (!(f = fopen(file, "w")))
    tc_error_str("plaum_save: unable to create file '%s'", file);
  fprintf(f, "%lf\n", c->b);
  weights_save(c->w, f);
  fclose(f);
}
/* -------------------------------------------------------------------- */
void plaum_append_save(void* cp, gchar *file) {
  Plaum c = (Plaum)cp;
  FILE *f;
  if (!c)
    tc_error("plaum_append_save: plaum structure not initialized");  
  if (!(f = fopen(file, "a+")))
    tc_error_str("plaum_save: unable to create file '%s'", file);
  fprintf(f, "%lf\n", c->b);
  weights_save(c->w, f);
  fclose(f);
}
/* -------------------------------------------------------------------- */
void plaum_save_ascii(void* cp, gchar *file, IntStrTable term_str, GArray *idid) {
  Plaum c = (Plaum)cp;
  FILE *f;
  if (!c)
    tc_error("plaum_save: plaum structure not initialized");  
  if (!(f = fopen(file, "w")))
    tc_error_str("plaum_save: unable to create file '%s'", file);
  printf("SAVING %s\n", file);
  fprintf(f, "PARAMS --------\nb\t%lf\nWEIGHTS -------\n", c->b);
  weights_save_ascii(c->w, f, term_str, idid);
  fclose(f);
}
/* -------------------------------------------------------------------- */
void *plaum_new_read(gchar *file) {
  Plaum c;
  FILE *f = fopen(file, "r");
  if (!f)
    tc_error_str("plaum_new_read: unable to open file '%s'", file);
  c = (Plaum) plaum_new(NULL);
  fscanf(f, "%lf\n", &c->b);
  c->w = weights_new_read(f);
  fclose(f);
  return c;
}
/* -------------------------------------------------------------------- */
void *plaum_new_read_at(FILE *f) {
  Plaum c;
  //  if (!f)
  //  tc_error_str("plaum_new_read: unable to open file '%s'", file);
  c = (Plaum) plaum_new(NULL);
  //fseek (f,offset,SEEK_SET);
  fscanf(f, "%lf\n", &c->b);
  c->w = weights_new_read(f);
  return c;
}
/* -------------------------------------------------------------------- */
/* This function returns <w, x_i> + b */
double plaum_classify(void *cp, Sample s) {
  return weights_product_by_sample(((Plaum)cp)->w, s) + ((Plaum)cp)->b;
}

/* -------------------------------------------------------------------- */
void plaum_dump (void *cp)
{
  Plaum c = (Plaum)cp;
  printf ("pos_tau: %f\n", c->pos_tau);
  printf ("neg_tau: %f\n", c->neg_tau);
  printf ("T: %d\n", c->T);
  printf ("eta: %f\n", c->eta);
  printf ("b: %f\n", c->b);
  printf ("lambda: %f\n", c->lambda);
}
/* PLAUM learning ----------------------------------------------------- */

/* -------------------------------------------------------------------- */
/* This product adds to weights w the product eta*y*x */
void __add_product (Weights * w, double etay, Sample s)
{
  Feature t;
  sample_it_reset (&s);
  while (!sample_it_end (s)) {
	t = sample_it_next (&s);
	weights_add_weight (w, t->id, etay * t->w);
  }
}
/* -------------------------------------------------------------------- */
void plaum_learn(void **cp, Samples ss)
{
  Plaum c = (Plaum) *cp;
  Sample s;
  double tau;		/* tau to compare to */
  gboolean updated = TRUE;
  int epoch = 0, doc_i=0;
  int *alpha;                           /* counter for weight updates (for lambda-trick) */

  //plaum_dump(c);
  c->w = weights_new (ss->max_feature_id+ss->total);
  c->b = 0.0;

  /* We create the update counter for lambda-trick. The solution is to
	 implement the product as: 

	 <w,x_i>_\lambda = <w,x_i> + \lambda*\eta*\alpha_i

	 where \alpha_i counts the positive assignments minus negative
	 ones for sample i.
  */
  alpha = g_new0(gint, ss->total);
  
  while (updated && (epoch < c->T)) {
    updated = FALSE;
    samples_it_reset (&ss);
	doc_i=0;
    while (!samples_it_end (ss)) {
      s = samples_it_next (&ss);
      tau = s->y < 0.0 ? -c->neg_tau : c->pos_tau;
      if (s->y * (plaum_classify(*cp, s) + c->lambda*c->eta*alpha[doc_i]) <= tau) {
		__add_product (&c->w, c->eta * s->y, s);
		c->b += c->eta * s->y * ss->r2;
		updated = TRUE;
		alpha[doc_i] += s->y;
      }
	  doc_i++;
    }
    epoch++;
  }
  g_free(alpha);
  //printf("[builtin] done in %d iterations, %d samples\n", epoch, ss->total);
}
