#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <errno.h>
#include <error.h>
#include <math.h>

#include "rocchio.h"
#include "feature.h"
#include "sample.h"
#include "samples.h"
#include "weights.h"

extern gboolean verbose;

/* -------------------------------------------------------------------- */
void *rocchio_new(gchar *args)
{
  Rocchio c = (Rocchio) g_new0(struct rocchio_struct, 1);
  gint i = 1, len, pos=0;
  gchar **arg_val = NULL, **argv;

  if (!c)
    tc_error("rocchio_new: No memory for rocchio classifier\n");

  /* default values */
  c->alpha = 0.0;
  c->beta = 1.0;
  c->gamma = 1.0;
  c->nonnegative = TRUE;
  c->t = 0.0;

  /* no arguments defined, we return it with default values */
  if (!args)
    return ((void*)c);
  
  len = strlen(args);
  argv = g_strsplit(args, ":", 0);
  pos = strlen(argv[0]);
  if (len == pos)       /* no arguments */
    return ((void*)c);

  /* user defined values */  
  while (pos < len) {
    arg_val = g_strsplit(argv[i], "=", 2);
    if (!strncmp(arg_val[0], "alpha", 5)) 
      c->alpha = atof(arg_val[1]);
    else if (!strncmp(arg_val[0], "beta", 4)) 
      c->beta = atof(arg_val[1]);
    else if (!strncmp(arg_val[0], "gamma", 5)) 
		c->gamma = atof(arg_val[1]);
    else if (!strncmp(arg_val[0], "nonnegative", 5)) 
      c->nonnegative = (atoi(arg_val[1]) == 1);
    else
      tc_error("Invalid arguments for rocchio");
    pos += strlen(argv[i++])+1; /* to control not to exceed the number of arguments */
    g_strfreev(arg_val);
  }
  g_strfreev(argv);
  return ((void*)c);
}

/* -------------------------------------------------------------------- */
void rocchio_free(void **cp) {
  Rocchio c = (Rocchio) *cp;
  if (c) {
    weights_free(&c->w);
    g_free(c);
  }
}
/* -------------------------------------------------------------------- */
void rocchio_save(void* cp, gchar *file) {
  Rocchio c = (Rocchio) cp;
  FILE *f;
  if (!c)
    tc_error("rocchio_save: rocchio structure not initialized");  
  if (!(f = fopen(file, "w")))
    tc_error_str("rocchio_save: unable to create file '%s'", file);
  fprintf(f, "%lf\n", c->t);
  weights_save(c->w, f);
  fclose(f);
}
/* -------------------------------------------------------------------- */
void *rocchio_new_read(gchar *file) {
  Rocchio c;
  FILE *f = fopen(file, "r");
  if (!f)
    tc_error_str("rocchio_new_read: unable to open file '%s'", file);
  c = (Rocchio) rocchio_new(NULL);
  fscanf (f, "%lf", &c->t);
  c->w = weights_new_read(f);
  fclose(f);
  return c;
}
/* -------------------------------------------------------------------- */
double rocchio_classify(void *cp, Sample s) {
  return weights_product_by_sample(((Rocchio) cp)->w, s);
}
/* -------------------------------------------------------------------- */
void rocchio_dump (void *cp)
{
  Rocchio c = (Rocchio) cp;
  printf ("alpha: %f\n", c->alpha);
  printf ("beta: %f\n", c->beta);
  printf ("gamma: %f\n", c->gamma);
  if (c->nonnegative)
    printf("TRUE\n");
  else
    printf("FALSE\n");
}
/* ROCCHIO learning ----------------------------------------------------- */

/* -------------------------------------------------------------------- */
void rocchio_learn(void **cp, Samples ss)
{
  Rocchio c = (Rocchio) *cp;
  Sample s;
  Feature t;
  guint nc = 0, i;
  Weights sum1, sum2;
  double value;

  //rocchio_dump(c);
  c->w = weights_new (ss->max_feature_id);
  sum1 = weights_new (ss->max_feature_id);
  sum2 = weights_new (ss->max_feature_id);

  /* compute centroids of positive (sum1) and negative (sum2) samples */

  samples_it_reset (&ss);
  while (!samples_it_end (ss)) {
    s = samples_it_next (&ss);
	if (s->y > 0) {
	  nc++;
	  sample_it_reset(&s);
	  while (!sample_it_end(s)) {
		t = sample_it_next(&s);
		weights_add_weight(&sum1, t->id, t->w);
      }
	}
	else {
	  sample_it_reset(&s);
	  while (!sample_it_end(s)) {
		t = sample_it_next(&s);
		weights_add_weight(&sum2, t->id, t->w);
      }
	}
  }
  
  /* compute weights */
  for (i=0; i<c->w->len; i++) {
    value = c->alpha * weights_get(c->w, i) +
      c->beta * weights_get(sum1, i) / ((double) nc) -
      c->gamma * weights_get(sum2, i) / ((double) ss->total - nc);
	
    /* tipycally, Rocchio weights are nonnegative, so negative
       computed values are set to zero, but the user can modify it
       with the nonnegative option */
	
    value = (value < 0.0) && c->nonnegative ? 0.0 : value;
    weights_set(&c->w, i, value);
  }
  
  weights_free(&sum1);
  weights_free(&sum2);
  
  /* threshold */
  c->t = 0; //weights_threshold(c->w, ss);
}
