#include <math.h>
#include "tc_utils.h"
#include "weights.h"
#include "sample.h"

/* -------------------------------------------------------------------- */
Weights weights_new(guint max_index) {
  Weights w = g_array_new(FALSE, TRUE, sizeof(double));
  if (w) w = g_array_set_size(w, max_index+1);
  if (!w) 
    tc_error("No memory for weights");
  return w;
}
/* -------------------------------------------------------------------- */
Weights weights_new_from_sample(Sample s) {
  Weights w = g_array_new(FALSE, TRUE, sizeof(double));
  Feature t;
  if (w) 
    w = g_array_set_size(w, s->max_id+1);
  if (!w) 
    tc_error("No memory for weights");
  sample_it_reset(&s);
  while (!sample_it_end(s)) {
    t = sample_it_next(&s);
    weights_set(&w, t->id, t->w);
  }
  return w;
}
/* -------------------------------------------------------------------- */
Weights weights_clone(Weights w) {
  Weights w2 = g_array_new(FALSE, TRUE, sizeof(double));
  int i;
  Feature t;
  if (w2) 
    w2 = g_array_set_size(w2, w->len);
  if (!w2) 
    tc_error("No memory for weights");
  for (i=0; i<w->len; i++)
    g_array_index(w2, double, i) = g_array_index(w, double, i);
  return w2;
}
/* -------------------------------------------------------------------- */
void weights_resize(Weights *w, guint max_index) {
  if ((*w)->len <= max_index)
    *w = g_array_set_size(*w, max_index+1);
}
/* -------------------------------------------------------------------- */
void weights_free (Weights * w)
{
  g_array_free (*w, TRUE);
}
/* -------------------------------------------------------------------- */
void weights_dump (Weights w)
{
  guint i;
  for (i = 0; i < w->len; i++)
    printf("(%d,%.4f)", i, g_array_index (w, double, i));
  printf("\n");
}
/* -------------------------------------------------------------------- */
inline void weights_set(Weights *w, guint index, double value) {
  if (index >= (*w)->len)
    fprintf(stderr, "weights_set: index (%d) out of range (%d)\n", index, (*w)->len-1);
  g_array_index(*w, double, index) = value;
}
/* -------------------------------------------------------------------- */
inline double weights_get(Weights w, guint index) {
  if (index >= w->len)
    return 0.0;
  return g_array_index(w, double, index);
}
/* -------------------------------------------------------------------- */
void weights_add_weight(Weights *w, guint index, double value) {
  weights_resize(w, index);
  g_array_index(*w, double, index) += value;
}
/* -------------------------------------------------------------------- */
double weights_product_by_sample(Weights w, Sample s) {
  guint i;
  double product=0.0;
  Feature t;

  sample_it_reset(&s);
  while (!sample_it_end(s)) {
    t = sample_it_next(&s);
    product += weights_get(w, t->id) * t->w;
  }
  return product;
}
/* -------------------------------------------------------------------- */
void weights_product_weights(Weights *w, Weights w2) {
  guint i;

  if ((*w)->len != w2->len)
    tc_error("weights_product_weights: different vector lengths");
  for (i=0; i<(*w)->len; i++)
    g_array_index(*w, double, i) *= g_array_index(w2, double, i);
}
/* -------------------------------------------------------------------- */
void weights_divide_weights(Weights *w, Weights w2) {
  guint i;

  if ((*w)->len != w2->len)
    tc_error("weights_product_weights: different vector lengths");
  for (i=0; i<(*w)->len; i++)
    g_array_index(*w, double, i) /= g_array_index(w2, double, i);
}
/* -------------------------------------------------------------------- */
double weights_distance_to_sample(Weights w, Sample s) {
  double dist=0.0, a;
  Feature t;
  guint i=0;

  sample_it_reset(&s);
  while (!sample_it_end(s)) {
    t = sample_it_next(&s);
    for (;i<t->id;i++) {      
      a =  weights_get(w,i);
      if (a) dist += a*a;
    } 
    a = weights_get(w,i++);
    dist += (a-t->w)*(a-t->w);
  }
  for (;i<w->len;i++) {
    a =  weights_get(w,i);
    if (a) dist += a*a;
  }
  // printf(" %lf %lf %f %d\n", dist, sqrt(dist), dist, w->len);

  return sqrt(dist);
}
/* -------------------------------------------------------------------- */
void weights_set_all(Weights *w, double value) {
  guint i;
  for (i=0; i<(*w)->len; i++)
    g_array_index(*w, double, i) = value;
}
/* -------------------------------------------------------------------- */
void weights_add_weights(Weights *w, Weights w2) {
  guint i;

  if ((*w)->len != w2->len)
    tc_error("weights_add_weights: different vector lengths");
  for (i=0; i<(*w)->len; i++)
    g_array_index(*w, double, i) += g_array_index(w2, double, i);
}
/* -------------------------------------------------------------------- */
void weights_product_by_scalar(Weights *w, double v) {
  guint i;
  for (i=0; i<(*w)->len; i++)
    g_array_index(*w, double, i) *= v;
}
/* -------------------------------------------------------------------- */
void weights_save(Weights w, FILE *f)
{
  guint i;
  fprintf(f, "%d\n", w->len);
  for (i = 0; i < w->len; i++)
    fprintf(f, "%lf ", weights_get(w, i));
  fprintf(f, "\n");
}
/* -------------------------------------------------------------------- */
void weights_save_ascii(Weights w, FILE *f, IntStrTable item_str, GArray *idid)
{
  guint i;
  gdouble val;

  for (i = 0; i < w->len; i++) {
    val = weights_get(w, i);    
    if (val > 0.0)
      fprintf(f, "%s\t%lf\n", intstr_table_get(item_str, g_array_index(idid, guint, i)), weights_get(w, i));
  }
  printf("done\n");
}
/* -------------------------------------------------------------------- */
Weights weights_new_read(FILE *f)
{
  Weights w;
  guint len, i;

  fscanf(f, "%d", &len);
  w = weights_new(len-1);
  for (i = 0; i < len; i++)
    fscanf(f, "%lf", &g_array_index(w, double, i));
  //weights_dump(w);
  return w;  
}
/* -------------------------------------------------------------------- */
double weights_threshold(Weights w, Samples ss) {
  Sample s;
  double pos=0.0, neg=0.0, value;
  guint npos=0, nneg=0;

  /* compute threshold */
  samples_it_reset (&ss);
  while (!samples_it_end (ss)) {
    s = samples_it_next (&ss);
    value = weights_product_by_sample(w, s);
    if (s->y > 0.0) {
      pos += value;
      npos++;
    }
    else {
      neg += value;
      nneg++;
    }
  }
  return (pos/(double)npos + neg/(double)nneg)/2.0;
}
