#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <glib.h>
#include <errno.h>
#include <error.h>
#include <math.h>

#include "widrow_hoff.h"
#include "feature.h"
#include "sample.h"
#include "samples.h"
#include "weights.h"

extern gboolean verbose;

/* -------------------------------------------------------------------- */
void *widrow_hoff_new(gchar *args)
{
  WidrowHoff c = (WidrowHoff) g_new0(struct widrow_hoff_struct, 1);
  gint i = 1, len, pos=0;
  gchar **arg_val = NULL, **argv;

  if (!c)
    tc_error("widrow_hoff_new: No memory for widrow_hoff classifier\n");

  /* default values */
  c->eta = -1.0; /* will be translated when learning to default value */
  c->average = TRUE;
  c->t = 0.0;

  /* no arguments defined, we return it with default values */
  if (!args)
    return ((void*)c);
  
  len = strlen(args);
  argv = g_strsplit(args, ":", 0);
  pos = strlen(argv[0]);
  if (len == pos)       /* no arguments */
    return ((void*)c);

  /* user defined values */  
  while (pos < len) {
    arg_val = g_strsplit(argv[i], "=", 2);
    if (!strncmp(arg_val[0], "eta", 3)) {
      c->eta = atof(arg_val[1]);
      if (c->eta <= 0.0)
	tc_error("Invalid value for eta in Widrow-Hoff algorithm: must be positive");
    }
    else if (!strncmp(arg_val[0], "average", 7))
      c->average = (atoi(arg_val[1]) == 1);
    else
      tc_error("Invalid arguments for widrow_hoff");
    pos += strlen(argv[i++])+1; /* to control not to exceed the number of arguments */
    g_strfreev(arg_val);
  }
  g_strfreev(argv);
  return ((void*)c);
}

/* -------------------------------------------------------------------- */
void widrow_hoff_free(void **cp) {
  WidrowHoff c = (WidrowHoff) *cp;
  if (c) {
    weights_free(&c->w);
    g_free(c);
  }
}
/* -------------------------------------------------------------------- */
void widrow_hoff_save(void* cp, gchar *file) {
  WidrowHoff c = (WidrowHoff) cp;
  FILE *f;
  if (!c)
    tc_error("widrow_hoff_save: widrow_hoff structure not initialized");  
  if (!(f = fopen(file, "w")))
    tc_error_str("widrow_hoff_save: unable to create file '%s'", file);
  fprintf(f, "%lf\n", c->t);
  weights_save(c->w, f);
  fclose(f);
}
/* -------------------------------------------------------------------- */
void *widrow_hoff_new_read(gchar *file) {
  WidrowHoff c;
  FILE *f = fopen(file, "r");
  if (!f)
    tc_error_str("widrow_hoff_new_read: unable to open file '%s'", file);
  c = (WidrowHoff) widrow_hoff_new(NULL);
  fscanf (f, "%lf", &c->t);
  c->w = weights_new_read(f);
  fclose(f);
  return c;
}
/* -------------------------------------------------------------------- */
double widrow_hoff_classify(void *cp, Sample s) {
  WidrowHoff c = (WidrowHoff) cp;
  return weights_product_by_sample(c->w, s);
}
/* -------------------------------------------------------------------- */
void widrow_hoff_dump (void *cp)
{
  WidrowHoff c = (WidrowHoff) cp;
  printf ("eta: %f\n", c->eta);
  printf ("averaged weights: ");
  if (c->average)
    printf("TRUE\n");
  else
    printf("FALSE\n");
}
/* WIDROW_HOFF learning ----------------------------------------------------- */

/* -------------------------------------------------------------------- */
void widrow_hoff_learn(void **cp, Samples ss)
{
  WidrowHoff c = (WidrowHoff) *cp;
  Sample s;
  Feature t;
  Weights prev, next, accum, tmp;
  double dot, yi;

  //widrow_hoff_dump(c);

  if (c->eta == -1.0)
    c->eta = 1.0/(4.0*ss->r2); /* default */

  prev = weights_new (ss->max_feature_id);
  next = weights_new (ss->max_feature_id);
  if (c->average)
    accum = weights_new (ss->max_feature_id);

  samples_it_reset (&ss);
  while (!samples_it_end (ss)) {
    s = samples_it_next (&ss);
    dot = weights_product_by_sample(prev, s);
	yi = s->y < 0.0 ? 0.0 : s->y; /* because this algorithm considers lables {1,0} */
    sample_it_reset(&s);
    while (!sample_it_end(s)) {
      t = sample_it_next(&s);	  
      weights_set(&next, t->id, weights_get(prev, t->id) - 2.0 * c->eta * (dot - yi) * t->w);      
    }
    if (c->average)
      weights_add_weights(&accum, next);
    tmp = next;
    next = prev;
    prev = tmp;
  }

  if (c->average) {
    weights_product_by_scalar(&accum, 1.0/((double)ss->total));
    c->w = accum;
	weights_free(&prev);
  }
  else
    c->w = prev;    
  weights_free(&next);  

  /* threshold */
  c->t = 0.0; //weights_threshold(c->w, ss);
}
