/***************************************************************************
 *            fold.c
 *
 *  Thu Dec 11 19:42:10 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/

#include <glib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include "fold.h"
#include "chrono.h"
#include "classifiers/plaum.h"
#include "classifiers/mltree.h"
#include "classifiers/classrank.h"
#include "intstr_table.h"

#define STR_BUFFER_LEN 512

#define DO_CLASS_FILTERING       1<<1
#define DO_SAMPLE_CLASSIFICATION 1<<2
#define ONLY_POSITIVES           1<<3
#define ALSO_NEGATIVES           1<<4

extern gboolean verbose;

inline int file_exists(gchar *filename);

Fold
__fold_new (gchar * path, guint i, InvertedIndex folds)
{
  Fold fold = g_new0 (struct fold_struct, 1);
  gchar fold_dir[10];

  fold->id = i;

  /* paths */
  sprintf (fold_dir, "fold%d", i);
  fold->path = g_strconcat (path, PATH_CONNECTOR, fold_dir, NULL);
  fold->filepath_train_samples = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_TRAIN_SAMPLES, NULL);
  fold->filepath_eval_samples = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_EVAL_SAMPLES, NULL);
  fold->filepath_test_samples = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_TEST_SAMPLES, NULL);
  fold->filepath_train_samples_tmp = g_strconcat (fold->filepath_train_samples, ".tmp", NULL);
  fold->filepath_eval_samples_tmp = g_strconcat (fold->filepath_eval_samples, ".tmp", NULL);
  fold->filepath_test_samples_tmp = g_strconcat (fold->filepath_test_samples, ".tmp", NULL);
  fold->filepath_class_info = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_CLASS_INFO, NULL);
  fold->filepath_term_info = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_TERM_INFO, NULL);
  fold->filepath_sample_tmp = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_SAMPLE_TMP, NULL);
  fold->filepath_predictions = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_PREDICTIONS, NULL);
  fold->path_models = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_DIR_MODELS, NULL);
  fold->filepath_methods_eval = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_METHODS_EVAL, NULL);
  fold->filepath_methods_filter = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_METHODS_FILTER, NULL);
  if (!g_file_test(fold->filepath_methods_filter, G_FILE_TEST_EXISTS))
	fold->filepath_methods_filter = NULL;
  fold->filepath_methods_test = g_strconcat (fold->path, PATH_CONNECTOR, FOLD_FILE_METHODS_TEST, NULL);
  fold_get_docs (&fold, folds);

  fold->methods = NULL;
  fold->class_info = NULL;
  fold->term_info = NULL;
	
  return fold;
}
/* -------------------------------------------------------------------------- */
/* Here weights are computed and new samples generated with new term ids */
void
__fold_weight_features (FILE * fdocsv, TermInfoArray* term_info, GArray * doc_info,
			gchar * newdocsv, BitSet docs, FoldTrainParams *params)
{
  FILE *fhnd = fopen (newdocsv, "w");
  tc_id d;
  Sample s, new_s;

  bitset_it_reset(&docs);
  while (!bitset_it_end(docs))
    {
      d = bitset_it_next(&docs);
      //s = sample_new_deserialize(fdocsv, doc_info_get(doc_info, d)->offset, TRUE);
      s = sample_new_deserialize(fdocsv, doc_info_get(doc_info, d)->offset, FALSE);
      new_s = sample_new_reweight(s, *term_info, docs, params->ftrans);
      sample_serialize(new_s, fhnd, FALSE); /* we save in ASCII format now */
      sample_free(&s);
      sample_free(&new_s);
    }

  /* Finalization */
  fclose (fhnd);
}

/* -------------------------------------------------------------------------- */
Fold
fold_new (Model m, InvertedIndex folds, guint i, FoldTrainParams *params)
{
  Fold fold = __fold_new (m->datadir, i, folds);

  /* we create fold directory */
  if (!g_file_test (fold->path, G_FILE_TEST_IS_DIR))
    mkdir (fold->path, 0750);

  /* 
     Compute information and reduce features and classes 
  */
	
  fold->term_info = term_info_new_compute(m->term_doc, m->class_doc, fold->train_docs, params->fig_size);
  term_info_reduce (&fold->term_info, params->fig_size, params->fdf_min);	
  fold->class_info = class_info_new_compute(m->class_doc, fold->train_docs, fold->eval_docs, fold->test_docs);
  class_info_reduce (&fold->class_info, params->cdf_min);

  /* If entropy is used we have to compute it before */
  if (params->ftrans & TC_WEIGHT_ENTROPY)
    term_info_compute_entropy(m->fdocsv, &fold->term_info, m->doc_info, fold->train_docs);
	
  /* Now we recompute weights */
  tc_message ("\tWeighting and saving features...");
  __fold_weight_features (m->fdocsv, &fold->term_info, m->doc_info, fold->filepath_train_samples, fold->train_docs, params);
  __fold_weight_features (m->fdocsv, &fold->term_info, m->doc_info, fold->filepath_eval_samples,  fold->eval_docs, params);
  __fold_weight_features (m->fdocsv, &fold->term_info, m->doc_info, fold->filepath_test_samples,  fold->test_docs, params);
  tc_message("done\n");

  /* save term info vector */	
  term_info_save(fold->term_info, fold->filepath_term_info);

  /* We save classes only at this point, in the trainig phases
     also some classes will be discarded if the method doesn't
     give any results, but since it could be different for
     another training run, we don't save which were the classes
     removed at training. */
  class_info_save (fold->class_info, fold->filepath_class_info);
  return fold;
}

/* -------------------------------------------------------------------------- */
/* 
   Sets of documents
   For fold 'i' 
   i+1, ..., i+6 are for training
   i+7, ..., i+9 are for evaluation
   i is for testing
*/
void
fold_get_docs (Fold * f, InvertedIndex folds)
{
  guint j;
	
  guint ntrain = (guint) ((double) folds->max_id / 3.0 * 2.0);
  guint neval = folds->max_id - ntrain;
	
  (*f)->train_docs = inverted_index_get (folds, ((*f)->id + 1) % (folds->max_id + 1));
  (*f)->eval_docs = inverted_index_get (folds, ((*f)->id + 1 + ntrain) % (folds->max_id + 1));
  (*f)->test_docs = inverted_index_get (folds, (*f)->id);
  for (j = 2; j <= ntrain; j++)
    {
      BitSet add_docs = inverted_index_get (folds, ((*f)->id + j) % (folds->max_id + 1));
      bitset_add_set (&(*f)->train_docs, add_docs);
      bitset_free (&add_docs);
    }
  for (j = j + 1; j <= ntrain + neval; j++)
    {
      BitSet add_docs = inverted_index_get (folds, ((*f)->id + j) % (folds->max_id + 1));
      bitset_add_set (&(*f)->eval_docs, add_docs);
      bitset_free (&add_docs);
    }
}

/* -------------------------------------------------------------------------- */
Fold
fold_new_read (gchar * path, guint i, InvertedIndex folds)
{
  Fold fold = __fold_new (path, i, folds);
  fold->class_info = class_info_new_read (fold->filepath_class_info);
  fold->term_info = term_info_new_read (fold->filepath_term_info);
  return fold;
}

extern gboolean verbose;

/* -------------------------------------------------------------------------- */
void
__fold_create_assigned_samples_file (gchar *samples_path, gchar *file_path, BitSet posdocs, BitSet negdocs, BitSet *docs)
{
  FILE *fout = fopen (file_path, "w");
  FILE *fin = fopen (samples_path, "r");
  gchar buffer[STR_BUFFER_LEN];

  if (!fout) {
    printf("%s\n", strerror(errno));
    tc_error("__fold_create_assigned_samples: unable to create temporal file");
  }
  if (!fin) {
    printf("%s\n", strerror(errno));
    tc_error("__fold_create_assigned_samples: unable to open samples file");
  }
  bitset_it_reset(docs);
  while (!bitset_it_end(*docs) && !feof(fin)) {
    /* class label */
    tc_id di = bitset_it_next(docs);		
    fgets(buffer, STR_BUFFER_LEN, fin);
    if (bitset_in(posdocs, di))
      fprintf(fout, "+1 ");
    else if (bitset_in(negdocs, di))		
      fprintf(fout, "-1 ");
    else {         /* we jump sample */
      for (;fgetc(fin) != '\n' && !feof(fin););
      continue;
    }
    if (!feof(fin)) 
      fputs(buffer, fout);
    while (strlen(buffer) == (STR_BUFFER_LEN-1) && buffer[STR_BUFFER_LEN-2] != '\n') {			
      fgets(buffer, STR_BUFFER_LEN, fin);
      if (!feof(fin)) fputs(buffer, fout);
      if (feof(fin))
	buffer[1] = '\0';
    }
  }
  fclose(fin);
  fclose(fout);
}
/* -------------------------------------------------------------------------- */
void
__fold_train_and_evaluate_method(Method *m, Fold f, tc_id class_id, BitSet poseval_docs, BitSet negeval_docs, 
				 Samples train_samples, Samples eval_samples, FoldTrainParams *params)
{
  Method minf = *m;
  BitSet assignment;
  void *c;
  static guint mcount=0;
  BitSet eval_docs;
  Sample s;
  char *model_file;

  /* sets model file name to absolute path */	
  if (minf->model_file)
    g_free(minf->model_file);
  minf->model_file = g_strdup_printf ("%s%s%s%d", f->path_models, PATH_CONNECTOR, params->modelstr, mcount);

  /* train classifier and evaluate it */
  if (minf->classifier == CLASSIFIER_COMMAND) {
    eval_docs = bitset_new_union(poseval_docs, negeval_docs);
    method_exec_cmd(minf, f->filepath_train_samples_tmp, "");                     /* learn */
    method_exec_cmd(minf, f->filepath_eval_samples_tmp, f->filepath_predictions); /* classify */
    s = method_new_sample_assignment_from_external(&minf, poseval_docs, eval_docs, f->filepath_predictions);  /* evaluate */
    if (params->compute_scut)
      method_compute_scut(&minf, s, params->selection_measure);
    method_evaluate_sample_based(&minf, s);
    sample_free(&s);
    bitset_free(&eval_docs);
  }
  else {
    c = method_learn_builtin(minf, train_samples);                      /* learn */
    s = method_new_sample_assignment_from_builtin(minf, eval_samples, c); /* classify */ 
    if (params->compute_scut)
      method_compute_scut(&minf, s, params->selection_measure);
    method_evaluate_sample_based(&minf, s);                            /* evaluate */	
    sample_free(&s);
    classifier_free[minf->classifier](&c);
  }

  /* resets model filename to relative path */
  g_free(minf->model_file);
  minf->model_file = g_strdup_printf ("%s%d", params->modelstr, mcount++);
}
/* -------------------------------------------------------------------------- */
Method fold_train_class(Fold *f, tc_id class_id, BitSet postrain_docs, BitSet negtrain_docs, 
			BitSet poseval_docs, BitSet negeval_docs, 
			FoldTrainParams* params, Samples train_samples, Samples eval_samples) {
  
  Samples train_class_samples, eval_class_samples;
  Method minf, method;
  tc_id j;
  gdouble val;
  char *model_file;

  /* ignore filtered class */
  ClassInfo *class = class_info_get ((*f)->class_info, class_id);
  if (class->new_id == -1) {
    tc_message("(-) !\n");
    return NULL;
  }      

  /* we create assigned samples in disk */
  if (methods_any_command(params->methods)) {
	if (!strcmp(params->modelstr, "model-filter"))
	  tc_error("fold_train_class: command based classifiers are not allowed for training class filters");
    __fold_create_assigned_samples_file ((*f)->filepath_train_samples, (*f)->filepath_train_samples_tmp, 
					 postrain_docs, negtrain_docs, &(*f)->train_docs);
    __fold_create_assigned_samples_file ((*f)->filepath_eval_samples, (*f)->filepath_eval_samples_tmp, 
					 poseval_docs, negeval_docs, &(*f)->eval_docs);
  }

  /* we create assigned samples in memory */
  if (methods_any_builtin(params->methods)) {
    train_class_samples = samples_shallow_clone(train_samples);
    eval_class_samples = samples_shallow_clone(eval_samples);    
    samples_assign (&train_class_samples, postrain_docs, negtrain_docs);
    samples_assign (&eval_class_samples, poseval_docs, negeval_docs);
  }

  /* 
     For each method we train the class and evaluate the method 
  */
  for (j = 0; j < params->methods->len; j++) {
    minf = methods_get (params->methods, j);
    minf->beta = params->beta;
    minf->pos_factor = (gdouble) negeval_docs->total / (gdouble) poseval_docs->total;
    __fold_train_and_evaluate_method(&minf, *f, class_id, poseval_docs, negeval_docs, train_class_samples, eval_class_samples, params);
  }

  /* select best method */
  minf = methods_get_best (params->methods, params->selection_measure);
  minf->class_id = class_id;

  /* remove other model files */
  for (j = 0; j < params->methods->len; j++) {
    method = methods_get (params->methods, j);
    model_file = g_strdup_printf ("%s%s%s", (*f)->path_models, PATH_CONNECTOR, method->model_file);
    if (method != minf && file_exists(model_file))
      remove (model_file);
    g_free(model_file);
  }

  /* remove assigned samples and temporal files */
  if (methods_any_command(params->methods)) {
    remove ((*f)->filepath_train_samples_tmp);
    remove ((*f)->filepath_eval_samples_tmp);
    remove ((*f)->filepath_predictions);
  }

  if (methods_any_builtin(params->methods)) {
    samples_shallow_free(&train_class_samples);
    samples_shallow_free(&eval_class_samples);
  }

  /* if method is not good we also delete its file */
  if (!method_valid (minf, params->selection_measure, params->selection_threshold, &val)) {
    model_file = g_strdup_printf ("%s%s%s", (*f)->path_models, PATH_CONNECTOR, method->model_file);
    remove(model_file);
    g_free(model_file);
    return NULL;
  }

  return minf;    
}
/* -------------------------------------------------------------------------- */
void
__fold_train_by_class (Fold * f, InvertedIndex class_doc, Samples train_samples, Samples eval_samples, FoldTrainParams *params)
{
  BitSet postrain_docs, poseval_docs, negtrain_docs, negeval_docs, class_docs;
  tc_id i;
  Method method;
  ClassInfo *class;
  gchar *methods_path;

  /* create array to store method selected for each class */
  (*f)->methods = methods_new();

  /* for each class */
  for (i = 0; i < (*f)->class_info->len; i++) {

    /* ignore filtered class */
    class = class_info_get ((*f)->class_info, i);
    if (class->new_id == -1) 
      continue;
    
    class_docs = inverted_index_get (class_doc, i);
    postrain_docs = bitset_new_intersection((*f)->train_docs, class_docs);
    negtrain_docs = bitset_new_diff((*f)->train_docs, class_docs);
    poseval_docs =  bitset_new_intersection((*f)->eval_docs, class_docs);
    negeval_docs = bitset_new_diff((*f)->eval_docs, class_docs);
    bitset_free (&class_docs);

    if (!strcmp(params->modelstr, "model-filter")) {
      /* we have to take out the class as feature from samples */
      samples_remove_feature(&train_samples, i, postrain_docs);
      samples_remove_feature(&eval_samples, i, poseval_docs);
    }

    method = fold_train_class(f, i, postrain_docs, negtrain_docs, poseval_docs, negeval_docs, params, train_samples, eval_samples);

    if (!strcmp(params->modelstr, "model-filter")) {
      /* now, we add it again */
      samples_add_feature(&train_samples, i, postrain_docs);
      samples_add_feature(&eval_samples, i, poseval_docs);
    }

    if (method) {
      methods_add (&(*f)->methods, method_new_clone (method));
      if (verbose) 
        printf(" [%.2f/%.2f]", method->precision*100.0, method->recall*100.0);  
    }
    else
      tc_message (" !"); /* class discarded, cannot find a valid method */
    tc_message("\n");
    
    bitset_free (&postrain_docs);
    bitset_free (&negtrain_docs);
    bitset_free (&poseval_docs);
    bitset_free (&negeval_docs);
  }

  /* we store array of methods */
  if (!strcmp(params->modelstr, "model-filter")) {
	g_free((*f)->filepath_methods_eval);
	(*f)->filepath_methods_eval = g_strconcat ((*f)->path, PATH_CONNECTOR, FOLD_FILE_METHODS_FILTER, NULL);
  } 
  methods_serialize ((*f)->methods, (*f)->filepath_methods_eval);

  tc_message_int ("\n\tdone (%d classes successfully trained) \n",  (*f)->methods->len);
} 
/* -------------------------------------------------------------------------- */
void
__fold_train_mltree (Fold * f, InvertedIndex class_doc, Samples train_samples, Samples eval_samples, FoldTrainParams *params) {
  MLNode *node;
  guint i;
  BitSet classes, *mclass_docs;
  gchar *tree_file;
  ClassInfo *class;

  classes = bitset_new((*f)->class_info->len-1);  
  mclass_docs = (BitSet *)g_new0(BitSet, (*f)->class_info->len);

  for (i=0; i<(*f)->class_info->len; i++) {
    class = class_info_get((*f)->class_info, i);
    if (class->new_id != -1) {
      bitset_insert(&classes, i);
      mclass_docs[i] = inverted_index_get(class_doc, i);
    }
  }

  tc_message("Creating tree\n");
  node = mltree_new(f, 
		    classes, 
		    (*f)->train_docs, (*f)->eval_docs, 
		    train_samples, eval_samples,
		    mclass_docs, 
		    params, 0);
  tc_message("\n");
  bitset_free(&classes);

  for (i=0; i<(*f)->class_info->len; i++)
    if (mclass_docs[i])
	  bitset_free(&mclass_docs[i]);
  g_free(mclass_docs);
  
  tc_message("Saving to disk\n");
  tree_file = g_strconcat((*f)->path, PATH_CONNECTOR, FOLD_FILE_TREE, NULL);
  mltree_save(node, tree_file);
  mltree_dump(node);
  g_free(tree_file);

  mltree_free(node);
}
/* -------------------------------------------------------------------------- */
void
__fold_train_all_together (Fold * f, InvertedIndex class_doc, FoldTrainParams *params) {

  guint i, j, first;
  BitSet docs;
  gchar *command, *filestem, *filestem_names, *filestem_data, *filestem_test, c;
  ClassInfo *class;
  Sample s;
  FILE *fnames, *fdata, *fsamples;
  Method m;

  tc_message("Creating files");
  filestem = g_strconcat((*f)->path, PATH_CONNECTOR, "all", NULL);
  filestem_names = g_strconcat(filestem, ".names", NULL);
  filestem_data = g_strconcat(filestem, ".data", NULL);
  filestem_test = g_strconcat(filestem, "-test", NULL);

  // names file
  printf("\n");
  fnames = fopen(filestem_names, "w+");
  first = 1;
  for (i=0; i<(*f)->class_info->len; i++) {
    class = class_info_get((*f)->class_info, i);
    if (class->new_id != -1) {
      if (!first)
	fprintf(fnames, ", ");
      fprintf(fnames, "%d", class->new_id);
      first = 0;
    }
  }
  fprintf(fnames, ".\ncontent: scoredtext.\n");
  fclose(fnames);

  // data file (train samples)
  fdata = fopen(filestem_data, "w+");
  bitset_it_reset(&(*f)->train_docs);
  fsamples = fopen((*f)->filepath_train_samples, "r");
  while (!bitset_it_end((*f)->train_docs)) {
    i = bitset_it_next(&(*f)->train_docs);
    // sample
    for (c = fgetc(fsamples); c!='\n' && c != EOF;c = fgetc(fsamples)) {
      if (c == ':')
	fprintf(fdata, " ");
      else 
	fputc(c, fdata);
    }
    fprintf(fdata, ",");
    // classes
    for (j=0; j<(*f)->class_info->len; j++) {
      class = class_info_get((*f)->class_info, j);
      if (class->new_id != -1) {
	docs = inverted_index_get(class_doc, class->new_id);
	if (bitset_in(docs, i))
	  fprintf(fdata, " %d", class->new_id);
	bitset_free(&docs);      
      }
    }
    fprintf(fdata, ".\n");
  }
  fclose(fsamples);

  // data file (evaluation samples)
  bitset_it_reset(&(*f)->eval_docs);
  fsamples = fopen((*f)->filepath_eval_samples, "r");
  while (!bitset_it_end((*f)->eval_docs)) {
    i = bitset_it_next(&(*f)->eval_docs);
    // sample
    for (c = fgetc(fsamples); c!='\n' && c != EOF; c = fgetc(fsamples)) {
      if (c == ':')
	fprintf(fdata, " ");
      else 
	fputc(c, fdata);
    }
    fprintf(fdata, ",");
    // classes
    for (j=0; j<(*f)->class_info->len; j++) {
      class = class_info_get((*f)->class_info, j);
      if (class->new_id != -1) {
	docs = inverted_index_get(class_doc, class->new_id);
	if (bitset_in(docs, i))
	  fprintf(fdata, " %d", class->new_id);
	bitset_free(&docs);      
      }
    }
    fprintf(fdata, ".\n");
  }
  fprintf(fdata, "\n");
  fclose(fsamples);
  fclose(fdata);

  // data file (test samples)
  fdata = fopen(filestem_test, "w+");
  bitset_it_reset(&(*f)->test_docs);
  fsamples = fopen((*f)->filepath_test_samples, "r");
  while (!bitset_it_end((*f)->test_docs)) {
    i = bitset_it_next(&(*f)->test_docs);
    // sample
    for (c = fgetc(fsamples); c!='\n' && c != EOF; c = fgetc(fsamples)) {
      if (c == ':')
	fprintf(fdata, " ");
      else 
	fputc(c, fdata);
    }
    fprintf(fdata, ",");
    // classes
    for (j=0; j<(*f)->class_info->len; j++) {
      class = class_info_get((*f)->class_info, j);
      if (class->new_id != -1) {
	docs = inverted_index_get(class_doc, class->new_id);
	if (bitset_in(docs, i))
	  fprintf(fdata, " %d", class->new_id);
	bitset_free(&docs);      
      }
    }
    fprintf(fdata, ".\n");
  }
  fprintf(fdata, "\n");
  fclose(fsamples);
  fclose(fdata);

  tc_message("Running command\n");

  m = methods_get(params->methods, 0);
  command = g_strconcat(m->learn_command, " ", filestem, " > /dev/null 2> /dev/null", NULL);

  if (system(command))
    tc_error("Unable to run command");

  g_free(filestem);
  g_free(filestem_names);
  g_free(filestem_data);
  g_free(command);
}
/* -------------------------------------------------------------------------- */
void
fold_train (Fold * f, Model m, FoldTrainParams *params)  {

  Samples train_samples, eval_samples;

  /* we create models directory */
  if (!g_file_test ((*f)->path_models, G_FILE_TEST_IS_DIR))
    mkdir ((*f)->path_models, 0750);

  /* we create samples in memory */
  if (methods_any_builtin(params->methods)) {
    train_samples = samples_new_read((*f)->filepath_train_samples, (*f)->train_docs);
    eval_samples = samples_new_read((*f)->filepath_eval_samples, (*f)->eval_docs);
  }

  params->modelstr = g_strdup("model");
  params->methods = params->stage1methods;
  if (params->approach & FOLD_TRAIN_BY_CLASS) {
    tc_message("Training based on documents\n");
    __fold_train_by_class(f, m->class_doc, train_samples, eval_samples, params);
  }
  else if (params->approach & FOLD_TRAIN_MLTREE)
    __fold_train_mltree(f, m->class_doc, train_samples, eval_samples, params);
  else if (params->approach & FOLD_TRAIN_ALL_TOGETHER)
    __fold_train_all_together(f, m->class_doc, params);
  g_free(params->modelstr);

  /* remove assigned samples in memory */
  if (methods_any_builtin(params->methods)) {
    samples_free(&train_samples);
    samples_free(&eval_samples);
  }

  /* now, we train classes filter if selected */
  if (params->approach & FOLD_TRAIN_CLASS_FILTER) {	
    tc_message("Training based on assignments\n");
    /* samples creation */
    train_samples = samples_new_for_class_filter((*f)->class_info->len, m->class_doc, (*f)->train_docs);
    eval_samples = samples_new_for_class_filter((*f)->class_info->len, m->class_doc, (*f)->eval_docs);
    
    /* training */
    params->modelstr = g_strdup("model-filter");
    params->methods = params->stage2methods;
    __fold_train_by_class(f, m->class_doc, train_samples, eval_samples, params);
    g_free(params->modelstr);
  }	

}
/* -------------------------------------------------------------------------- */
/* This function takes the vector of classes and sets the pointer of the method
   to the proper one */
void
__fold_point_classes_to_methods(Methods methods, Methods filters, GArray **class_info)
{
  guint i;
  Method minf, filter;

  /* if the class couldn't be trained the method will be a pointer to NULL */  
  for (i=0; i<(*class_info)->len; i++) {	
    g_array_index(*class_info, ClassInfo, i).method = 
      g_array_index(*class_info, ClassInfo, i).filter = NULL;	  
  }
  for (i=0; i<methods->len; i++) {
    minf = methods_get(methods, i);
    if (minf)
      g_array_index(*class_info, ClassInfo, minf->class_id).method = minf;
    if (filters) {
      filter = methods_get(filters, i);
      if (filter)
	g_array_index(*class_info, ClassInfo, filter->class_id).filter = filter;
    }
  }

  /* we disable those classes that couldn't be trained 
	 for (i=0; i<(*class_info)->len; i++) {
	 ClassInfo *class = &g_array_index(*class_info, ClassInfo, i);
	 if (class->method == NULL)
	 class->new_id = -1;		
	 } 
  */
}
/* -------------------------------------------------------------------------- */
BitSet *__fold_compute_docs_assignment(InvertedIndex class_doc, BitSet docs, ClassInfoArray class_info) {
  BitSet *docs_assign, cdocs;
  tc_id i, d;
  ClassInfo *class;

  /* Allocate memory */
  docs_assign = (BitSet *)g_new0(BitSet, docs->max_id+1);
  for (i=0; i<= docs->max_id; i++)
    docs_assign[i] = bitset_new(class_info->len);

  /* Assign classes (only those valid) */
  for (i=0; i< class_info->len; i++) {
	class = class_info_get(class_info, i);
	if (class->new_id == -1)
	  continue;
    cdocs = inverted_index_get(class_doc, i);
    bitset_it_reset(&cdocs);
    while (!bitset_it_end(cdocs)) {
      d = bitset_it_next(&cdocs);
      if (d <= docs->max_id)
		bitset_insert(&docs_assign[d], i);
    }
    bitset_free(&cdocs);
  }
  return docs_assign;
}
/* -------------------------------------------------------------------------- */
void
__fold_test_all_together(Fold * f, Model m, Methods *microavg, Methods *macroavg, FoldTrainParams *params) 
{
  tc_id i, count=1, total_test_classes=0;
  BitSet classes, assign;
  Methods docs_stats;
  Method method, fold_microavg, fold_macroavg;
  gchar *filestem, *filestem_test, *filestem_results, *command;
  double avgprec=0.0, avgrec=0.0;
  FILE *ftest, *fres;
  gint classtype, class;

  tc_message("Classifying documents\n");

  // prepare statistics
  docs_stats = methods_new();
  for (i=0; i<(*f)->test_docs->max_id+1; i++)
    methods_add(&docs_stats, method_new());
  for (i=0; i<(*f)->class_info->len; i++)
    if (class_info_get((*f)->class_info, i)->new_id != -1)
      total_test_classes++;

  // prepare test file
  filestem = g_strconcat((*f)->path, PATH_CONNECTOR, "all", NULL);
  filestem_test = g_strconcat(filestem, "-test", NULL);
  filestem_results = g_strconcat(filestem, "-res", NULL);

  // perform classification
  tc_message("Running command\n");
  method = methods_get(params->methods, 0);
  command = g_strconcat(method->classify_command, " ", filestem, " ", filestem_test, " ", filestem_results, 
			" ", NULL);
  if (system(command))
    tc_error("Unable to run command");
  g_free(command);

  // read results and compute measures
  fres = fopen(filestem_results, "r");
  classes = bitset_new((*f)->class_info->len);
  i = 0;
  classtype = 0;
  while (fscanf(fres, "%d", &class) != EOF) {
    if (class == -1) { // correct classes read
      classtype = 1;
      assign = bitset_new((*f)->class_info->len);      
    } else if (class == -10) { // system assigned classes read
      classtype = 0;   
      method = methods_get(docs_stats, i);
      method->class_id = i; /* we use class_id to store document id */
      method_evaluate_assignment(&method, assign, classes, total_test_classes);
      if (verbose) {
	avgprec += method->precision;
	avgrec += method->recall;
	printf("(%.4f/%.4f) (avg: %.4f/%.4f)\n", method->precision, method->recall,
	       avgprec / (double)count, avgrec / (double)count);
      }
      bitset_free(&classes);
      classes = bitset_new((*f)->class_info->len);
      i++;
      count++;
    } else if (classtype == 0) 
      bitset_insert(&classes, class);
    else
      bitset_insert(&assign, class);
  } 
  if (assign)
    bitset_free(&assign);
  if (classes)
    bitset_free(&classes);
  fclose(fres);

  fold_microavg = methods_micro_avg(docs_stats);
  fold_macroavg = methods_macro_avg(docs_stats);
  methods_free(&docs_stats);
	
  if (verbose) {
    printf("         TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
    printf("avg) ");
    method_dump_measures(fold_microavg);
    printf("AVG) ");
    method_dump_measures(fold_macroavg);
  }

  fold_microavg->class_id = fold_macroavg->class_id = 0; /* to avoid be ignored in averaging */
  methods_add(microavg, fold_microavg);
  methods_add(macroavg, fold_macroavg);	
  g_free(filestem);
  g_free(filestem_test);
  g_free(filestem_results);
}
/* -------------------------------------------------------------------------- */
void
__fold_test_tree(Fold * f, Model m, Methods *microavg, Methods *macroavg, FoldTrainParams *params) 
{
  MLNode *tree;
  gchar *tree_file;
  Samples samples;
  Sample s;
  tc_id i, count=0, total_test_classes=0;
  BitSet classes, *docs_assign;
  Methods docs_stats;
  Method method, fold_microavg, fold_macroavg;
  double avgprec=0.0, avgrec=0.0;

  tc_message("Reading tree\n");
  tree_file = g_strconcat((*f)->path, PATH_CONNECTOR, FOLD_FILE_TREE, NULL);  
  tree = mltree_new_read(tree_file);  
  g_free(tree_file);
  mltree_dump(tree);

  tc_message("Classifying documents\n");
  docs_stats = methods_new();
  for (i=0; i<(*f)->test_docs->max_id+1; i++)
    methods_add(&docs_stats, method_new());
  for (i=0; i<(*f)->class_info->len; i++)
    if (class_info_get((*f)->class_info, i)->new_id != -1)
      total_test_classes++;
  docs_assign = __fold_compute_docs_assignment(m->class_doc, (*f)->test_docs, (*f)->class_info);
  samples = samples_new_read((*f)->filepath_test_samples, (*f)->test_docs);
  samples_it_reset(&samples);
  while (!samples_it_end(samples)) {
    s = samples_it_next(&samples);
    if (verbose) {
      count++;
      printf("[%d/%d]", count, samples->total);
    }
    classes = mltree_classify(tree, s, *f, 0, params);
    if (verbose) {
      printf("\n");
      bitset_dump(classes);
      bitset_dump(docs_assign[s->id]);
    }
    method = methods_get(docs_stats, s->id);
    method->class_id = s->id; /* we use class_id to store document id */
    method_evaluate_assignment(&method, classes, docs_assign[s->id], total_test_classes);
    if (verbose) {
      avgprec += method->precision;
      avgrec += method->recall;
      printf("(%.4f/%.4f) (avg: %.4f/%.4f)\n", method->precision, method->recall,
	     avgprec / (double)count, avgrec / (double)count);
    }
    bitset_free(&classes);
  }

  for (i=0; i<= (*f)->test_docs->max_id; i++)
    bitset_free(&docs_assign[i]);
  g_free(docs_assign);
  samples_free(&samples);
  mltree_free(tree);

  fold_microavg = methods_micro_avg(docs_stats);
  fold_macroavg = methods_macro_avg(docs_stats);
  methods_free(&docs_stats);
	
  if (verbose) {
    printf("         TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
    printf("avg) ");
    method_dump_measures(fold_microavg);
    printf("AVG) ");
    method_dump_measures(fold_macroavg);
  }

  fold_microavg->class_id = fold_macroavg->class_id = 0; /* to avoid be ignored in averaging */
  methods_add(microavg, fold_microavg);
  methods_add(macroavg, fold_macroavg);	
}

/* -------------------------------------------------------------------------- */
/* Classifies a sample using ranking */
ClassRank __fold_rank(Fold f, Sample s, gchar flags) {

  ClassInfo *class;
  ClassRank cr = classrank_new(f->class_info->len);
  Method method;
  FILE *fout, *file;
  gdouble val;
  tc_id i;  
  void *c;
  char *model_file;
  
  /* we create sample file */
  fout = fopen (f->filepath_sample_tmp, "w");
  if (!fout)
	tc_error("mltree_classify: unable to create temporal file");
  fprintf(fout, "+0 "); /* no assignment */
  sample_serialize(s, fout, FALSE);
  fclose(fout);	

  /* we classify against each class */
  for (i = 0; i < f->class_info->len; i++) {

	/* select method (for classifying or for filtering */
	class = class_info_get (f->class_info, i);
	if (flags & DO_SAMPLE_CLASSIFICATION)
	  method = class->method;
	else if (flags & DO_CLASS_FILTERING)
	  method = class->filter;
	else
	  tc_error("__fold_rank: wrong flags");

	/* ignore filtered class */
	if (class->new_id == -1 || !method)
	  continue;

	val = 0.0;

	/* model local path to absolute path */
	model_file = g_strdup_printf ("%s%s%s", f->path_models, PATH_CONNECTOR, method->model_file);
	g_free(method->model_file);
	method->model_file = model_file;

	if (method->classifier == CLASSIFIER_COMMAND) {
	  method_exec_cmd(method, f->filepath_sample_tmp, f->filepath_predictions);
	  file = fopen(f->filepath_predictions, "r");
	  fscanf(file, "%lf", &val);
	  fclose(file);
	  remove(f->filepath_predictions);
	} else  {
	  c = classifier_new_read[method->classifier](method->model_file);
	  val = classifier_classify[method->classifier](c, s);
	  classifier_free[method->classifier](&c);
	}	
	//printf("%lf %lf\n", val, method->cut);
	if (val > method->cut || flags & ALSO_NEGATIVES)
	  classrank_add(&cr, i, val); 	 
  }
  remove (f->filepath_sample_tmp);

  /* now, we apply filter */
  //printf("-----\n");
  return cr;
}
/* -------------------------------------------------------------------------- */
/* With this testing we classify document by document, selecting the n
   top ranked classes */

void
__fold_test_ranked(Fold * f, Model m, Methods *microavg, Methods *macroavg, FoldTrainParams *params) 
{
  Samples samples;
  Sample s, rank;
  gchar rank_flags;
  tc_id i, count=0, total_test_classes=0;
  BitSet classes, *docs_assign;
  Methods docs_stats, methods;
  Method method, fold_microavg, fold_macroavg;
  gdouble avgprec=0.0, avgrec=0.0;
  ClassRank cr;

  tc_message("Classifying documents\n");
  docs_stats = methods_new();
  for (i=0; i<(*f)->test_docs->max_id+1; i++)
    methods_add(&docs_stats, method_new());

  for (i=0; i<(*f)->class_info->len; i++)
	if (class_info_get((*f)->class_info, i)->new_id != -1)
	  total_test_classes++;

  docs_assign = __fold_compute_docs_assignment(m->class_doc, (*f)->test_docs, (*f)->class_info);
  methods = methods_new_deserialize((*f)->filepath_methods_eval);
  __fold_point_classes_to_methods(methods, NULL, &(*f)->class_info);

  samples = samples_new_read((*f)->filepath_test_samples, (*f)->test_docs);
  samples_it_reset(&samples);

  while (!samples_it_end(samples)) {
    s = samples_it_next(&samples);
    if (verbose) {
      count++;
      printf("[%d/%d]", count, samples->total);
    }

    /* get ranked list of classes */
    if (params->positive_ranked)
      rank_flags = 0 | ONLY_POSITIVES;
    else
      rank_flags = 0 | ALSO_NEGATIVES;
    cr = __fold_rank(*f, s, rank_flags | DO_SAMPLE_CLASSIFICATION);
    rank = classrank_get_top(cr, params->ntop);
    classrank_free(&cr);
    classes = sample_to_bitset(rank);
    
    /* update statistics */
    method = methods_get(docs_stats, s->id);
    method->beta = 1.0;
    method->class_id = s->id; /* we use class_id to store document id */
    method_evaluate_assignment(&method, classes, docs_assign[s->id], total_test_classes);

    bitset_free(&classes);
    sample_free(&rank);

    if (verbose) {
      avgprec += method->precision;
      avgrec += method->recall;
      printf("(%.4f/%.4f) (avg: %.4f/%.4f)\n", method->precision, method->recall,
	     avgprec / (double)count, avgrec / (double)count);
    }
  }

  for (i=0; i<= (*f)->test_docs->max_id; i++)
    bitset_free(&docs_assign[i]);
  g_free(docs_assign);
  samples_free(&samples);

  fold_microavg = methods_micro_avg(docs_stats);
  fold_macroavg = methods_macro_avg(docs_stats);
  methods_free(&docs_stats);
	
  if (verbose) {
    printf("         TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
    printf("avg) ");
    method_dump_measures(fold_microavg);
    printf("AVG) ");
    method_dump_measures(fold_macroavg);
  }

  fold_microavg->class_id = fold_macroavg->class_id = 0; /* to avoid be ignored in averaging */
  methods_add(microavg, fold_microavg);
  methods_add(macroavg, fold_macroavg);	
}
/* -------------------------------------------------------------------------- */
/* With this testing we classify document by document, and then, for
   each document, classes assigned are passed to the second-stage
   classifier class by class */
void
__fold_test_filtering(Fold * f, Model m, Methods *microavg, Methods *macroavg, FoldTrainParams *params) 
{
  Samples samples;
  Sample s, rank1, rank2;
  tc_id i, count=0, total_test_classes=0;
  BitSet classes1, classes2, classes, *docs_assign;
  gchar rank_flags;
  Methods docs_stats, methods, filters;
  Method method, fold_microavg, fold_macroavg;
  gdouble avgprec=0.0, avgrec=0.0;
  ClassRank cr;

  tc_message("Classifying documents\n");
  docs_stats = methods_new();
  for (i=0; i<(*f)->test_docs->max_id+1; i++)
    methods_add(&docs_stats, method_new());

  for (i=0; i<(*f)->class_info->len; i++)
	if (class_info_get((*f)->class_info, i)->new_id != -1)
	  total_test_classes++;

  docs_assign = __fold_compute_docs_assignment(m->class_doc, (*f)->test_docs, (*f)->class_info);
  methods = methods_new_deserialize((*f)->filepath_methods_eval);
  filters = methods_new_deserialize((*f)->filepath_methods_filter);
  __fold_point_classes_to_methods(methods, filters, &(*f)->class_info);

  samples = samples_new_read((*f)->filepath_test_samples, (*f)->test_docs);
  samples_it_reset(&samples);

  while (!samples_it_end(samples)) {
    s = samples_it_next(&samples);
    if (verbose) {
      count++;
      printf("[%d/%d]", count, samples->total);
    }
    /* get ranked list of classes */
    if (params->stage1_positive)
      rank_flags = 0 | ONLY_POSITIVES;
    else
      rank_flags = 0 | ALSO_NEGATIVES;
    cr = __fold_rank(*f, s, rank_flags | DO_SAMPLE_CLASSIFICATION);
    rank1 = classrank_get_top(cr, params->stage1_ntop);
    classrank_free(&cr);
    
    /* filter them */
    if (params->stage2_positive)
      rank_flags = 0 | ONLY_POSITIVES;
    else
      rank_flags = 0 | ALSO_NEGATIVES;
    cr = __fold_rank(*f, rank1, rank_flags | DO_CLASS_FILTERING);
    if (params->stage2_ntop > 0)
      rank2 = classrank_get_top(cr, params->stage2_ntop);
    else if (params->stage2_nbottom > 0)
      rank2 = classrank_get_bottom(cr, params->stage2_nbottom);
    else
      rank2 = classrank_get_top(cr, 0);
    classrank_free(&cr);
    
    /* evaluate */
    classes1 = sample_to_bitset(rank1);
    classes2 = sample_to_bitset(rank2);
    //sample_dump(rank2);
    sample_free(&rank1);
    sample_free(&rank2);
    
    if (params->stage2_merge == FOLD_STAGE2_UNION)
      classes = bitset_new_union(classes1, classes2);
    else if (params->stage2_merge == FOLD_STAGE2_INTERSECTION)
      classes = bitset_new_intersection(classes1, classes2);
    else if (params->stage2_merge == FOLD_STAGE2_DIFFERENCE)
      classes = bitset_new_diff(classes1, classes2);
    else /* anything else is considered to prefer just second set of classes */
      classes = bitset_new_copy(classes2);
    
    bitset_dump(classes1);
    bitset_dump(classes2);
    bitset_dump(classes);
    method = methods_get(docs_stats, s->id);
    method->class_id = s->id; /* we use class_id to store document id */
    method->beta = 1.0;
    method_evaluate_assignment(&method, classes, docs_assign[s->id], total_test_classes);
    if (verbose) {
      avgprec += method->precision;
      avgrec += method->recall;
      printf("(%.4f/%.4f) (avg: %.4f/%.4f)\n", method->precision, method->recall,
	     avgprec / (double)count, avgrec / (double)count);
    }
    bitset_free(&classes);
    bitset_free(&classes1);
    bitset_free(&classes2);
  }

  for (i=0; i<= (*f)->test_docs->max_id; i++)
    bitset_free(&docs_assign[i]);
  g_free(docs_assign);
  samples_free(&samples);

  fold_microavg = methods_micro_avg(docs_stats);
  fold_macroavg = methods_macro_avg(docs_stats);
  methods_free(&docs_stats);
  methods_free(&methods);
  methods_free(&filters);
	
  if (verbose) {
    printf("         TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
    printf("avg) ");
    method_dump_measures(fold_microavg);
    printf("AVG) ");
    method_dump_measures(fold_macroavg);
  }

  fold_microavg->class_id = fold_macroavg->class_id = 0; /* to avoid be ignored in averaging */
  methods_add(microavg, fold_microavg);
  methods_add(macroavg, fold_macroavg);	
}
/* -------------------------------------------------------------------------- */
void
fold_test(Fold * f, Model m, Methods *cmicroavg, Methods *cmacroavg, Methods *dmicroavg, Methods *dmacroavg, guint by_what, FoldTrainParams *params)
{
  guint i, nctested=0, nctrained=0;
  Samples test_samples;
  Sample s;
  BitSet pos_docs, neg_docs, assignment;
  Method fold_cmicroavg, fold_cmacroavg, fold_dmicroavg, fold_dmacroavg, method;
  Methods methods, docs_stats;
  void *c;
  gboolean verbose_tmp;
  ClassInfo *class;
  IntStrTable classstr;
  gchar *model_file;

  /* for each class */
  if (verbose)
	printf("--------------------------------------------------------------------------------------------------------- (Fold %2d, %d docs)\n", (*f)->id+1, (*f)->test_docs->total);

  /* We have trained by tree and test using it */
  if (by_what == FOLD_TEST_BY_TREE) {
    __fold_test_tree(f, m, dmicroavg, dmacroavg, params);
    return;
  }
  
  if (by_what == FOLD_TEST_ALL_TOGETHER) {
    params->methods = params->stage1methods;
    __fold_test_all_together(f, m, dmicroavg, dmacroavg, params);
    return;
  }

  /* We want to test by ranked n top */
  if (by_what == FOLD_TEST_BY_RANK) {
    __fold_test_ranked(f, m, dmicroavg, dmacroavg, params);
    return;
  }

  /* We want to test by filtering classes */
  if (by_what == FOLD_TEST_FILTERING) {
    __fold_test_filtering(f, m, dmicroavg, dmacroavg, params);
    return;
  }

  methods = methods_new_deserialize((*f)->filepath_methods_eval);
  __fold_point_classes_to_methods(methods, NULL, &(*f)->class_info);
  methods_reset_measures(&methods);

  /* Here we store a method for each doc to compute statistics about each doc */
  docs_stats = methods_new_sized((*f)->test_docs->max_id+1);    

  for (i = 0; i < (*f)->class_info->len; i++) {			      

    if (verbose) {
      printf ("class %4d/%d     \r", i, m->class_id->max_id);
      fflush(stdout);
    }

    /* ignore filtered class */
    class = class_info_get ((*f)->class_info, i);
    if (class->new_id == -1)
      continue;

    /* create training, evaluation and test data */
    pos_docs = inverted_index_get (m->class_doc, i);
    neg_docs = bitset_new_diff((*f)->test_docs, pos_docs);
    
    verbose_tmp = verbose;
    verbose = FALSE;

    if (class->method == NULL) { 
	  /* this class couldn't be trained so we just create a void
		 method and evaluate it anyhow */
      class->method = method_new();
      class->method->beta = 1.0;
      class->method->class_id = i;
      methods_add(&methods, class->method);
      assignment = bitset_new((*f)->test_docs->max_id);
      test_samples = samples_new_read((*f)->filepath_test_samples, (*f)->test_docs);                 
      samples_assign (&test_samples, pos_docs, neg_docs);
      method_evaluate_builtin(&class->method, test_samples, assignment);		
      methods_evaluate_builtin_in_docs(&docs_stats, test_samples, assignment); 
      bitset_free(&assignment);
      samples_free(&test_samples);
    }
    else if (class->method->classifier == CLASSIFIER_COMMAND) {

      /* model local path to absolute path */
      model_file = g_strdup_printf ("%s%s%s", (*f)->path_models, PATH_CONNECTOR, class->method->model_file);
      g_free(class->method->model_file);
      class->method->model_file = model_file;

      class->method->beta = 1.0;
      __fold_create_assigned_samples_file ((*f)->filepath_test_samples, (*f)->filepath_test_samples_tmp, pos_docs, neg_docs, &(*f)->test_docs);                    
      method_exec_cmd(class->method, (*f)->filepath_test_samples_tmp, (*f)->filepath_predictions);
      s = method_new_sample_assignment_from_external(&class->method, pos_docs, (*f)->test_docs, (*f)->filepath_predictions);
      method_evaluate_sample_based(&class->method, s);
      methods_evaluate_external_in_docs(&docs_stats, pos_docs, (*f)->test_docs, (*f)->filepath_predictions);
      sample_free(&s);
      remove ((*f)->filepath_test_samples_tmp);
      remove ((*f)->filepath_predictions);
      nctrained++;
    }
    else {

      /* model local path to absolute path */
      model_file = g_strdup_printf ("%s%s%s", (*f)->path_models, PATH_CONNECTOR, class->method->model_file);

      class->method->beta = 1.0;
      test_samples = samples_new_read((*f)->filepath_test_samples, (*f)->test_docs);                 
      samples_assign (&test_samples, pos_docs, neg_docs);
      c = classifier_new_read[class->method->classifier](model_file);                   
      assignment = method_classify_builtin(class->method, test_samples, c);                            
      method_evaluate_builtin(&class->method, test_samples, assignment);
      methods_evaluate_builtin_in_docs(&docs_stats, test_samples, assignment); 	  
      bitset_free(&assignment);
      classifier_free[class->method->classifier](&c);
      samples_free(&test_samples);
      nctrained++;
      
      g_free(model_file);
    }
    verbose = verbose_tmp;
    bitset_free (&pos_docs);
    bitset_free (&neg_docs);
    
    nctested++;
  }	/* end classes */
  
  /* we save results of test for each method */
  methods_serialize(methods, (*f)->filepath_methods_test);
  
  for (i=0; i<docs_stats->len; i++) {
    method = methods_get(docs_stats, i);
    if (method != NULL) {
      method->beta = 1.0;
      method->class_id = 0;
      method_compute_measures(&method);
    }
  }
  fold_dmicroavg = methods_micro_avg(docs_stats);
  fold_dmacroavg = methods_macro_avg(docs_stats);
  methods_free(&docs_stats);	  
  fold_cmicroavg = methods_micro_avg(methods);
  fold_cmacroavg = methods_macro_avg(methods);
  
  if (verbose) {
    printf("BY CLASS TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
    printf("avg) ");
    method_dump_measures(fold_cmicroavg);
    printf("AVG) ");
    method_dump_measures(fold_cmacroavg);
    printf("BY DOC   TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
    printf("avg) ");
    method_dump_measures(fold_dmicroavg);
    printf("AVG) ");
    method_dump_measures(fold_dmacroavg);
    printf("%d classes tested, %d classes where successfully trained\n", nctested, nctrained);

    fold_cmacroavg->nctrain = (double)nctrained;
    fold_cmacroavg->nctest = (double)nctested;

    /* ten most frequent classes */
    printf("10 Most frequent classes:\n");
    nctested=i=0;
    classstr = intstr_table_new(m->class_id);
    class_info_sort_by_df(&(*f)->class_info);
    while (i<(*f)->class_info->len && nctested<params->ntop_shown) {
      class = class_info_get ((*f)->class_info, i);
      if (class->new_id != -1) {
        nctested++;
        printf("--------------------- '%s' (in %d docs)\n     ", intstr_table_get(classstr, class->id), class->train_df+class->eval_df+class->test_df);
        method_dump_measures(class->method);
      }
      i++;
    }
    intstr_table_free(&classstr);
  }

  /* to avoid be ignored in averaging */
  fold_cmicroavg->class_id = fold_cmacroavg->class_id =  fold_dmicroavg->class_id = fold_dmacroavg->class_id = 0; 
  methods_add(cmicroavg, fold_cmicroavg);
  methods_add(cmacroavg, fold_cmacroavg);	
  methods_add(dmicroavg, fold_dmicroavg);
  methods_add(dmacroavg, fold_dmacroavg);	
  methods_free(&methods);
} 
/* -------------------------------------------------------------------------- */
gdouble __fold_assign_value (gchar *file_path)
{
  FILE *fhnd = fopen(file_path, "r");
  double value;

  if (!fhnd)
    tc_error("__fold_add_class: unable to read prediction");

  /* decission value */
  fscanf(fhnd, "%lf", &value);
  fclose(fhnd);
	
  return value;
}
/* -------------------------------------------------------------------------- */
/* 
   Automatically assings classes to a given sample generated from a full text
   processed by model_parse_doc()
*/
Sample
fold_classify (Fold f, Sample s, FeatureTransformations ftrans)
{
  guint i;
  Sample classes;
  Sample new_s;
  Methods methods;
  FILE *fout;
  ClassInfo *class;
  gchar *cmd, *model_file;
  void *c;
  double val;

  /* set of assigned classes (empty) */
  classes = sample_new();
	
  methods = methods_new_deserialize(f->filepath_methods_eval);
  __fold_point_classes_to_methods(methods, NULL, &f->class_info);
	
  /* we reweight document */
  new_s = sample_new_reweight(s, f->term_info, f->train_docs, ftrans);	
	
  if (methods_any_command(methods)) {
	/* we create sample file */
	fout = fopen (f->filepath_sample_tmp, "w");
	if (!fout)
	  tc_error("fold_classify: unable to create temporal file");
	fprintf(fout, "+0 "); /* no assignment */
	sample_serialize(new_s, fout, FALSE);
	fclose(fout);
  }

  tc_message("Searching classes\n");	
  /* for each class we assign to document */
  for (i = 0; i < f->class_info->len; i++)
    {				  
      tc_message(".");
      /* ignore filtered class */
      class = class_info_get (f->class_info, i);
      if (class->new_id == -1 || !class->method)
	continue;
   
      /* model local path to absolute path */
      model_file = g_strdup_printf ("%s%s%s", f->path_models, PATH_CONNECTOR, class->method->model_file);
      g_free(class->method->model_file);
      class->method->model_file = model_file;

      if (class->method->classifier == CLASSIFIER_COMMAND) {
	method_exec_cmd(class->method, f->filepath_sample_tmp, f->filepath_predictions);       
	val = __fold_assign_value (f->filepath_predictions);
	if (val > class->method->cut)
	  sample_add_feature(&classes, feature_new(i, val));
	remove (f->filepath_predictions);
      } else {
	c = classifier_new_read[class->method->classifier](class->method->model_file);
	val = classifier_classify[class->method->classifier](c, new_s);
	if (val > class->method->cut) {
	    /* printf("%d %lf\n", i, val);*/
	  sample_add_feature(&classes, feature_new(i, val));
	}
	classifier_free[class->method->classifier](&c);
      }
    }	/* end classes */
  tc_message("\n");
  sample_free(&new_s);
  if (methods_any_command(methods)) 
	remove (f->filepath_sample_tmp);
  methods_free(&methods);
  return  classes;
}
/* -------------------------------------------------------------------------- */
/* 
   This is the new method that reads only the HugeModel file and classifies accordingly
*/
Sample
fold_classify_huge (Fold f, Sample s, FeatureTransformations ftrans)
{
  guint i;
  Sample classes;
  Sample new_s;
  Methods methods;
  FILE *fout;
  ClassInfo *class;
  void *c;
  double val;
  gchar *modelName = "HugeModel";
  gchar *huge_path = " ";
  guint idx = 0;
  FILE *pFile;
  guint posit;
  fpos_t pos;
  char buf[BUFSIZ] = "Nothing";
  guint total, num;
  
  /* set of assigned classes (empty) */
  classes = sample_new();

  methods = methods_new_deserialize(f->filepath_methods_eval);
  __fold_point_classes_to_methods(methods, NULL, &f->class_info);

  /* we reweight document */
  new_s = sample_new_reweight(s, f->term_info, f->train_docs, ftrans);	

  /*  classification based on packled models is not available for external classifiers */
  if (methods_any_command(methods))
    tc_error("Huge models only available for PLAUM builtin classifier");
		
  huge_path = g_strconcat(f->path_models, PATH_CONNECTOR, modelName, NULL);
  pFile = fopen (huge_path,"r");

  tc_message("Searching classes\n");	
  /* for each class we assign to document */
  for (i = 0; i < f->class_info->len; i++)  {				  
    tc_message(".");
    /* ignore filtered class */
    class = class_info_get (f->class_info, i);
    if (class->new_id == -1 || !class->method)
      continue;
    
    if (class->method->classifier != CLASSIFIER_PLAUM)
      tc_error("Huge models only available for PLAUM classifier");

    c = classifier_new_read_at[class->method->classifier](pFile);
    val = classifier_classify[class->method->classifier](c, new_s);
    if (val > class->method->cut) {
	/* printf("%d %lf\n", i, val); */
      sample_add_feature(&classes, feature_new(i, val));
    }
    classifier_free[class->method->classifier](&c);
  }
  fclose (pFile);
  sample_free(&new_s);
  methods_free(&methods);
  return  classes;
}
/* -------------------------------------------------------------------------- */
/* 
   This is the new method that reads only the HugeModel file and classifies accordingly
*/
Sample
fold_classify_huge_alberto (Fold f, Sample s, FeatureTransformations ftrans)
{
  guint i;
  Sample classes;
  Sample new_s;
  Methods methods;
  FILE *fout;
  ClassInfo *class;
  void *c;
  double val;
  gchar *modelName = "HugeModel";
  gchar *huge_path = " ";
  guint idx = 0;
  FILE *pFile;
  guint posit;
  fpos_t pos;
  char buf[BUFSIZ] = "Nothing";
  guint total, num;
  
  huge_path = g_strconcat(f->path_models, PATH_CONNECTOR, modelName, NULL);

  /* set of assigned classes (empty) */
  classes = sample_new();

  methods = methods_new_deserialize(f->filepath_methods_eval);
  __fold_point_classes_to_methods(methods, NULL, &f->class_info);

  /* we reweight document */
  new_s = sample_new_reweight(s, f->term_info, f->train_docs, ftrans);	

  /*  classification based on packled models is not available for external classifiers */
  if (methods_any_command(methods))
    tc_error("fold_classify: unable to create temporal file");
		
  tc_message("Reading single model file...\n");

  //here I need to add a check to make sure that method classifier of model file is plaum

  pFile = fopen (huge_path,"r");
  //while (!feof(pFile)) //feof does not realy work fine.. using the other method below instead  
  total = 0;
  while (fscanf(pFile, "%d", &num) == 1)
  {
      fseek (pFile, 0, SEEK_CUR);
      posit=ftell(pFile);
      printf("Current position is %ld \n", posit);
      c = classifier_new_read_at[CLASSIFIER_PLAUM](pFile);
      fgetpos (pFile,&pos);
      fsetpos (pFile,&pos);
      val = classifier_classify[CLASSIFIER_PLAUM](c, new_s);
      // if (val > class->method->cut){       I DON'T KNOW HOW TO MAKE THIS CHECK.. LEADS TO SEGMENTATION FAULT
        printf("%d %lf\n",idx, val);
        sample_add_feature(&classes, feature_new(0, val));
        //}
      classifier_free[CLASSIFIER_PLAUM](&c);
      idx = idx + 1;
      total+=num;
  }
  fclose (pFile);
  sample_free(&new_s);
  return  classes;
}
/* -------------------------------------------------------------------------- */
/* Function to compute correlation matrix between classes */
void fold_dump_correlation_matrix(Fold f, Model m) {
  BitSet *mclass_docs, common;
  ClassInfo *ci, *cj;
  gint ni, nj, nij, N, i, j;
  IntStrTable cid_str;

  N = f->train_docs->total + f->eval_docs->total + f->test_docs->total;
  mclass_docs = (BitSet *)g_new0(BitSet, f->class_info->len);
  cid_str = intstr_table_new(m->class_id);

  for (i=0; i<f->class_info->len; i++) {
    ci = class_info_get(f->class_info, i);
    if (ci->new_id != -1)
      mclass_docs[i] = inverted_index_get(m->class_doc, i);
  }
  for (i=0; i<f->class_info->len; i++) {
    ci = class_info_get(f->class_info, i);
    if (ci->new_id != -1)
	  for (j=i; j<f->class_info->len; j++) {
		cj = class_info_get(f->class_info, j);
		if (cj->new_id != -1) {
		  ni = mclass_docs[i]->total;
		  nj = mclass_docs[j]->total;
		  common = bitset_new_intersection(mclass_docs[i], mclass_docs[j]);
		  nij = common->total;
		  if (nij)
			printf("%lf %d %d %d '%s' '%s'\n", 
				   ((double)(N*nij-ni*nj))/sqrt(((double)ni*(N-ni)*nj*(N-nj))),
				   ni, nj, nij,
				   intstr_table_get(cid_str, i),
				   intstr_table_get(cid_str, j));
		  else
			printf("0.0 %d %d %d '%s' '%s'\n", 
				   ni, nj, nij,
				   intstr_table_get(cid_str, i),
				   intstr_table_get(cid_str, j));

		}
	  }  
  }
  for (i=0; i<f->class_info->len; i++)
    if (ci->new_id != -1)   
	  bitset_free(&mclass_docs[i]);
  g_free(mclass_docs);

  intstr_table_free(&cid_str);
}
/* -------------------------------------------------------------------------- */
/* Function to export classifiers models */
void fold_export_models(Fold f, Model m, gchar *dir) {
  guint i, new_id;
  Methods methods;
  void *c;
  ClassInfo *class;
  IntStrTable class_str, term_str, term_new_str;
  gchar *path;
  GArray *ii; /* translation of new term id to term id */

  /* data directory creation when missing */
  if (!g_file_test (dir, G_FILE_TEST_IS_DIR))
    mkdir (dir, 0750);

  /* compute translation tables */
  class_str = intstr_table_new(m->class_id);
  term_str = intstr_table_new(m->term_id);
  
  /* we want to map new term ids to old ones */
  ii = g_array_new(FALSE, TRUE, sizeof(guint));
  g_array_set_size(ii, f->term_info->len+1);
  for (i=0; i<term_str->len; i++) {
    new_id = term_info_get(f->term_info, i)->new_id;
    if (new_id >= 0)
      g_array_index(ii, guint, new_id) = i;
  }

  /* get classes methods */
  methods = methods_new_deserialize(f->filepath_methods_eval);
  __fold_point_classes_to_methods(methods, NULL, &f->class_info);

  /* class by class, by export models */
  for (i = 0; i < f->class_info->len; i++) {	      

    path = g_strconcat(dir, PATH_CONNECTOR, intstr_table_get(class_str, i), NULL);

    /* ignore filtered class or invalid ones*/
    class = class_info_get (f->class_info, i);
    if (class->new_id == -1 || strlen(intstr_table_get(class_str, i)) == 0) {
      g_free(path);
      continue;
    }

    if (class->method == NULL) /* this class couldn't be trained */
      tc_message("Untrained class\n");
    else if (class->method->classifier == CLASSIFIER_COMMAND)
      tc_message("External method, cannot export classifier model\n");
    else {
      if (class->method->classifier != CLASSIFIER_PLAUM)
	tc_error("Exporting only implemented for PLAUM");
      c = classifier_new_read[class->method->classifier](class->method->model_file);      
      classifier_save_ascii[class->method->classifier](c, path, term_str, ii);
      classifier_free[class->method->classifier](&c);
    }
    g_free(path);
  } 	/* end classes */
  g_array_free(ii, FALSE);
  intstr_table_free(&class_str);
  intstr_table_free(&term_str);
  printf("done!\n");
}
/* -------------------------------------------------------------------------- */
void
fold_free (Fold * f)
{
  if ((*f)->class_info)
    class_info_free (&(*f)->class_info);
  if ((*f)->term_info)
    term_info_free (&(*f)->term_info);
  if ((*f)->methods)
    methods_free (&(*f)->methods);
  bitset_free (&(*f)->train_docs);
  bitset_free (&(*f)->eval_docs);
  bitset_free (&(*f)->test_docs);
  g_free ((*f)->path);
  g_free ((*f)->filepath_train_samples);
  g_free ((*f)->filepath_eval_samples);
  g_free ((*f)->filepath_test_samples);
  g_free ((*f)->filepath_train_samples_tmp);
  g_free ((*f)->filepath_eval_samples_tmp);
  g_free ((*f)->filepath_sample_tmp);
  g_free ((*f)->filepath_test_samples_tmp);
  g_free ((*f)->filepath_predictions);
  g_free ((*f)->filepath_class_info);
  g_free ((*f)->filepath_term_info);
  g_free ((*f)->filepath_methods_eval);
  g_free ((*f)->filepath_methods_test);
  if ((*f)->filepath_methods_filter)
	g_free ((*f)->filepath_methods_filter);
  g_free ((*f)->path_models);
  g_free (*f);
}
/* -------------------------------------------------------------------------- */
void
fold_pack_models (Fold f)
{
  void *c;
  ClassInfo *class;
  Methods methods;
  gchar *path;
  guint filtercount = 0;
  guint nullcount = 0;
  guint count = 0;
  guint i;
  gchar *modelName = "HugeModel", *model_file;

  path = g_strconcat(f->path_models, PATH_CONNECTOR, modelName, NULL);
  printf("Packing models into file: ");
  printf(path);
  printf("\n");   

  /* get classes methods */
  methods = methods_new_deserialize(f->filepath_methods_eval);
  __fold_point_classes_to_methods(methods, NULL, &f->class_info);

  for (i = 0; i < f->class_info->len; i++) {

      /* ignore filtered classes */
      class = class_info_get (f->class_info, i);
      if (class->new_id == -1) {
        filtercount++;
	continue;
      }
      if (class->method==NULL) {
	    nullcount++;
            tc_message("!");
      	    continue;
      }
      if (class->method->classifier != CLASSIFIER_PLAUM)
	tc_error("Packing only implemented for PLAUM");

      /* model local path to absolute path */
      model_file = g_strdup_printf ("%s%s%s", f->path_models, PATH_CONNECTOR, class->method->model_file);
      g_free(class->method->model_file);
      class->method->model_file = model_file;

      //the class is fine.. we need to read the model and store it
      c = classifier_new_read[class->method->classifier](class->method->model_file);      
      //here I use new method append_save of plaum
      classifier_append_save[class->method->classifier](c, path);
      classifier_free[class->method->classifier](&c);
      count++;
      tc_message("."); 
  }
  g_free(path);
  printf("\n%d classes filtered", filtercount);
  printf("\n%d classes with null method", nullcount);
  printf("\n%d valid models parsed", count);
  printf("\ndone!\n");
}
/* -------------------------------------------------------------------------- */
inline int file_exists(gchar *filename) {
  FILE *f = fopen(filename,"r");
  int ret = (f == NULL ? 0 : 1);
  fclose(f);
  return ret;
}
