/***************************************************************************
 *            fold.h
 *
 *  Thu Nov 27 16:40:40 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/
#ifndef _FOLD_H
#define _FOLD_H

#include <glib.h>
#include "class_info.h"
#include "term_info.h"
#include "methods.h"
#include "bitset.h"
#include "model.h"
#include "inverted_index.h"

/* 
   Fold attributes
*/
typedef struct fold_struct {
  ClassInfoArray class_info;
  TermInfoArray term_info;
  guint id;
  gchar *path;
  gchar *filepath_class_info;
  gchar *filepath_term_info;
  gchar *filepath_predictions;
  gchar *filepath_methods_eval;
  gchar *filepath_methods_filter;
  gchar *filepath_methods_test;
  gchar *filepath_train_samples;
  gchar *filepath_eval_samples;
  gchar *filepath_test_samples;
  gchar *filepath_train_samples_tmp;
  gchar *filepath_eval_samples_tmp;
  gchar *filepath_test_samples_tmp;
  gchar *filepath_sample_tmp;
  gchar *path_models;
  Methods methods;
  BitSet train_docs, eval_docs, test_docs;  
} *Fold;

/* 
   Transformation parameters 
*/
typedef struct fold_train_params_struct
{
  FeatureTransformations ftrans;	
  guint nfolds;                 /* number of folds for cross
				   validation */
  guint fig_size;		/* number of features with the highest
				   information gain value to be selected (0
				   means no filtering) */
  guint fdf_min;		/* minimal document frequency to allow a
				   feature (0 means no filtering) */       
  guint cdf_min; 	        /* minimal document frequency to allow a
				   class (0 means no filtering) */
  
  double beta;                  /* beta value for F measure */
  guint tree_balance;           /* method for balance: 0 id, 1 1/ig, 2
				   id/ig */
  guint tree_max_children;      /* maximum number of children in
				   m3tree */
  guint tree_max_depth;         /* maximum allowed depth for the tree
				   in its construction */
  gboolean tree_duplicates;     /* determines whether classes trained
				   in one level may appear in lower
				   levels */
  gboolean tree_traverse_all;   /* determines whether both positive
				   and negative subtress must be
				   explored */
  
  MethodSelection selection_measure; /* measure used for selecting
					best method */
  gdouble selection_threshold;       /* threshold for selection
					measure */
  gboolean compute_scut;             /* tells the system to compute
					the scut measure
					automatically */
  Methods methods;                   /* pointer to methods to be used
					for training (either stage1 or
					stage2 below) */
  
  Methods stage1methods, stage2methods;   /* first stage and second
					     stage list of methods for learning
					     and classifying */
  
  gchar *modelstr;              /* basename for model files */
  
  guint ntop;                   /* number of classes to return when
				   doing ranked classification */
  gboolean positive_ranked;     /* if TRUE when using ranking, only
				   classes above the classifier
				   threshold will be returned */

  /* test param */
  guint ntop_shown;             /* number of top classes showing its performance measures */
  
  /* stage 1 and 2 options (for filtering) */
  gboolean stage1_positive;
  guint stage1_ntop;
  gboolean stage2_positive;
  guint stage2_ntop;
  guint stage2_nbottom;
  gchar stage2_merge;

  /* Training approach 
     FOLD_TRAIN_BY_CLASS    We train a binary classifier for each class in a one-against-all fashion
     FOLD_TRAIN_MLTREE      We train a m3-tree with a class and a classifier per node
     FOLD_TRAIN_MATRIX      We train a dual system term-doc-doc-keyword obtaining a final matrix term-keyword 
  */
  guint approach;         
  	
} FoldTrainParams;

#define FOLD_FILE_TRAIN_SAMPLES "train_samples_ascii.vcs"
#define FOLD_FILE_EVAL_SAMPLES "eval_samples_ascii.vcs"
#define FOLD_FILE_TEST_SAMPLES "test_samples_ascii.vcs"
#define FOLD_FILE_SAMPLE_TMP "sample.tmp"
#define FOLD_FILE_CLASS_INFO "class.inf"
#define FOLD_FILE_TERM_INFO "term.inf"
#define FOLD_DIR_MODELS "models"
#define FOLD_FILE_PREDICTIONS "predictions.tmp"
#define FOLD_FILE_METHODS_EVAL "methods_eval.inf" /* methods with results of evaluation */
#define FOLD_FILE_METHODS_FILTER "methods_filter.inf" /* methods with results of class filtering */
#define FOLD_FILE_METHODS_TEST "methods_test.inf" /* methods with results of test */
#define FOLD_FILE_TREE "m3tree.inf" /* methods with results of test */
#define FOLD_TEST_BY_CLASS    1
#define FOLD_TEST_BY_TREE     2
#define FOLD_TEST_BY_RANK   3
#define FOLD_TEST_FILTERING   4
#define FOLD_TRAIN_BY_CLASS   1<<1  
#define FOLD_TRAIN_CLASS_FILTER   1<<2  
#define FOLD_TRAIN_MLTREE  1<<3
#define FOLD_TRAIN_MATRIX 1<<4
#define FOLD_TRAIN_ALL_TOGETHER 1<<5
#define FOLD_TEST_ALL_TOGETHER 1<<6

#define FOLD_STAGE2_UNION 1
#define FOLD_STAGE2_INTERSECTION 2
#define FOLD_STAGE2_DIFFERENCE 3

typedef void (*fold_test_func) (Fold * f, Model m, Methods *microavg, Methods *macroavg);

Fold fold_new(Model m, InvertedIndex folds, guint i, FoldTrainParams *params);
void fold_free (Fold * f);
Fold fold_new_read (gchar * path, guint i, InvertedIndex folds);
void fold_train (Fold * f, Model m, FoldTrainParams *params);
Method fold_train_class(Fold *f, tc_id class_id, 
			BitSet postrain_docs, BitSet negtrain_docs, 
			BitSet poseval_docs, BitSet negeval_docs, 
			FoldTrainParams* params, 
			Samples train_samples, Samples eval_samples);
void fold_test(Fold * f, Model m, Methods *cmicroavg, Methods *cmacroavg, Methods *dmicroavg, Methods *dmacroavg, guint by_what, FoldTrainParams *params);
void fold_get_docs (Fold * f, InvertedIndex folds);
Sample  fold_classify (Fold f, Sample s, FeatureTransformations ftrans);
Sample  fold_classify_huge (Fold f, Sample s, FeatureTransformations ftrans);
/* Function to compute correlation matrix between classes */
void fold_dump_correlation_matrix(Fold f, Model m);

void fold_export_models(Fold f, Model m, gchar *dir);
void fold_pack_models(Fold f);

#endif /* _FOLD_H */
