/***************************************************************************
 *            folds.c
 *
 *  Thu Nov 27 16:20:01 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/

#include "folds.h"
#include "fold.h"
#include "tc_utils.h"
#include "time.h"

#define RANDOM_FROM_CERO_TO(max) 

Folds __folds_new(gchar *datadir)
{
  Folds folds = g_new0(struct folds_struct, 1);
  folds->filepath_ii = g_strconcat(datadir, PATH_CONNECTOR, FOLDS_FILE_II, NULL);
  folds->filepath_idx = g_strconcat(datadir, PATH_CONNECTOR, FOLDS_FILE_IDX, NULL); 
  folds->path = g_strdup (datadir);
  return folds;
}

/* -------------------------------------------------------------------------- */
/* Distributes documents in model 'model' into 'n' different folds */
Folds folds_new(Model m, FoldTrainParams *params)
{
  tc_id doc;
  Folds folds;
  guint i;

  tc_message ("Distributing files in folds...");

  folds = __folds_new(m->datadir);
  folds->folds = inverted_index_new ();
  folds->len = params->nfolds;
  folds->it = 0;
  doc = m->doc_id->max_id;
	
  /* we set random seed */
  srand ((long) time(NULL));

  /* we put documents in folds randomly, creating inverted index */
  while (doc >= 0) {
    tc_id fold = ((tc_id)(((double)(folds->len-1)+1.0)*rand()/(RAND_MAX+1.0))); /* from 0 to n-1 */
    inverted_index_add (&folds->folds, fold, doc--);
  }		
  inverted_index_save(folds->folds, folds->filepath_ii, folds->filepath_idx);
  inverted_index_free(&folds->folds);
  folds->folds = inverted_index_new_read(folds->filepath_ii, folds->filepath_idx);
  tc_message ("done\n");
	
  for (i = 0; i < folds->len; i++)
    {
      Fold fold;
      tc_message_int("\t-------------- Working on fold %d\n", i+1);
		
      /* create fold data */
      fold = fold_new(m, folds->folds, i, params);
      fold_free(&fold);
		
      tc_message("\tdone\n");
    }	
	
  return folds;
}
/* -------------------------------------------------------------------------- */
void folds_train(Folds *folds, Model m, FoldTrainParams *params) {

  tc_id i;	
	
  tc_message ("Training folds...\n");

  /* for each fold */
  for (i = 0; i < (*folds)->len; i++)
    {
      Fold fold;
      tc_message_int("\t-------------- Working on fold %d\n", i+1);
		
      fold = fold_new_read((*folds)->path, i, (*folds)->folds);
      fold_train(&fold, m, params);
      fold_free(&fold);
      tc_message("going to next fold\n");
    }
	
  tc_message ("done\n");
}
/* -------------------------------------------------------------------------- */
void folds_test(Folds *folds, Model m, guint by_what, FoldTrainParams *params) 
{
  tc_id i;	
  Methods cmicroavg, cmacroavg, dmicroavg, dmacroavg;
  Method cfinal_microavg, cfinal_macroavg, dfinal_microavg, dfinal_macroavg;
	
  tc_message ("Testing folds...\n");

  /* methods averaging */
  if (by_what == FOLD_TEST_BY_CLASS) {
	cmicroavg = methods_new();
	cmacroavg = methods_new();
  }

  dmicroavg = methods_new();
  dmacroavg = methods_new();

  /* for each fold */
  for (i = 0; i < (*folds)->len; i++) {			
	Fold fold = fold_new_read((*folds)->path, i, (*folds)->folds);
	fold_test(&fold, m, &cmicroavg, &cmacroavg, &dmicroavg, &dmacroavg, by_what, params);
	fold_free(&fold);
  }
	
  printf("-----------------------------------------------------------------------------------------------------------------\n");
  printf("FINAL RESULTS\n");
  
  if (by_what == FOLD_TEST_BY_CLASS) {
	cfinal_microavg = methods_micro_avg(cmicroavg);	
	cfinal_macroavg = methods_macro_avg(cmacroavg);	
	printf("%.2lf classes tested, %.2lf classes where successfully trained\n", cfinal_macroavg->nctest, cfinal_macroavg->nctrain);
   	printf("BY CLASS TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
	printf("avg) ");
	method_dump_measures(cfinal_microavg);
	printf("AVG) ");
	method_dump_measures(cfinal_macroavg);

	method_free(&cfinal_microavg);
	method_free(&cfinal_macroavg);
	methods_free(&cmicroavg);
	methods_free(&cmacroavg);
  }

  dfinal_microavg = methods_micro_avg(dmicroavg);	
  dfinal_macroavg = methods_macro_avg(dmacroavg);	

  printf("BY DOC.  TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
  printf("avg) ");
  method_dump_measures(dfinal_microavg);
  printf("AVG) ");
  method_dump_measures(dfinal_macroavg);

  method_free(&dfinal_microavg);
  method_free(&dfinal_macroavg);
  methods_free(&dmicroavg);
  methods_free(&dmacroavg);

  tc_message ("done\n");
}
/* -------------------------------------------------------------------------- */
void folds_classify(Folds *folds, Model m, FeatureTransformations ftrans, gchar *document)
{
  tc_id i;	
  Sample s, classes;
	
  /* we parse the document */
  tc_message_str ("Parsing document '%s' ...", document);
  s = model_plain_text_process(&m, document, 0);
  tc_message ("done\n");
	
  /* for each fold */
  for (i = 0; i < /*(*folds)->len*/1; i++) {	
	Fold fold = fold_new_read((*folds)->path, i, (*folds)->folds);
	classes = fold_classify(fold, s, ftrans);
	model_show_classes_strings(m, classes);
	sample_free(&classes);
	fold_free(&fold);
  }	
  sample_free(&s);
}
/* -------------------------------------------------------------------------- */
void folds_classify_huge(Folds *folds, Model m, FeatureTransformations ftrans, gchar *document)
{
  tc_id i;	
  Sample s, classes;
	
  /* we parse the document */
  tc_message_str ("Parsing document '%s' ...", document);
  s = model_plain_text_process(&m, document, 0);
  tc_message ("done\n");
	
  /* for each fold */
  for (i = 0; i < /*(*folds)->len*/1; i++) {	
	Fold fold = fold_new_read((*folds)->path, i, (*folds)->folds);
	classes = fold_classify_huge(fold, s, ftrans);
	model_show_classes_strings(m, classes);
	sample_free(&classes);
	fold_free(&fold);
  }	
  sample_free(&s);
}
/* -------------------------------------------------------------------------- */
/* Reads from disk */
Folds 
folds_new_read(gchar *path)
{
  Folds folds = __folds_new(path);
  folds->folds = inverted_index_new_read(folds->filepath_ii, folds->filepath_idx);
  folds->len = folds->folds->max_id+1;
  return folds;
}
/* -------------------------------------------------------------------------- */
/* Frees memory */
void 
folds_free(Folds *folds)
{
  if ((*folds)->folds)
    inverted_index_free(&(*folds)->folds);	
  if ((*folds)->path)
    g_free((*folds)->path);
  if ((*folds)->filepath_ii)
    g_free((*folds)->filepath_ii);
  if ((*folds)->filepath_idx)
    g_free((*folds)->filepath_idx);
  g_free(*folds);	
}
