#include <stdio.h>
#include <stdlib.h>
#include "inverted_index.h"
#include "tc_utils.h"

/* -------------------------------------------------------------------------- */
/* Creates a new inverted index  */
InvertedIndex
inverted_index_new (void)
{
  InvertedIndex ii = g_new(struct inverted_index_struct, 1);

  ii->vmax_id = g_array_new (FALSE, TRUE, sizeof (tc_id));
  if (!ii->vmax_id)
    tc_error("inverted_index_new: no memory for vector of max_id\n");
  ii->voffset = g_array_new (FALSE, TRUE, sizeof (long));
  if (!ii->voffset)
    tc_error("inverted_index_new: no memory for vector of offset\n");
  ii->fdata = tmpfile ();	/* we create a new temporal file */
  if (!ii->fdata)
    tc_error("inverted_index_new: unable to create temporal file\n");	
  ii->max_id = -1;
  /* dummy fill of the data file to avoid have data at offset 0 */
  fwrite(&ii->max_id, sizeof(tc_id), 1, ii->fdata);
  return ii;
}

/* -------------------------------------------------------------------------- */
/* Adds to the inverted index the element 'e' to the list at position 'i' */
void
inverted_index_add (InvertedIndex * ii, tc_id i, tc_id e)
{
  long prevpos, curpos, *prevposp;
  tc_id *pmax_id;

  /* resize vectors when needed */
  if ((*ii)->voffset->len <= i)		
    {			/* no list at 'i' in the vector */
      (*ii)->voffset = g_array_set_size ((*ii)->voffset, i + 1);
      (*ii)->vmax_id = g_array_set_size ((*ii)->vmax_id, i + 1);
    }
	
  /* get position of previous added item */
  if (g_array_index((*ii)->voffset, long, i))
    {
      prevposp = &g_array_index ((*ii)->voffset, long, i);
      prevpos = *prevposp;
    }
  else       /* first element added to the list */
    {
      prevpos = -1;
      prevposp = &g_array_index ((*ii)->voffset, long, i);
    }

  /* actual position in file */
  curpos = ftell ((*ii)->fdata);
  fwrite (&e, sizeof (tc_id), 1, (*ii)->fdata);	/* writes element */
  fwrite (&prevpos, sizeof (long), 1, (*ii)->fdata);	/* writes position to 
							 * next one in file */
  /* now we update offset */
  *prevposp = curpos;

  /* We update max_id in the list*/
  pmax_id = &g_array_index ((*ii)->vmax_id, tc_id, i);
  if (*pmax_id < e)
    *pmax_id = e;
	
  /* We update max_id in for the inverted index */
  if ((*ii)->max_id < i) 
    (*ii)->max_id = i;
}

/* -------------------------------------------------------------------------- */
/* Save inverted index to disk
   data -- is the name of the file with the inverted index
   index -- is the name of the file with the index (offsets) 

   NOTE: no further operations should be performed with the data after 
   calling this function. 'inverted_index_free' should be called right
   after this function.
*/
void
inverted_index_save (InvertedIndex ii, char *data, char *index)
{
  FILE *fidx = fopen (index, "wb");
  FILE *fdata = fopen (data, "wb");
  guint i;
  long offset;

  if (!fidx)
    tc_error_str("inverted_index_save: unable to create file '%s'\n", index);
  if (!fdata)
    tc_error_str("inverted_index_save: unable to create file '%s'\n", data);

  /* We create bitsets and serialize them into a file, 
   * afterwards we destroy them! 
   * Also the index is created at same time 
   */

  /* we write size of vector of offsets */
  fwrite (&ii->voffset->len, sizeof (guint), 1, fidx);

  fflush(ii->fdata);
  for (i = 0; i < ii->voffset->len; i++)
    {
      BitSet bs;
      long next;
      tc_id element;

      /* index of offsets */
      offset = ftell (fdata);
      fwrite (&offset, sizeof (long), 1, fidx);

      /* inverted index */
      bs =
	bitset_new ((tc_id)
		    g_array_index (ii->vmax_id, tc_id, i));
		
      /* we pass through the file reading linked elements */
      next = g_array_index (ii->voffset, long, i);
      while (next != -1)
	{
	  fseek (ii->fdata, next, 0);
	  if (fread (&element, sizeof (tc_id), 1, ii->fdata)!=1)
	    tc_error ("inverted_index_save: invalid format");
	  if (fread (&next, sizeof (long), 1, ii->fdata) != 1)
	    tc_error ("inverted_index_save: invalid format");
	  bitset_insert (&bs, element);
	}
      bitset_serialize (bs, fdata);
      bitset_free (&bs);
    }

  fclose (fidx);
  fclose (fdata);
}

/* -------------------------------------------------------------------------- */
/* Loads and returns an inverted index from the disk 
   data -- is the name of the file with the inverted index
   index -- is the name of the file with the index (offsets)

   NOTE: this function can be called ONLY when the inverted index has been
   created using 'inverted_index_new' and never when the inverted index
   was load from another file.
*/
InvertedIndex
inverted_index_new_read (char *data, char *index)
{
  InvertedIndex ii = g_new(struct inverted_index_struct, 1);
  FILE *fidx;
  guint size = 0;

  ii->vmax_id = NULL;	/* not used */
  ii->voffset = NULL;

  /* We read the index of offsets */
  fidx = fopen (index, "rb");
  if (!fidx)
    tc_error_str("inverted_index_new_read: unable to open file '%s'\n", index);


  if (fread (&size, sizeof (guint), 1, fidx) != 1)
    {
      fclose (fidx);
      tc_error("inverted_index_new_read: invalid index file\n");
    }
  
  ii->voffset = g_array_new (FALSE, TRUE, sizeof (long));
  ii->voffset = g_array_set_size (ii->voffset, size);

  if (!ii->voffset)
    {
      fclose (fidx);
      tc_error("inverted_index_new_read: no memory of vector of offsets\n");
    }

  if (fread (ii->voffset->data, sizeof (long), size, fidx) != size)
    {
      fclose (fidx);
      tc_error("inverted_index_new_read: invalid index file\n");
    }

  fclose(fidx);

  /* Now we leave open the file of data and that's all folks! */
  ii->fdata = fopen (data, "rb");
  if (!ii->fdata)
    tc_error_str("inverted_index_new_read: unable to open file '%s'\n", data);

  ii->max_id = size - 1;
  return ii;
}

/* -------------------------------------------------------------------------- */
/* Returns the bitset for a given index */
BitSet
inverted_index_get (InvertedIndex ii, tc_id id)
{
  return bitset_new_deserialize (ii->fdata,
				 g_array_index (ii->voffset, long, id));
}

/* -------------------------------------------------------------------------- */
/* Frees memory and close files if needed */
void
inverted_index_free (InvertedIndex * ii)
{
  if (!(*ii))
    return;

  /* we remove arrays */
  if ((*ii)->vmax_id)
    g_array_free ((*ii)->vmax_id, TRUE);
  if ((*ii)->voffset)
    g_array_free((*ii)->voffset, TRUE);

  /* and we liberate the file if open */
  if ((*ii)->fdata)
    fclose ((*ii)->fdata);

  /* finally, we destroy struct */
  g_free (*ii);

}

/* -------------------------------------------------------------------------- */
/* Dumps the whole inverted index only for readed from file inverted indexes */
void
inverted_index_dump (InvertedIndex ii)
{
  tc_id i;
  BitSet bs;
  for (i = 0; i <= ii->max_id; i++)
    {
      bs = inverted_index_get (ii, i);
      printf ("%d -> ", i);
      bitset_dump (bs);
    }
}
/* -------------------------------------------------------------------------- */
/* FOR TESTING */
//~ int
//~ main (void)
//~ {
//~ InvertedIndex ii = inverted_index_new ();
//~ guint i, total = 10000, j;
//~ printf ("adding\n");
//~ for (i = 0; i < total; i++)
//~ {
//~ //printf("%d -> ", i);
//~ guint totalj =      1 +     (guint) (((float) total) * rand () / (RAND_MAX + 1.0));
//~ //printf("(%d) ", totalj);
//~ for (j = 0; j < 200; j++)
//~ {
//~ //printf(" %d", j);
//~ inverted_index_add (&ii, i, j);
//~ }
//~ //printf("\n");
//~ printf ("%d \r", i);
//~ fflush (stdout);
//~ }
//~ printf ("saving\n");
//~ inverted_index_save (ii, "ii.data", "ii.idx");
//~ printf ("saved!\nfreeing\n");
//~ inverted_index_free (&ii);
//~ printf ("free!\n");

//~ InvertedIndex ii2 = inverted_index_new_read ("ii.data", "ii.idx");
//~ inverted_index_dump (ii2);
//~ inverted_index_free (&ii2);
//~ }
