#ifndef __INVERTED_INDEX_H_
#define __INVERTED_INDEX_H_

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include "bitset.h"
#include "tc_types.h"

/* This is the Inverted Index object,
   it implements a vector of integers pointing to a list of integers for large
   data 
   The first vector would be in memory, and the second in disk.
   The way of use is (in the given order)
   
   1. CREATION:
      - 1 inverted_index_new
	  - N inverted_index_add
	  - 1 inverted_index_save
	  - 1 inverted_index_free
	  
   2. USE:
      - 1 inverted_index_new_read
	  - N inverted_index_get
	  - 1 inverted_index_free

	1 represents only one invocation per inverted index, N representes several
	invocations per inverted index are allowed.
	
	We CANNOT use inverted_index_get for a newly created inverted index.

*/

typedef struct inverted_index_struct 
{
	GArray *vmax_id;	    /* vector of maximal values in lists */
	GArray *voffset;        /* vector of offsets to the file */
	tc_id max_id;			/* max identificator in the inverted index (maximal index) */
	FILE *fdata;            /* file to data */
} *InvertedIndex;

/* Creates a new inverted index ready to store, for the time being, 
  'initial_size' elements. If more memory is needed it will resize 
  automatically (TBD) 
  */
InvertedIndex inverted_index_new(void);

/* Adds to the inverted index the element 'e' to the list at position 'i' */
void inverted_index_add(InvertedIndex *ii, tc_id i, tc_id e);

/* Save inverted index to disk
   data -- is the name of the file with the inverted index
   index -- is the name of the file with the index (offsets) 

   NOTE: no further operations should be performed with the data after 
   calling this function. 'inverted_index_free' should be called right
   after this function.
*/
void inverted_index_save(InvertedIndex ii, char *data, char *index);

/* Loads and returns an inverted index from the disk 
   data -- is the name of the file with the inverted index
   index -- is the name of the file with the index (offsets)

   NOTE: this function can be called ONLY when the inverted index has been
   created using 'inverted_index_new' and never when the inverted index
   was load from another file.
*/
InvertedIndex inverted_index_new_read(char *data, char *index);

/* Returns the bitset for a given index, the bitset MUST be destroy with 
bitset_free() */
BitSet inverted_index_get(InvertedIndex ii, tc_id id);

/* Frees memory and close files if needed */
void inverted_index_free(InvertedIndex *ii);

/* Dumps the whole inverted index */
void inverted_index_dump (InvertedIndex ii);

#endif /*__INVERTED_INDEX_H_*/
