#include <argp.h>
#include <glib.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <config.h>

#include "pt_parser.h"
#include "tc_types.h"
#include "tc_utils.h"
#include "model.h"
#include "methods.h"
#include "folds.h"
#include "fold.h"
#include "samples.h"
#include "classifiers/classifiers.h"

#define DEFAULT_STOP_WORDS "stop.txt"

gchar *str, *learn_command = NULL;

/* Possible actions performed by the program */
typedef enum {
  TC_BUILD_MODEL = 1 << 0,
  TC_MAKE_FOLDS = 1 << 1,
  TC_TRAIN_BY_CLASS = 1 << 2,
  TC_TRAIN_ALL  = 1 << 3,
  TC_TEST_CLASSES = 1 << 4,
  TC_TEST_RANKED = 1 << 5,
  TC_TEST_TREE = 1 << 6,
  TC_CLASSIFY = 1 << 7,
  TC_CLASS_CORRELATION = 1 << 8,
  TC_EXPORT_MODELS = 1 << 9,
  TC_DUMP_TERM_IDS = 1 << 10,
  TC_DUMP_CLASS_IDS = 1 << 11,
  TC_TEST_FILTERING = 1 << 12,
  TC_PACK_MODELS = 1 << 13,
  TC_TEST_ALL = 1 << 14,
  TC_DUMP_FOLD_TERM_IDS = 1 << 15,
} TCAction;

/* program parameters */
typedef struct params_struct
{
  gchar *datadir;                /* directory for model data */
  gchar *ptextdir;                  /* directory where full text files are */
  gchar *classdir;                   /* directory where classes files are */
  TCAction action;               /* action to perform */
	
  /* Documents processing params */
  PTParserParams pt_params;      /* parameters for plain text parsing */
  guint minfreq;				   /* minimal frequency allowed */
  gchar *document;			   /* document to be classify */
  gboolean hugemodel;			   /* huge model file to be classified */
	
  /* Training params */
  FoldTrainParams train_params;

  /* Fold to be dumped */
  int fold;
	
  gboolean verbose;
} TCParams;

TCParams params;
gboolean verbose;

/* Program documentation */

const char *argp_program_version = "tecat 0.3";
const char *argp_program_bug_address = "<amontejo@ujaen.es>";
static char doc[] =
"TECAT -- The TExt CATegorization framework\nArturo Montejo Raez - University of Jaen 2006 \n\
\vYou can distribute this software freely.";

/* A description of the arguments we accept. */
static char args_doc[] = "DATADIR";

/* Suboptions for classifier */
enum {
  METHOD = 0,
  LEARN,
  CLASSIFY,
  FORMAT
};

const char * classifier_opts[] = {
  [METHOD] = "method",
  [LEARN] = "learn",
  [CLASSIFY] = "classify",
  [FORMAT] = "format"
};

/* Suboptions for filtering stages */
enum {
  POSITIVE = 0,
  NTOP,
  NBOTTOM,
  MERGE
};

const char * stage1_opts[] = {
  [POSITIVE] = "positive",
  [NTOP] = "ntop"
};


const char * stage2_opts[] = {
  [POSITIVE] = "positive",
  [NTOP] = "ntop",
  [NBOTTOM] = "nbottom",
  [MERGE] = "merge"
};

/* The options we understand. */
static struct argp_option options[] = {

  {"index-collection", 'i', 0, 0, "Index full text documents (see INDEX OPTIONS)"}, 
  {"make-folds", 'k', 0, 0, "Prepares folds for training classifiersm (see FOLDING OPTIONS)"}, 
  {"train-by-class", 't', 0, 0, "Train the system using given classifiers against each class (see TRAIN BY CLASS OPTIONS)"}, 
  {"train-tree", 'T', 0, 0, "Train the system with a method which consider all the classes at once (the MLTree algorithm). NOTE: THIS CLASSIFIER IS NOT OPERATIONAL YET."},
  {"train-all", 'a', 0, 0, "Train the system with an external method which consider all the classes at once (e.g. boostexter)."},
  {"test-classes", 'z', 0, 0, "Tests the trained system of a serial of classes returning only classes ranked with positive or over threshold score."}, 
  {"test-ranked", 'R', 0, 0, "Tests the trained system of a serial of classes getting top ranked classes using classification score."}, 
  {"test-filtering", 'O', 0, 0, "Tests the trained system of a serial of classes applying second-stage filtering on proposed classes."}, 
  {"test-all", 'o', 0, 0, "Tests the trained system using an external command that produces directly the set of labels (e.g. boostexter)."}, 
  {"test-tree", 'K', 0, 0, "Tests the trained system using the MLTree algorithm. NOTE: THIS CLASSIFIER IS NOT OPERATIONAL YET."}, 
  {"classify", 'X', 0, 0, "Classifies a document."}, 
	
  {0, 0, 0, 0,
   "DATADIR is the directory where the model is or will be stored"},

  /* --------------- BUILD -------------------------------------------- */
  {0, 0, 0, 0,
   "INDEX OPTIONS -----------------------------------------------\n"},
	 
  {"full-text-dir", 'F', "DIR", 0, 
   "Directory where full text files are stored"}, 
		
  {"classes-dir", 'C', "DIR", 0, 
   "Directory where classes for each file are stored"}, 
		
  {"max-term-length", 'x', "N", 0, 
   "Maximal length for a term allowed (default 40, 0 means no filtering)"},
		
  {"min-term-length", 'm', "N", 0,
   "Minimal length for a term allowed (default 0, means no filtering)"},

  {"min-freq", 'r', "N", 0,
   "Minimal frequency for a term allowed (default 0, 0 means no filtering)"},

  {"stop-words", 's', "FILE", 0,
   "File containing stop words for removal (default 'stop.txt' at current directory"},
		
  {"no-stop-words-removal", 'S', 0, 0, "Disable perform stop words removal (default enabled)"},
		
  {"no-stemming", 'E', 0, 0, "Disable perform Porter stemming (default enabled)"},

  /* --------------- FOLDS PREPARATION --------------------------------- */
  {0, 0, 0, 0,
   "FOLDING OPTIONS ---------------------------------------------\n"},
		 
  {"folds", 'f', "N", 0,
   "Number of folds for cross-validation (default 10)"},
	 
  {"weighting-entropy", 'w', 0, 0,
   "Weighting scheme using entropy (default TF.IDF)"},
		 
  {"weighting-tfidf", 'W', 0, 0,
   "Weighting scheme using TF.IDF (default)"},
		 
  {"apply-icf", 'e', 0, 0,
   "Apply inverse class frequency as global weight for every term."},

  /* TBD {"global-weighting-level", 'G', "N", 0,
     "Level at which we apply global weighting: 0 (at document level), 1 (at class level) or 2 (both)"}, */

  {"ig-feature-selection", 'g', "N", 0,
   "Selects only N features with the highest information gain (default 50000, 0 means all selected)"},

  {"no-cosine-normalization", 'N', 0, 0,
   "Term weights will not be normalized by the cosine factor"},
		 
  {"df-feature-selection", 'd', "N", 0,
   "Selects only features appearing in at least N documents of the training set (default 1, 0 means all selected)"},
		 
  {"df-class-selection", 'j', "N", 0, 
   "Selects only classes appearing in training, evaluation and test sets and with at least N documents for training (default 1, 0 means select all)"},

  /* --------------- TRAINING ------------------------------------------ */

  {0, 0, 0, 0,
   "TeCaT works in two modes: training a binary classifier for each class or training a tree based on binary classifiers. If you specify '--train-by-class' option a binary classifier for each class will be trained in a one-agains-all basis. If '--train-tree' is specified, then a ML-Tree will be constructed. For both approaches, we can specify as many binary calssifiers as wanted (see BINARY CLASSIFIERS).\n"},

  {0, 0, 0, 0,
   "CLASSIFIERS OPTIONS --------------------------------\n"},
	 
  {"train-filter", 'I', 0, 0,
   "With this option, second-stage classification will be performed. That means training additional per-class classifiers based on class assignments as samples. Thus, once classes have been produced by first-stage classifiers (using the --classifier's entries defined before this option), filters will be ftrained applying additional classifiers (defined with next --classifier's entries)."},

  {"classifier", 'c', "OPTIONS", 0,
   "A binary classifier to be used by TECAT. See options below."},

  {"beta-F", 'b', "VALUE", 0,
   "Beta value for F measure (default 1.0)."},

  {"selection-measure", 'u', "N", 0,
   "Measure to use for selecting best method. Values of N: 0 (by precision), 1 (by recall), 2 (by fallout), 3 (by accuracy), 4 (by error), 5 (by distance to optimal point), 6 (by F measure), 7 (by F1 measure, default), 8 (by break-even-point)."},

  {"selection-threshold", 'U', "N", 0,
   "Threshold to be used for the selection measure. This value is 0.0 by default. Please, not that this value use to be in [0,1] interval and depends in the selection measure used."},

  {"compute-scut", 'Q', 0, 0,
   "The S-Cut approach will be apply to compute, using the validation samples, an automatic threshold on the decision value (default off, the decision value is then 0.0). NOTE: No effect for PLAUM classifier."},

  {0, 0, 0, 0,
   "ALL-TOGETHER OPTIONS ------------------------------------"},

  {0, 0, 0, 0,
   "This method uses an external command for training and classify, so --learn and --classify commands must be provided.\n"},


  {0, 0, 0, 0,
   "ML-TREE OPTIONS ------------------------------------\n\nNOTE: THIS CLASSIFIER IS NOT OPERATIONAL YET."},

  {0, 0, 0, 0,
   "Above options are valid here. Additional, to control tree generation and behaviour you have:\n"},

  {"tree-max-children", 'H', "N", 0,
   "Maximum number of children per node when training a MLTree (default 4, 0 means all children will be accepted until one of them fails in the evaluation). Be carefull when setting this parameter, since the complexity of the MLTree training algorithm is of order O(n^log2(m)) being m the number of classes and n the number of children per level (the tree-max-children parameter)."},

  {"tree-max-depth", 'L', "N", 0,
   "Maximum depth of the MLTree allowed when generating it (default 7, 0 means no limit)."},

  {"tree-allow-duplicates", 'Z', 0, 0,
   "Determines whether classes already trained in a level may appear in lower levels (default off, add this parameter to set it on)."},

  {"tree-balance", 'B', "N", 0,
   "Method to be used for determining the balance of a class: 0 (id, default), 1 (1/ig), 2 (id/ig), 3 (Sum_children(id)), 4 (Sum_children(id)/ig); where id is the imbalance degree, ig the information gain and Sum_children the summatory of imbalance degrees for potential children."},

  {0, 0, 0, 0,
   "BINARY CLASSIFIERS \n"},

  {0,0,0,0,
   "There are two types of classifiers: external and built-in. External classifiers are external programs which will be called by TECAT following. To call them properly some options has to be passed. Built-in classifiers are classification algorithms implemented internally by TECAT, so you don't need to specify anything but certain parameters that the classification method may require."},

  {0,0,0,0,
   "OPTIONS for built-in classifiers:"},

  {0,0,0,0,
   "method=[plaum|rocchio|wh|eg|lvq|nb|genetic]\n Selects a built-in classifier to be added to the list of trained classifiers. They are the PLAUM perceptron (plaum), the Rocchio algorithm (rocchio), the Widrow-Hoff algorithm (wh), the Kivinen-Warmuth algorithm (kw), the LVQ neural network (lvq), and the Naive-Bayes network (nb). Each built-in classifier has additional parameters which can be specified using colons, e.g.:\n\n    --classifier=method=plaum:pos_tau=2.0:neg_tau=-1.5:iter=100"},

  {0,0,0,0,
   "** PLAUM (plaum)\npos_tau:    margin for positive samples (default +1.0)\nneg_tau:    margin for negative samples (default -1.0)\niterations: number of maximum iterations (default 50)\neta:        value for eta (default 1.0)\nFor more information about his algorith, see [1]." },

  {0,0,0,0,
   "** Rocchio (rocchio)\nalpha:       weight for previous weights (default 0.0)\nbeta:        weight for positive samples (default 1.0)\ngamma:       weight for negative samples (default 1.0)\nnonnegative: 0 (negative weights are preserved) or 1 (negative weights are set to 0, default)"},

  {0,0,0,0,
   "** Naive-Bayes (nb)\nNOT AVAILABLE"},

  {0,0,0,0,
   "** Widrow-Hoff (wh)\neta:     learning rate (default 1/(4*||x||^2), being ||x|| the maximum norm in the set of samples). This value must be positive.\naverage: 0 (use final computed weight) or 1 (use average of all computed weights, default)"},

  {0,0,0,0,
   "** Exponentiated Gradient (eg)\nAlso known as \"Kivinen & Warmuth\" or EG algorithm\neta:     learning rate (default 1/(3*R^2) where R is the difference between the maximum weight minus the minimum weight of a term.\nNOTE: THIS CLASSIFIER IS NOT OPERATIONAL YET."},

  {0,0,0,0,
   "** LVQ (lvq)\niterations: number of maximum iterations of the LVQ loop (default 50*ncbv)\nncbv:       number of codebook vectors for positive and negative sides (default 5)\nncbv_pos:      number of positive codebook vectors\nncbv_neg:      number of negative codebook vectors\nnormcbv:    normalization base for imbalanced codebookvectors\nalpha:      initial value for learning rate (default 0.05)\nWhen using normcbv the value passed represents the number of codebook vectors used for the less represented side (positive or negative samples), being assigned to the most represented side a number of codebook vectors proportional to the imbalance degree in basis of the value passed; e.g. if there are 10 positive samples and 100 negative ones and the value passed is 5, then there will be 5 positive codebook vectors and 50 negative ones."},

#ifdef HAVE_LIBGA
  {0,0,0,0,
   "** Genetic algorithm (genetic)\npopsize:    population size (default 100)\nngen:       number of generations (default 100)\npmut:       probability of mutation (default 0.1)\npcross:     probability of crossing over (default 0.6)\nNOTE: THIS CLASSIFIER IS NOT OPERATIONAL YET."},
#else
  {0,0,0,0,
   "** Genetic algorithm (genetic)\nNOT AVAILABLE"},
#endif

  {0,0,0,0,
   "--- OPTIONS for external classifiers:"},

  {0, 0, 0, 0, 
   "learn=<learn command>\nThis is the command used to train the classifier. It is an string following the format specified below.\n \nclassify=<classify command> \nThis is the command use for classifying using the computed model in the learning phase. It is an string following the format specified below. \n\nformat=[vectors|arff]\nThis is the format used create the samples for learning and classification.\n\nFormat for 'learn' and 'classify' strings: \nSince we need to place model, samples and predictions files properly in the invocation of the program, we will use [m], [s], [p] as patterns to represent the location of those files in the invocation of the program. Also, some algorithms accept a weight factor for positive samples as a measure against imbalanced classes, for that you can use the [f] pattern, which will be replaced by the fraction of positive samples over negative samples. See FORMATS below"},

  {0, 0, 0, 0,
   "FORMATS:"},

  {0, 0, 0, 0, 
   "** vectors\nThe samples file following the 'vectors' format will contain a sample (document) at each line in ASCII format, with the class label set to +1 or -1 before list of pairs <features>:<weight>; e.g.\n\n+1 1:3.4 56:23.4 1002:-0.423233\n-1 23:0.0001 435:1.3232002 2023:23.34232\n-1 532:3.0 5000:-0.34"},

  {0, 0, 0, 0, 
   "** arff\nThe samples file following the 'arff' format will be compliant with standard ARFF files used by Weka (see http://www.cs.waikato.ac.nz/~ml/weka/arff.html).\nNOTE: THIS FORMAT IS NOT AVAILABLE YET."},

  {0, 0, 0, 0, 
   "You can specify as many pairs classifiers as you want. Examples: \n\n./tecat --train                                           \\\n      --classifier=method=plaum                           \\\n      --classifier=learn=\"svm_learn -j [f] [s] [m]\",             \\\n                   classify=\"svm_classify [s] [m] [p]\",   \\\n                   format=vectors\n      --classifier=learn=\"svm_learn [s] [m]\",             \\\n                   classify=\"svm_classify [s] [m] [p]\",   \\\n                   format=vectors"},


  /* --------------- CLASSIFICATION OF A DOCUMENT -------------------------------- */
  {0, 0, 0, 0,
   "TESTING OPTIONS ---------------------------------------------\n"},

  {"top-shown", 'M', "N", 0,
   "In the verbose output, it determines the number of top classes that will be shown per fold with its individual performance measures (default 10).\n"},  

  {"tree-traverse-all", 'Y', 0, 0,
   "Dives down both positive and negative subtrees, no matter the classification sign returned by a classifier node (disabled by default).\n"},  

  {"top-classes", 'P', "N", 0,
   "When --test-ranked is used, it determines that N classes with the hights ranked status value will returned.\n"},  

  {"only-positive", 'A', "[0|1]", 0,
   "When --test-ranked is used, it determines that only classes with a score above the classifier threshold (a positive class) should be considered in the top ranking (default 1, true).\n"},  

  {0, 0, 0, 0,
   "Testing options when using --test-filtering\n"},  
  
  {"stage1", '1', "<stage1 options>", 0, 0},

  {0, 0, 0, 0,
   "Sets the parameters to be used in the first stage. Parameters are given in the form \"param=value,param=value\". The params are:\n\npositive=0|1:      0 no thresholded classes, 1 classes will be thresholded\nntop=N:            N top ranked classes will be returned\n"},

  {"stage2", '2', "<stage2 options>", 0, 0},

  {0, 0, 0, 0,
   "Sets the parameters to be used in the second stage (the one for filtering classes using classifiers trained on classes). Parameters are given in the form \"param=value,param=value\". The params are:\n\npositive=0|1:      0 no thresholded classes, 1 classes will be thresholded\nntop=N:            N top ranked classes will be returned\nnbottom=N:         N bottom ranked classes will be returned (this is useful for discarding classes)\nmerge=<operation>: Determines the operation used to merge classes from both stages. <operation> can be 'union' (for returning results from both stages), 'intersection' (for returning only those classes appearing in both stages) or 'difference' (for deleting from classes returned at first stage the ones obtained in the second stage)"},

  {0, 0, 0, 0,
   "CLASSIFICATION OPTIONS --------------------------------------\n"},
		
  {"document", 'D', "FULLTEXT", 0,
   "Path to the plaintext document to be classify."},

  {"hugemodel", '9', 0, 0,
   "Classify using a single model file"},
		
  {0, 0, 0, 0, 
   "Option -w (entropy weighting) or -W  (TF.IDF weighting) may be specify to comply with the way documents were processed in training (default TF.IDF)."},

  {0, 0, 0, 0,
   "References --------------------------------------------------"},

  {0, 0, 0, 0, 
   "[1]. Li. Y. and Zaragoza H., Herbrich R., Shawe-Taylor J. and Kandola J. 'The Perceptron Algorithm with Uneven Margins'. Proceedings of the International Conference of Machine Learning (ICML'2002), 2002."},

  {0, 0, 0, 0,
   "Other options -----------------------------------------------\n"},

  {"quiet", 'q', 0, 0, "Produces no verbose output on processing"}, 

  {"class-correlation", '3', 0, 0, "Output the correlation matrix class-to-class of every folder"}, 

  {"export-models", '4', "DIR", 0, "Models for each classifier from the first fold will be saved to the given directory, using the string of the class as file name."}, 

  {"dump-term-ids", '5', 0, 0, "Dumps the translation table of terms to integer ids. Note that these ids are those generated at first parsing, as each fold has its own term translation. Therefore, ids found in fold samples and models do not correspond to those generated by this option and option '8' should be used instead."}, 

  {"dump-fold-term-ids", '8', "N", 0, "Dumps the translation table of terms to integer ids for the fold number 'N'."}, 

  {"dump-class-ids", '6', 0, 0, "Dumps the translation table of classes to integer ids."}, 

  {"pack-models", '7', 0, 0, "Packs all models in every fold into a huge file"}, 

  {0}
};

/* Parse a single option. */
static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{

  /* Get the input argument from argp_parse, which we
   * know is a pointer to our plaum_params structure. */

  guint i;
  char * subopts, *value;
  Method method;
  gchar *learn_cmd = NULL, *classify_cmd = NULL, *format = NULL;
  gchar **argsv;
	
  switch (key)
    {
    case 'q':
      params.verbose = FALSE; 
      break;
	
    case 'i':
      params.action = TC_BUILD_MODEL;
      break;

    case 'k':
      params.action = TC_MAKE_FOLDS;
      break;

    case 'I':
      params.action = TC_TRAIN_BY_CLASS;
      params.train_params.approach |= FOLD_TRAIN_CLASS_FILTER;
      break;

    case 't':
      params.action = TC_TRAIN_BY_CLASS;
      params.train_params.approach |= FOLD_TRAIN_BY_CLASS;
      break;

    case 'T':
      params.action = TC_TRAIN_ALL;
      params.train_params.approach |= FOLD_TRAIN_MLTREE;
      break;

    case 'a':
      params.action = TC_TRAIN_ALL;
      params.train_params.approach |= FOLD_TRAIN_ALL_TOGETHER;
      break;

    case 'z':
      params.action = TC_TEST_CLASSES;
      break;

    case 'R':
      params.action = TC_TEST_RANKED;
      break;

    case 'K':
      params.action = TC_TEST_TREE;
      break;

    case 'O':
      params.action = TC_TEST_FILTERING;
      break;

    case 'o':
      params.action = TC_TEST_ALL;
      break;

    case 'X':
      params.action = TC_CLASSIFY;
      break;

    case '1':
      /* Here we read classifiers with their options */
	  if (params.action != TC_TEST_FILTERING) {
		fprintf(stderr, "ERROR: stage parameters are only for --test-filtering option\n");
		argp_usage(state);
	  }
      subopts = arg;
      while (*subopts != '\0')
		switch (getsubopt (&subopts, stage1_opts, &value)) 
		  {
		  case POSITIVE:
			params.train_params.stage1_positive = (strcmp(value, "1") == 0);
			break;
		  case NTOP:
			params.train_params.stage1_ntop = strtol(value, (char **)NULL, 10);
			if (errno == ERANGE) {
			  fprintf(stderr, "Invalid number of top classes for stage 1\n");
			  argp_usage(state);	
			}  
			break;
		  default:
			tc_error("ERROR: invalid stage 1 options");
		  }		  
	  break;

    case '2':
      /* Here we read classifiers with their options */
	  if (params.action != TC_TEST_FILTERING) {
		fprintf(stderr, "ERROR: stage parameters are only for --test-filtering option\n");
		argp_usage(state);
	  }
      subopts = arg;
      while (*subopts != '\0')
		switch (getsubopt (&subopts, stage2_opts, &value)) 
		  {
		  case POSITIVE:
			params.train_params.stage2_positive = (strcmp(value, "1") == 0);
			break;
		  case NTOP:
			params.train_params.stage2_ntop = strtol(value, (char **)NULL, 10);
			if (errno == ERANGE) {
			  fprintf(stderr, "Invalid number of top classes for stage 2\n");
			  argp_usage(state);	
			}  
			break;
		  case NBOTTOM:
			params.train_params.stage2_nbottom = strtol(value, (char **)NULL, 10);
			if (errno == ERANGE) {
			  fprintf(stderr, "Invalid number of bottom classes for stage 2\n");
			  argp_usage(state);	
			}  
			break;
		  case MERGE:
			if (!strcmp(value, "union"))
				params.train_params.stage2_merge = FOLD_STAGE2_UNION;
			else if (!strcmp(value, "intersection"))
				params.train_params.stage2_merge = FOLD_STAGE2_INTERSECTION;
			else if (!strcmp(value, "difference"))
				params.train_params.stage2_merge = FOLD_STAGE2_DIFFERENCE;
			else 
			  tc_error("ERROR: invalid merge option");
			break;
		  default:
			tc_error("ERROR: invalid stage 2 options");
		  }		  
	  break;

    case '3':
      params.action = TC_CLASS_CORRELATION;
      break;

    case '4':
      params.action = TC_EXPORT_MODELS;
      params.classdir = arg;
      break;

    case '5':
      params.action = TC_DUMP_TERM_IDS;
      break;

    case '8':
      params.action = TC_DUMP_FOLD_TERM_IDS;
      params.fold = atoi(arg);
      break;

    case '6':
      params.action = TC_DUMP_CLASS_IDS;
      break;

    case '7':
      params.action = TC_PACK_MODELS;
      break;

    case 'F':
      params.ptextdir = arg;
      break;

    case 'C':
      params.classdir = arg;
      break;

    case 'x':
      params.pt_params.maxlen = atoi(arg);
      break;

    case 'm':
      params.pt_params.minlen = atoi(arg);
      break;

    case 'r':
      params.minfreq = atoi(arg);
      break;

    case 's':
      params.pt_params.stopwdlist = arg;
      break;

    case 'S':
      params.pt_params.flags = params.pt_params.flags & (!TC_STOPPER);
      break;

    case 'E':
      params.pt_params.flags = params.pt_params.flags & (!TC_STEMMER);
      break;
	
    case 'g':
      params.train_params.fig_size = atoi(arg);
      break;

    case 'w':
      params.train_params.ftrans &= ~TC_WEIGHT_TFIDF;		
      params.train_params.ftrans |= TC_WEIGHT_ENTROPY;
      break;

    case 'W':
      params.train_params.ftrans &= ~TC_WEIGHT_ENTROPY;
      params.train_params.ftrans |= TC_WEIGHT_TFIDF;
      break;

    case 'e':
      params.train_params.ftrans |= TC_WEIGHT_ICF;
      break;

    case 'N':
      params.train_params.ftrans &= ~TC_COSINE_NORMALIZATION;
      break;

      /* TBD case 'G':
	 i = atoi(arg);
	 switch (i) {
	 case 0: params.train_params.ftrans |= TC_GLOBAL_WEIGHT_DOC; break;
	 case 1: params.train_params.ftrans |= TC_GLOBAL_WEIGHT_CLASS; break;
	 case 2: params.train_params.ftrans |= TC_GLOBAL_WEIGHT_DOC | TC_GLOBAL_WEIGHT_CLASS; break;
	 default: 
	 fprintf(stderr, "Invalid global weighting scheme\n");
	 argp_usage(state);
	 break;
	 }
	 break;
      */

    case 'b':
      params.train_params.beta = atof(arg);
      break;

    case 'f':
      params.train_params.nfolds = strtol(arg, (char **)NULL, 10);
      if (errno == ERANGE || !params.train_params.nfolds) {
        fprintf(stderr, "Invalid number of folds\n");
        argp_usage(state);	
      }  
      break;

    case 'd':
      params.train_params.fdf_min = strtol(arg, (char **)NULL, 10);
      if (errno == ERANGE) {
        fprintf(stderr, "Invalid minimal feature document frequency\n");
        argp_usage(state);	
      }  
      break;

    case 'j':
      params.train_params.cdf_min = strtol(arg, (char **)NULL, 10);
	  if (errno == ERANGE) {
		fprintf(stderr, "Invalid minimal class document frequency\n");
		argp_usage(state);	
	  }  
      break;

    case 'u':
      i = atoi(arg);
      switch (i) {
      case 0: params.train_params.selection_measure = METHOD_SELECT_BY_PRECISION; break;
      case 1: params.train_params.selection_measure = METHOD_SELECT_BY_RECALL; break;
      case 2: params.train_params.selection_measure = METHOD_SELECT_BY_FALLOUT; break;
      case 3: params.train_params.selection_measure = METHOD_SELECT_BY_ACCURACY; break;
      case 4: params.train_params.selection_measure = METHOD_SELECT_BY_ERROR; break;
      case 5: params.train_params.selection_measure = METHOD_SELECT_BY_OPT_DISTANCE; break;
      case 6: params.train_params.selection_measure = METHOD_SELECT_BY_F; break;
      case 7: params.train_params.selection_measure = METHOD_SELECT_BY_F1; break;
      case 8: params.train_params.selection_measure = METHOD_SELECT_BY_BEP; break;
      default: 
		fprintf(stderr, "Invalid selection method\n");
		argp_usage(state);
		break;
      }
      break;

    case 'U':
      params.train_params.selection_threshold = atof(arg);
      break;

    case 'M':
      params.train_params.ntop_shown = atoi(arg);
      break;

    case 'Q':
      params.train_params.compute_scut = TRUE;
      break;

    case 'P':
      params.train_params.ntop = strtol(arg, (char **)NULL, 10);
	  if (errno == ERANGE) {
		fprintf(stderr, "Invalid top number of classes to return\n");
		argp_usage(state);	
	  }  
      break;

    case 'A':
      params.train_params.positive_ranked = strtol(arg, (char **)NULL, 10) == 1;
	  if (errno == ERANGE) {
		fprintf(stderr, "Must be either 1 or 0 to determine whether to consider only positive classes or not in ranked classification\n");
		argp_usage(state);	
	  }  
      break;

    case 'H':
      params.train_params.tree_max_children = strtol(arg, (char **)NULL, 10);
	  if (errno == ERANGE) {
		fprintf(stderr, "Invalid number of maximum children per node\n");
		argp_usage(state);	
	  }  
      break;

    case 'B':
      params.train_params.tree_balance = strtol(arg, (char **)NULL, 10);
	  if (errno == ERANGE) {
		fprintf(stderr, "Invalid tree balance\n");
		argp_usage(state);	
	  }  
      break;

    case 'L':
      params.train_params.tree_max_depth = strtol(arg, (char **)NULL, 10);
	  if (errno == ERANGE) {
		fprintf(stderr, "Invalid maximal tree depth\n");
		argp_usage(state);	
	  }  
      break;

    case 'Z':
      params.train_params.tree_duplicates = TRUE;
      break;

    case 'Y':
      params.train_params.tree_traverse_all = TRUE;
      break;

    case 'c':
      /* Here we read classifiers with their options */
      subopts = arg;
      method = method_new();
      method->classifier = -1;
      learn_cmd = NULL, classify_cmd = NULL, format = NULL;
      while (*subopts != '\0')
		switch (getsubopt (&subopts, classifier_opts, &value)) 
		  {
		  case METHOD:
			if (method->learn_command || method->classify_command || method->format) {
			  fprintf(stderr, "ERROR: built-in methods are incompatible with 'learn', 'classify' and 'format' options\n");
			  argp_usage(state);
			}
			method->classifier = 0;
			method->args = g_strdup(value);
			argsv = g_strsplit(value, ":", 0);		
			for (i=1; i<CLASSIFIER_NULL; i++)
			  if (!strcmp(argsv[0], classifier_str_ids[i]))
				method->classifier = i;
			if (method->classifier == 0) {
			  fprintf(stderr, "ERROR: invalid built-in method\n");
			  argp_usage(state);
			}	
			g_strfreev(argsv);
			break;
		  case LEARN:		
			if (method->classifier > CLASSIFIER_COMMAND) {
			  fprintf(stderr, "ERROR: built-in methods do not take learn commands\n");
			  argp_usage(state);
			}
			method->classifier = CLASSIFIER_COMMAND;
			method->learn_command = g_strdup(value);
			break;
		  case CLASSIFY:
			if (method->classifier > CLASSIFIER_COMMAND) {
			  fprintf(stderr, "ERROR: built-in methods do not take classify commands\n");
			  argp_usage(state);
			}
			method->classify_command = g_strdup(value);
			if (method->learn_command == NULL) {
			  fprintf(stderr, "ERROR: 'classify' command must be given after 'learn' command\n");
			  argp_usage(state);		  
			}
			break;
		  case FORMAT:
			if (method->classifier > CLASSIFIER_COMMAND) {
			  fprintf(stderr, "ERROR: built-in methods do not need formats\n");
			  argp_usage(state);
			}
			if (strcmp(value, "vectors") && strcmp(value, "arff")) {
			  fprintf(stderr, "ERROR: invalid format for samples\n");
			  argp_usage(state);
			}
			if (!strcmp(value, "vectors"))
			  method->format = FORMAT_VECTORS;
			else if (!strcmp(value, "arff"))
			  method->format = FORMAT_ARFF;
			else {
			  fprintf(stderr, "ERROR: invalid format for samples\n");
			  argp_usage(state);
			}
			break;
		  default:
			tc_error("ERROR: invalid classifier syntax");
		  }
      if (params.train_params.approach & FOLD_TRAIN_CLASS_FILTER)
		methods_add(&params.train_params.stage2methods, method);
	  else
		methods_add(&params.train_params.stage1methods, method);
	  
      break;
	  
    case 'D':
      params.document = arg;
      break;

    case '9':
      params.hugemodel = TRUE;
      break;
		
    case ARGP_KEY_NO_ARGS:
      argp_usage(state);
      break;

    case ARGP_KEY_ARG:
      /* Here we know that state->arg_num == 0, since we
       * force argument parsing to end before any more plaum_params can
       * get here. */
      if (state->arg_num == 0)
	params.datadir = arg;
      else
	argp_usage(state);

      /* Now we consume all the rest of the plaum_params.
       * state->next is the index in state->argv of the
       * next argument to be parsed, which is the first string
       * we're interested in, so we can just use
       * &state->argv[state->next] as the value for
       * plaum_params->strings.
       * 
       * In addition, by setting state->next to the end
       * of the plaum_params, we can force argp to stop parsing here and
       * return. */
      state->next = state->argc;
      break;

    case ARGP_KEY_END:
      /* state->arg_num contains the number of plaum_params passed (non
       * parameters), except in ARGP_KEY_ARG, where arg_num is the
       * actual argument index. */
      if (state->arg_num > 1) 
	argp_usage (state);	/* not enough params */
		
      /* missing parameters */
      if (params.action == TC_BUILD_MODEL)
	{
	  if (!(params.datadir && params.ptextdir && params.classdir))
	    {
	      fprintf(stderr, "Invalid directories\n");
	      argp_usage(state);
	    }
	}
      if (params.action == TC_TRAIN_BY_CLASS | params.action == TC_TRAIN_ALL)
	{
	  if (params.train_params.methods->len == 0)
	    {
	      fprintf(stderr, "No methods provided\n");
	      argp_usage(state);
	    }
	}
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/* Our argp parser. */
static struct argp argp = { options, parse_opt, args_doc, doc };

int
main (int argc, char *argv[])
{
  Model m;
  Folds folds;
  Fold fold;
  guint i;
  IntStrTable idstr;

  //samples_to_css("/home/amontejo/research/data/tecat/byAuthors/fold0/train_samples_ascii.vcs", "/home/amontejo/out.css");
  //return 1;

  /* memory table for profile */	
	
  /* Default values. */
  params.pt_params.maxlen = 40;
  params.pt_params.minlen = 0;
  params.minfreq = 0;
  params.pt_params.flags = TC_PTPROC_DEFAULT;
  params.pt_params.stopwdlist = DEFAULT_STOP_WORDS;
  params.datadir = NULL;
  params.ptextdir = NULL;
  params.classdir = NULL;
  params.verbose = TRUE;
  params.hugemodel = FALSE;
  params.train_params.fig_size = 50000;
  params.train_params.fdf_min = 1;
  params.train_params.cdf_min = 1;
  params.train_params.ftrans = TC_WEIGHT_TFIDF | TC_COSINE_NORMALIZATION;
  params.train_params.nfolds = 10;
  params.train_params.ntop = 10;
  params.train_params.stage1methods = methods_new();
  params.train_params.stage2methods = methods_new();
  params.train_params.methods = params.train_params.stage1methods;
  params.train_params.beta = 1.0;
  params.train_params.tree_balance = 0;
  params.train_params.tree_max_children = 4;
  params.train_params.tree_max_depth = 7;
  params.train_params.tree_duplicates = FALSE;
  params.train_params.tree_traverse_all = FALSE;
  params.train_params.selection_measure = METHOD_SELECT_BY_F1;
  params.train_params.selection_threshold = 0.0;
  params.train_params.compute_scut = FALSE;
  params.train_params.positive_ranked = TRUE;
  params.train_params.ntop_shown = 10;
  params.train_params.stage1_ntop = 0;
  params.train_params.stage1_positive = TRUE;
  params.train_params.stage2_ntop = 0;
  params.train_params.stage2_nbottom = 0;
  params.train_params.stage2_positive = TRUE;  
  params.train_params.stage2_merge = FOLD_STAGE2_UNION;

  /* Parse our _params; every option seen by parse_opt will be
   * reflected in params. */
  argp_parse (&argp, argc, argv, 0, 0, &params);

  verbose = params.verbose;
	
  switch (params.action) {
	
  case TC_BUILD_MODEL:
    tc_message("\n*************************************\n");
    tc_message("*         Building model            *\n");
    tc_message("*************************************\n");
    m = model_new(params.datadir);	
    model_build(&m, params.ptextdir, params.classdir, &params.pt_params, params.minfreq);
    model_free(&m);
    break;
		
  case TC_MAKE_FOLDS:
    tc_message("\n*************************************\n");
    tc_message("*         Creating folds            *\n");
    tc_message("*************************************\n");
    m = model_new_read(params.datadir);
    folds = folds_new(m, &params.train_params);			
    folds_free(&folds);
    model_free(&m);
    break;

  case TC_TRAIN_BY_CLASS:
  case TC_TRAIN_ALL:
    tc_message("\n*************************************\n");
    tc_message("*         Training folds            *\n");
    tc_message("*************************************\n");
    m = model_new_read(params.datadir);
    folds = folds_new_read(params.datadir);
    folds_train(&folds, m, &params.train_params);
    folds_free(&folds);
    model_free(&m);
    break;
		
  case TC_TEST_CLASSES:
  case TC_TEST_TREE:
  case TC_TEST_RANKED:
  case TC_TEST_FILTERING:
  case TC_TEST_ALL:
    tc_message("\n*************************************\n");
    tc_message("*         Testing                   *\n");
    tc_message("*************************************\n");
    m = model_new_read(params.datadir);
    folds = folds_new_read(params.datadir);
    if (params.action == TC_TEST_CLASSES)
      folds_test(&folds, m, FOLD_TEST_BY_CLASS, &params.train_params);
    else if (params.action == TC_TEST_RANKED)
      folds_test(&folds, m, FOLD_TEST_BY_RANK, &params.train_params);
    else if (params.action == TC_TEST_TREE)
      folds_test(&folds, m, FOLD_TEST_BY_TREE, &params.train_params);
    else if (params.action == TC_TEST_FILTERING)
      folds_test(&folds, m, FOLD_TEST_FILTERING, &params.train_params);
    else if (params.action == TC_TEST_ALL)
      folds_test(&folds, m, FOLD_TEST_ALL_TOGETHER, &params.train_params);
    folds_free(&folds);
    model_free(&m);
    break;

  case TC_CLASSIFY:			
    tc_message("\n*************************************\n");
    tc_message("*         Finding classes           *\n");
    tc_message("*************************************\n");
    m = model_new_read(params.datadir);
    folds = folds_new_read(params.datadir);
    if (params.hugemodel)
        folds_classify_huge(&folds, m, params.train_params.ftrans, params.document);
    else
        folds_classify(&folds, m, params.train_params.ftrans, params.document);
    folds_free(&folds);
    model_free(&m);
	break;
  case TC_CLASS_CORRELATION:			
    tc_message("\n*************************************\n");
    tc_message("*    Correlation between classes    *\n");
    tc_message("*************************************\n");
    m = model_new_read(params.datadir);
    folds = folds_new_read(params.datadir);
	for (i = 0; i < folds->len; i++)
    {
      Fold fold;
      tc_message_int("\t-------------- Fold %d\n", i+1);		
      fold = fold_new_read(folds->path, i, folds->folds);
      fold_dump_correlation_matrix(fold, m);
      fold_free(&fold);
    }	
    folds_free(&folds);
    model_free(&m);
	break;
  case TC_EXPORT_MODELS:			
    tc_message("\n*************************************\n");
    tc_message("*    Exporting class models          *\n");
    tc_message("*************************************\n");    
    m = model_new_read(params.datadir);
    folds = folds_new_read(params.datadir);
    fold = fold_new_read(folds->path, 0, folds->folds);
    fold_export_models(fold, m, params.classdir);
    fold_free(&fold);
    folds_free(&folds);
    model_free(&m);
    break;
  case TC_PACK_MODELS:			
    tc_message("\n*************************************\n");
    tc_message("*      Packing class models         *\n");
    tc_message("*************************************\n");    
    folds = folds_new_read(params.datadir);
    //tc_message_int("Folds: %d", folds->len);
    //for now only work on one folder...
    //for (i = 0; i < folds->len; i++)
    //{
    tc_message("\t-------------- Working on Fold 0 \n");		
    fold = fold_new_read(folds->path, 0, folds->folds);
    fold_pack_models(fold);
    //}
    fold_free(&fold);
    folds_free(&folds);
    break;
  case TC_DUMP_TERM_IDS:			
    tc_message("\n*************************************\n");
    tc_message("*    Dump of term trans. table      *\n");
    tc_message("*************************************\n");    
    m = model_new_read(params.datadir);
    idstr = intstr_table_new(m->term_id);
    intstr_table_dump(idstr);
    intstr_table_free(&idstr);
    model_free(&m);
    break;
  case TC_DUMP_FOLD_TERM_IDS:			
    tc_message("\n********************************************\n");
    tc_message_int("*  Dump of term trans. table for fold %d   *\n", params.fold);
    tc_message("********************************************\n");    
    m = model_new_read(params.datadir);
    folds = folds_new_read(params.datadir);
    fold = fold_new_read(folds->path, params.fold, folds->folds);
    idstr = intstr_table_new(m->term_id);
    term_info_dump_new_id(fold->term_info, idstr);
    intstr_table_free(&idstr);
    model_free(&m);
    break;
  case TC_DUMP_CLASS_IDS:			
    tc_message("\n*************************************\n");
    tc_message("*    Dump of class trans. table     *\n");
    tc_message("*************************************\n");    
    m = model_new_read(params.datadir);
    idstr = intstr_table_new(m->class_id);
    intstr_table_dump(idstr);
    intstr_table_free(&idstr);
    model_free(&m);
    break;
  default:
	tc_message("ERROR: unknown action\n");
  }

  methods_free(&params.train_params.stage1methods);
  methods_free(&params.train_params.stage2methods);

  g_mem_profile();
  exit (0);
}
