/***************************************************************************
 *            method.c
 *
 *  Wed Nov 12 15:42:14 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/

#include <math.h>
#include <glib.h>
#include "method.h"
#include "tc_utils.h"

#define STR_BUFFER_LEN 256

extern gboolean verbose;

/* -------------------------------------------------------------------------- */
Method
method_new(void)
{
  Method minf = g_new0(struct method_struct, 1);
  minf->learn_command = NULL;
  minf->classify_command = NULL;
  minf->model_file = NULL;
  minf->tp = minf->tn = minf->fp = minf->fn = 0;
  minf->precision = minf->recall = minf->accuracy = minf->error = minf->bep = 0.0;
  minf->F = minf->F1 = minf->opt_distance = minf->fallout = minf->nctrain = minf->nctest = 0.0;
  minf->class_id = -1;
  minf->pos_factor = minf->beta = 1.0;
  minf->cut = 0.0;
  return minf;
}
/* -------------------------------------------------------------------------- */
Method
method_new_clone(Method minf)
{
  Method mnew = method_new();
  *mnew = *minf;
  mnew->learn_command = g_strdup(minf->learn_command);
  mnew->classify_command = g_strdup(minf->classify_command);
  mnew->model_file = g_strdup(minf->model_file);
  mnew->args = g_strdup(minf->args);
  return mnew;
}
/* -------------------------------------------------------------------------- */
void 
method_reset_measures (Method *m) {
  Method minf = *m;    
  minf->tp = minf->tn = minf->fp = minf->fn = 0;
  minf->precision = minf->recall = minf->accuracy = minf->error =  
    minf->F = minf->F1 = minf->opt_distance = minf->fallout = minf->bep = 0.0; 
}
/* -------------------------------------------------------------------------- */
void 
method_compute_measures (Method *m) 
{
  guint all = (*m)->tp+(*m)->fp+(*m)->tn+(*m)->fn; 
  (*m)->precision = (*m)->recall = (*m)->accuracy = (*m)->error = (*m)->F = (*m)->opt_distance = (*m)->fallout = 0.0; 
  if ((*m)->tp+(*m)->fp > 0) 
    (*m)->precision = ((double)(*m)->tp)/((double)((*m)->tp+(*m)->fp)); 
  else 
    (*m)->precision = 0.0; 
  if ((*m)->tp+(*m)->fn > 0) 
    (*m)->recall = ((double)(*m)->tp)/((double)((*m)->tp+(*m)->fn)); 
  else 
    (*m)->recall = 0.0; 
  if ((*m)->fp+(*m)->tn > 0) 
    (*m)->fallout = ((double)(*m)->fp)/((double)((*m)->fp+(*m)->tn)); 
  else 
    (*m)->fallout = 0.0; 
  (*m)->accuracy = ((double)((*m)->tp+(*m)->tn))/((double)all); 
  (*m)->error = ((double)((*m)->fp+(*m)->fn))/((double)all); 
  if ((*m)->beta*(*m)->beta*(*m)->precision+(*m)->recall != 0.0) 
    (*m)->F = (((*m)->beta*(*m)->beta+1.0)*(*m)->precision*(*m)->recall)/((*m)->beta*(*m)->beta*(*m)->precision+(*m)->recall); 
  else 
    (*m)->F = 0.0; 
  if ((*m)->precision+(*m)->recall > 0.0)
	(*m)->F1 = 2.0*(*m)->precision*(*m)->recall/((*m)->precision+(*m)->recall);
  else
	(*m)->F1 = 0.0;
  (*m)->opt_distance = 1.0 - sqrt(((1.0-(*m)->precision)*(1.0-(*m)->precision)+(1.0-(*m)->recall)*(1.0-(*m)->recall)) / 2.0);
  (*m)->bep = ((*m)->precision+(*m)->recall)/2.0;
}
/* -------------------------------------------------------------------------- */
/* Creates a sample where each feature represents a document, each
   feature id is either 0 (out) or 1 (in) and each feature weight
   the decision value returned by the external algorithm */
Sample 
method_new_sample_assignment_from_external(Method *method, BitSet class_docs, BitSet eval_docs, gchar *predictions_path)
{
  Method m = *method;
  FILE *fdec = fopen (predictions_path, "r");
  double value;
  Sample s;

  if (!fdec)
    tc_error("method_evaluate: unable to open decissions file");

  s = sample_new();
  bitset_it_reset(&eval_docs);
  m->tp = m->tn = m->fp = m->fn = 0; 
  while (!bitset_it_end(eval_docs) && !feof(fdec)) {
	tc_id di = bitset_it_next(&eval_docs);      
	/* decision value */
	fscanf(fdec, "%lf", &value);
	sample_add_feature(&s, feature_new(bitset_in(class_docs, di) ? 1 : 0, value));
  }
  fclose(fdec);
  return s;
}
/* -------------------------------------------------------------------------- */
void 
method_evaluate_sample_based(Method *method, Sample s) {
  Method m = *method;
  Feature f;

  method_reset_measures(&m);
  sample_it_reset(&s);
  while (!sample_it_end(s)) {
	f = sample_it_next(&s);		
	if (f->id)
	  if (f->w > m->cut)    /* value should be positive */
		m->tp++;
	  else
		m->fn++;		
	else                    /* value should be negative */
	  if (f->w > m->cut)
		m->fp++;
	  else
		m->tn++;
  }    
  method_compute_measures(&m);
  if (verbose) {
    printf("(%.2f/%.2f)", m->precision*100.0, m->recall*100.0);
    fflush(stdout);
  }
}
/* -------------------------------------------------------------------------- */
void
method_dump_measures(Method m)
{
  //printf("TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F  Opt.Dist.\n");
  printf("%6d %6d %6d %6d %10f %9f %9f %9f %9f %9f %9f %9f %10f %9f\n", 
	 m->tp, m->tn, m->fp, m->fn, m->precision, m->recall, m->accuracy, m->error, 
	 m->fallout, m->beta, m->F, m->F1, m->opt_distance, m->bep);
}
/* -------------------------------------------------------------------------- */
void 
method_dump(Method m)
{
  if (m->classifier == CLASSIFIER_COMMAND) {
    printf("Learn command: %s\n", m->learn_command);
    printf("Classifying command: %s\n", m->classify_command);
  }
  else {
    printf("Classifier: %s\n", classifier_str_ids[m->classifier]);
    printf("Built-in: %s\n", m->args);
  }
  printf("Model file: %s\n", m->model_file);
  printf("    TP     TN     FP     FN  Precision    Recall  Accuracy     Error   Fallout      Beta         F        F1   Opt.Dist.      BEP\n");
  method_dump_measures(m);
  printf("Cut: %lf\n", m->cut);
}
/* -------------------------------------------------------------------------- */
void
method_serialize(Method m, FILE *fhnd)
{
  fprintf(fhnd, "%d\n", m->id);
  fprintf(fhnd, "%d\n", m->classifier);
  fprintf(fhnd, "%d\n", m->class_id);
  fprintf(fhnd, "%d\n", m->format);
  fprintf(fhnd, "%s\n", m->args);
  fprintf(fhnd, "%s\n", m->learn_command);
  fprintf(fhnd, "%s\n", m->classify_command);
  fprintf(fhnd, "%s\n", m->model_file);
  fprintf(fhnd, "%f\n", m->cut);
  fprintf(fhnd, "TP: %d, TN: %d, FP: %d, FN: %d\n", m->tp, m->tn, m->fp, m->fn);
  fprintf(fhnd, "Precision/recall: %.4f/%.4f\n", m->precision, m->recall);
  fprintf(fhnd, "Accuracy/error: %.4f/%.4f\n", m->accuracy, m->error);
  fprintf(fhnd, "Fallout: %.4f\n", m->fallout);
  fprintf(fhnd, "Beta/F: %.4f/%.4f\n", m->beta, m->F);
  fprintf(fhnd, "F1: %.4f\n", m->F1);
  fprintf(fhnd, "Distance to optimum: %.4f\n", m->opt_distance);
  fprintf(fhnd, "BEP: %.4f\n", m->bep);
}
/* -------------------------------------------------------------------------- */
Method
method_new_deserialize(FILE *fhnd)
{
  Method m = method_new();
  gchar strbuff[TC_BUFFER_LENGTH];
	
  /* ids */
  fscanf(fhnd, "%d\n", &m->id);
  fscanf(fhnd, "%d\n", &m->classifier);  
  fscanf(fhnd, "%d\n", &m->class_id);
  fscanf(fhnd, "%d\n", &m->format);
	
  /* arguments used for training */
  tc_fscanline(fhnd, strbuff);
  m->args = g_strdup(strbuff);

  /* learn command */
  tc_fscanline(fhnd, strbuff);
  m->learn_command = g_strdup(strbuff);

  /* classify command */	
  tc_fscanline(fhnd, strbuff);
  m->classify_command = g_strdup(strbuff);

  /* model file */
  tc_fscanline(fhnd, strbuff);
  m->model_file = g_strdup(strbuff);
		
  /* cut */
  fscanf(fhnd, "%lf\n", &m->cut);

  /* meassures */
  fscanf(fhnd, "TP: %d, TN: %d, FP: %d, FN: %d\n", &m->tp, &m->tn, &m->fp, &m->fn);
  fscanf(fhnd, "Precision/recall: %lf/%lf\n", &m->precision, &m->recall);
  fscanf(fhnd, "Accuracy/error: %lf/%lf\n", &m->accuracy, &m->error);
  fscanf(fhnd, "Fallout: %lf\n", &m->fallout);
  fscanf(fhnd, "Beta/F: %lf/%lf\n", &m->beta, &m->F);
  fscanf(fhnd, "F1: %lf\n", &m->F1);
  fscanf(fhnd, "Distance to optimum: %lf\n", &m->opt_distance);
  fscanf(fhnd, "BEP: %lf\n", &m->bep);
	
  return m;
}
/* -------------------------------------------------------------------------- */
gboolean
method_valid(Method m, MethodSelection measure, gdouble threshold, gdouble *value)
{
  switch (measure) {
  case METHOD_SELECT_BY_PRECISION: 	*value = m->precision; break;
  case METHOD_SELECT_BY_RECALL: 	*value = m->recall; break;
  case METHOD_SELECT_BY_ACCURACY:	*value = m->accuracy; break;
  case METHOD_SELECT_BY_ERROR:		*value = -m->error; threshold *= -1.0; break;
  case METHOD_SELECT_BY_FALLOUT: 	*value = -m->fallout; threshold *= -1.0; break;
  case METHOD_SELECT_BY_OPT_DISTANCE:   *value = m->opt_distance; break;
  case METHOD_SELECT_BY_F: 		*value = m->F; break;
  case METHOD_SELECT_BY_F1: 		*value = m->F1; break;
  case METHOD_SELECT_BY_BEP: 		*value = m->bep; break;
  }	
  return (*value > threshold);
}

/* -------------------------------------------------------------------------- */
void method_exec_cmd(Method m, gchar *samples, gchar *predictions)
{
  #define MAX_BUFFER_LEN 1000

  int state=0, pos=0, offset=0;
  gchar cmd[MAX_BUFFER_LEN], *pfile, *cmd_noecho, *str, *pos_factor_str;

  if (strlen(predictions) == 0)
    str = m->learn_command;
  else
    str = m->classify_command;

  if (strlen(str)+strlen(m->model_file)+strlen(samples)+strlen(predictions)+3 >= MAX_BUFFER_LEN) /* overflow */
    tc_error("__fold_exec_cmd: overflow of command string buffer");

  /* we create command */
  pos_factor_str = g_strdup_printf("%.4lf", m->pos_factor);
  while (pos <= strlen(str)) {
    switch (state) 
      {
      case 0:
		if (str[pos] == '[')
		  state = 1;
		else
		  cmd[pos+offset] = str[pos];
		break;
      case 1:
		switch(str[pos]) 
		  {
		  case 'm': pfile = m->model_file; break;
		  case 's': pfile = samples; break;
		  case 'p': pfile = predictions; break;
		  case 'f': pfile = pos_factor_str; break;
		  }
		strcpy(&cmd[pos+offset-1], pfile); 
		offset += strlen(pfile)-3;
		pos++;
		state = 0;
		break;	
      }
    pos++;
  }
  g_free(pos_factor_str);

  /* we execute */
  cmd_noecho = g_strconcat(cmd, " > /dev/null", NULL);
  state = system (cmd_noecho);
  g_free (cmd_noecho);
  if (state == -1)
    tc_error ("Broken execution");
}
/* -------------------------------------------------------------------------- */
void* method_learn_builtin(Method minf, Samples ss) {
  /* if we have gamma_auto we have to replace it for the pos factor */
  
  void* c = classifier_new[minf->classifier](minf->args);
  classifier_learn[minf->classifier](&c, ss);
  classifier_save[minf->classifier](c, minf->model_file); 
  return c;  
}
/* -------------------------------------------------------------------------- */
Sample 
method_new_sample_assignment_from_builtin(Method minf, Samples ss, void *classifier)
{
  Sample score, s;
  double value;

  score = sample_new();
  samples_it_reset(&ss);
  while (!samples_it_end(ss)) {
	s = samples_it_next(&ss);
	value = classifier_classify[minf->classifier](classifier, s);
	sample_add_feature(&score, feature_new(s->y > 0.0 ? 1 : 0, value));	
  }
  return score;
}
/* -------------------------------------------------------------------------- */
BitSet method_classify_builtin(Method minf, Samples ss, void *classifier) {
  BitSet assignment = bitset_new(ss->max_id+1);
  samples_it_reset(&ss);
  while (!samples_it_end(ss)) {
    Sample s = samples_it_next(&ss);
    if (classifier_classify[minf->classifier](classifier, s) > minf->cut)
      bitset_insert(&assignment, s->id);
  }
  return assignment;
}
/* -------------------------------------------------------------------------- */
void  method_evaluate_builtin(Method *minf, Samples ss, BitSet assignment) {
  Method m = *minf;
  samples_it_reset(&ss);
  m->tp = m->tn = m->fp = m->fn = 0;
  while (!samples_it_end(ss)) {
    Sample s = samples_it_next(&ss);
    if (s->y > 0.0) /* value should be positive */ 
      if (bitset_in(assignment, s->id))
		m->tp++;
      else
		m->fn++;
    else            /* value should be negative */
      if (bitset_in(assignment, s->id))
		m->fp++;
      else
		m->tn++;
  }		
  method_compute_measures(&m);
  if (verbose) {
    printf("(%.2f/%.2f)", m->precision*100.0, m->recall*100.0);
    fflush(stdout);
  }
}
/* -------------------------------------------------------------------------- */
void method_evaluate_assignment(Method *minf, BitSet auto_assigned, BitSet  real_assigned, guint total_classes) {
  tc_id c;
  Method m = *minf;

  m->tp = m->tn = m->fp = m->fn = 0;

  bitset_it_reset(&auto_assigned);
  while (!bitset_it_end(auto_assigned)) {
    c = bitset_it_next(&auto_assigned);
    if (bitset_in(real_assigned, c))
      m->tp++;
  }

  m->fp = auto_assigned->total - m->tp;
  m->fn = real_assigned->total - m->tp;
  m->tn = total_classes - m->tp - m->fp - m->fn;

  method_compute_measures(&m);
}
/* -------------------------------------------------------------------------- */
void method_compute_scut(Method *method, Sample s, MethodSelection selection_measure) {
  gdouble scut = 0.0, best_measure = 0.0, val;
  Method minf = *method, m;
  GList* it;
  Feature fi, fj;

  /* margin based classifier work worse with scut */
  //if (minf->classifier & CLASSIFIER_PLAUM)
  //return;

  /* for non-margin */
  m = method_new();
  sample_it_reset(&s);
  while (!sample_it_end(s)) {
	fi = sample_it_next(&s);
	it = s->it;
	sample_it_reset(&s);
	method_reset_measures(&m);
	while (!sample_it_end(s)) {
	  fj = sample_it_next(&s);
	  if (fj->w > fi->w) {
		if (fj->id)
		  m->tp++;
		else
		  m->fp++;		
	  } else {
		if (fj->id)
		  m->fn++;
		else
		  m->tn++;
	  }
	}
	s->it = it;
	method_compute_measures(&m);
	if (method_valid(m, selection_measure, best_measure, &val)) {
	  best_measure = val;
	  scut = fi->w;
	}
  }
  method_free(&m);
  minf->cut = scut;
}
/* -------------------------------------------------------------------------- */

void
method_free(Method *m)
{
  if ((*m)->learn_command)
    g_free((*m)->learn_command);
  if ((*m)->classify_command)
    g_free((*m)->classify_command);
  if ((*m)->model_file)
    g_free((*m)->model_file);
  if ((*m)->args)
    g_free((*m)->args);
  g_free(*m);
}
