/***************************************************************************
 *            method.h
 *
 *  Wed Nov 12 15:40:41 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch 
 *
 * In this object the interface to all possible built-in methods is
 * specified. To add a new method you have to:
 *
 * 1. Add a new entry at "Methods indentifiers" (must be before METHOD_NULL)
 * 2. Add a new entry at "Methods strings names"
 *
 * 3. Add a new entry at "Methods learn commands", here you specify a
 *    pointer to a function of type: 
 *    void* (*learn_func)(void * method_struct, Samples samples)  TBD
 *
 * N. Add the corresponding help at main.c
 ****************************************************************************/
#ifndef _METHOD_H
#define _METHOD_H

#include <glib.h>
#include "bitset.h"
#include "tc_types.h"
#include "classifiers/classifiers.h"

/* DATA TYPE -------------------------------------------------------------- */

typedef enum {
  METHOD_SELECT_BY_PRECISION = 1 << 0,
  METHOD_SELECT_BY_RECALL = 1 << 1,
  METHOD_SELECT_BY_FALLOUT = 1 << 2,
  METHOD_SELECT_BY_ACCURACY = 1 << 3,
  METHOD_SELECT_BY_ERROR = 1 << 4,
  METHOD_SELECT_BY_OPT_DISTANCE = 1 << 5,
  METHOD_SELECT_BY_F = 1 << 6,
  METHOD_SELECT_BY_F1 = 1 << 7,
  METHOD_SELECT_BY_BEP = 1 << 8
} MethodSelection;

enum {
  /* Formats */
  FORMAT_VECTORS = 0,
  FORMAT_ARFF,
};

typedef struct method_struct
{
  int id;
  int classifier; /* type of classifier (see classifiers/classifiers.h) */
  gchar *args;    /* string with optional parameters for built-in algorithms */

  int format;    /* format of samples file (see 'Formats' above) */
  gchar *learn_command;
  gchar *classify_command;
  gchar *model_file; /* path to the file where classifier trained data is stored */	
  tc_id class_id; /* class to which this method has been applied */
  gdouble pos_factor; /* fraction of positive samples over negative samples in training */
  gdouble cut;       /* threshold of on decission factor, default 0,
						can be optimized according to a given
						measure using SCut approach */
		
  /* measures */
  guint tp, tn, fp, fn; /* true/false positive/negative predictions */
  double precision;   /* tp / (tp+fp) */
  double recall;      /* tp / (tp+fn) */
  double fallout;     /* fp / (fp+tn) */
  double accuracy;    /* (tp+tn) / (tp+fp+tn+fn) */
  double error;       /* (fp+fn) / (tp+fp+tn+fn) */
  double beta, F;     /* see Kjersti survey */
  double F1;          /* 2pr/(p+r) */
  double opt_distance; /* based on euclidean distance to the optimal point (100%/100%) */	
  double bep;         /* break-even-point (average between precision and recall) */ 

  /* for counting number of classes trained and test (since method is
	 also used to store fold statistics */
  double nctrain, nctest;
  
} *Method;

/* FUNCTIONS -------------------------------------------------------------- */

Method
method_new(void);

Method
method_new_clone(Method m);

void
method_free(Method *m);

void 
method_dump(Method m);

void 
method_reset_measures (Method *m);

void 
method_compute_measures (Method *m);

/* Creates a sample where each feature represents a document, each
   feature id is either 0 (out) or 1 (in) and each feature weight
   the decision value returned by the external algorithm */
Sample
method_new_sample_assignment_from_external(Method *method, BitSet class_docs, BitSet eval_docs, gchar *predictions_path);

Sample 
method_new_sample_assignment_from_builtin(Method minf, Samples ss, void *classifier);

void 
method_evaluate_sample_based(Method *method, Sample s);

void
method_compute_scut(Method *method, Sample s, MethodSelection selection_measure);

void
method_dump_measures(Method m);

void
method_serialize(Method m, FILE *fhnd);


void 
method_exec_cmd(Method m, gchar *samples, gchar *predictions);

Method
method_new_deserialize(FILE *fhnd);

/* Determines whether the method measure is over the threshold or not,
   returns the value of the measure in value */
gboolean
method_valid(Method m, MethodSelection measure, gdouble threshold, gdouble *value);

/* Returns a classifier that must be freezed using classifier_free[]() */
void* 
method_learn_builtin(Method minf, Samples ss);

BitSet 
method_classify_builtin(Method minf, Samples ss, void *classifier);

void
method_evaluate_builtin(Method *minf, Samples ss, BitSet assignment);

/* Evaluates measures for a given assignment to one document */
void 
method_evaluate_assignment(Method *minf, BitSet auto_assigned, BitSet  real_assigned, guint total_classes);

#endif /* _METHOD_H */
