/***************************************************************************
 *            methods.c
 *
 *  Tue Nov 11 17:34:25 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/

#include <math.h>
#include "methods.h"
#include "tc_utils.h"
#include "tc_types.h"
#include "classifiers/classifiers.h"

/* -------------------------------------------------------------------------- */
Methods 
methods_new(void)
{	
  return g_ptr_array_new();
}
/* -------------------------------------------------------------------------- */
Methods 
methods_new_sized(guint size)
{	
  Methods m = g_ptr_array_new();
  g_ptr_array_set_size(m, size);
  return m;
}
/* -------------------------------------------------------------------------- */
Methods
methods_new_clone(Methods m)
{
  Methods methods = methods_new();
  guint i;
  for (i=0; i<m->len; i++)
    methods_add(&methods, method_new_clone(methods_get(m, i)));
  return methods;
}
/* -------------------------------------------------------------------------- */
void 
methods_add(Methods *methods, Method minf)
{
  minf->id = (*methods)->len;
  g_ptr_array_add(*methods, minf);
}
/* -------------------------------------------------------------------------- */
void 
methods_add_with_commands(Methods *methods, gchar *learn_command, gchar *classify_command)
{
  Method minf = method_new();
  minf->learn_command = g_strdup(learn_command);
  minf->classify_command = g_strdup(classify_command);
  methods_add(methods, minf);
}
/* -------------------------------------------------------------------------- */
Method
methods_get(Methods methods, guint index)
{
  if (index >= methods->len)
    return NULL;
  return g_ptr_array_index(methods, index);
}
/* -------------------------------------------------------------------------- */
void
methods_free(Methods *methods)
{
  Methods m = *methods;
  guint k;
  for (k=0; k<m->len; k++)
    {
      Method minf = methods_get(m, k);
      if (minf != NULL)
	method_free(&minf);
    }
  g_ptr_array_free(m, TRUE);
}
/* -------------------------------------------------------------------------- */
void methods_remove(Methods *m, guint index) 
{
  Method minf = methods_get(*m, index);
  if (minf != NULL)
    method_free(&minf);
  g_ptr_array_remove_index(*m, index);
}
/* -------------------------------------------------------------------------- */
gboolean methods_any_command(Methods m) 
{
  gboolean any = FALSE;
  Method minf;
  int i;
  for (i=0; i<m->len && !any; i++)
    {
      Method minf = methods_get(m, i);
      if (minf->classifier == CLASSIFIER_COMMAND)
	any = TRUE;
    }
  return any;
}
/* -------------------------------------------------------------------------- */
gboolean methods_any_builtin(Methods m) 
{
  gboolean any = FALSE;
  Method minf;
  int i;
  for (i=0; i<m->len && !any; i++)
    {
      Method minf = methods_get(m, i);
      if (minf->classifier != CLASSIFIER_COMMAND)
	any = TRUE;
    }
  return any;
}
/* -------------------------------------------------------------------------- */
void
methods_dump(Methods m)
{
  guint k;
  for (k=0; k<m->len; k++)
    {
      Method minf = methods_get(m, k);
      printf("-----------------------\n");
      method_dump(minf);
    }
}
/* -------------------------------------------------------------------------- */
void methods_set(Methods *docs_stats, Method m, guint index)
{
  g_ptr_array_index((*docs_stats), index) = m;
}
/* -------------------------------------------------------------------------- */
Method
methods_get_best(Methods m, MethodSelection measure)
{
  guint i;
  gboolean better;
  Method next;
  Method best = methods_get(m, 0);
  
  if (m->len == 1)
    return best;
  for (i=1; i<m->len; i++)
    {
      next = methods_get(m, i);
      better = FALSE;
      switch (measure) {
      case METHOD_SELECT_BY_PRECISION: 	better = (next->precision > best->precision); break;
      case METHOD_SELECT_BY_RECALL: 		better = (next->recall > best->recall); break;
      case METHOD_SELECT_BY_ACCURACY:		better = (next->accuracy > best->accuracy); break;
      case METHOD_SELECT_BY_ERROR:		better = (next->error < best->error); break;
      case METHOD_SELECT_BY_FALLOUT: 		better = (next->fallout < best->fallout); break;
      case METHOD_SELECT_BY_OPT_DISTANCE: better = (next->opt_distance > best->opt_distance); break;
      case METHOD_SELECT_BY_F: 			better = (next->F > best->F); break;
      case METHOD_SELECT_BY_F1:			better = (next->F1 > best->F1); break;
      case METHOD_SELECT_BY_BEP:		better = (next->bep > best->bep); break;
      }
      if (better)
		best = next;
    }
  return best;
}
/* -------------------------------------------------------------------------- */
Method
methods_macro_avg(Methods m)
{
  guint i;
  Method minf;
  Method mavg = method_new();
  guint total = 0;
  for (i=0; i<m->len; i++) {
	minf = methods_get(m, i);
	if (minf == NULL || minf->class_id == -1)
	  continue;
	total++;
	/* Macroaverage implies the average of top measures */
	mavg->precision += minf->precision;
	mavg->recall += minf->recall;
	mavg->accuracy += minf->accuracy;
	mavg->error += minf->error;
	mavg->fallout += minf->fallout;
	mavg->opt_distance += minf->opt_distance;
	mavg->F += minf->F;
	mavg->F1 += minf->F1;
	mavg->bep += minf->bep;
	mavg->nctest += minf->nctest;
	mavg->nctrain += minf->nctrain;
  }
  mavg->precision /= (double)total;
  mavg->recall /= (double)total;
  mavg->accuracy /= (double)total;
  mavg->error /= (double)total;
  mavg->fallout /= (double)total;
  mavg->opt_distance /= (double)total;
  mavg->F /= (double)total;
  mavg->F1 /= (double)total;
  mavg->bep /= (double)total;
  mavg->nctrain /= (double)total;
  mavg->nctest /= (double)total;

  return mavg;
}
/* -------------------------------------------------------------------------- */
Method
methods_micro_avg(Methods m)
{
  guint i;
  Method minf;
  Method mavg = method_new();
  for (i=0; i<m->len; i++)
    {
      minf = methods_get(m, i);
      if (minf == NULL || minf->class_id == -1)
		continue;
      /* Macroaverage implies compute from basic summatories of matches */
      mavg->tp += minf->tp;
      mavg->tn += minf->tn;
      mavg->fp += minf->fp;
      mavg->fn += minf->fn;
    }
  method_compute_measures(&mavg);

  return mavg;
}
/* -------------------------------------------------------------------------- */
void
methods_serialize(Methods methods, gchar *file_path)
{
  guint i;	
  FILE *fhnd = fopen(file_path, "w");
  if (!fhnd)
    tc_error("methods_serialize: unable to create file");
  fprintf(fhnd, "%d\n", methods->len);
  for (i=0; i<methods->len; i++)
    method_serialize(methods_get(methods, i), fhnd);
	
  fclose(fhnd);
}
/* -------------------------------------------------------------------------- */
Methods 
methods_new_deserialize(gchar *file_path)
{
  Methods m = methods_new();
  guint i;
  guint len;
  FILE *fhnd = fopen(file_path, "r");
  if (!fhnd)
    tc_error("methods_new_deserialize: unable to open file");

  fscanf(fhnd, "%d\n", &len);
  for (i=0; i<len; i++)
    methods_add(&m, method_new_deserialize(fhnd));
	
  fclose(fhnd);
  return m;
}
/* -------------------------------------------------------------------------- */
void
methods_evaluate_external_in_docs(Methods *docs_stats, BitSet class_docs, BitSet eval_docs, char * predictions_path)
{
  FILE *fdec = fopen (predictions_path, "r");
  double value;
  int total=0;
  if (!fdec)
    tc_error("methods_evaluate_external_in_docs: unable to open decissions file");

  bitset_it_reset(&eval_docs);
  while (!bitset_it_end(eval_docs) && !feof(fdec))
    {
      /* doc id, we get statistics for that document */
      tc_id di = bitset_it_next(&eval_docs);
      Method m = methods_get(*docs_stats, di);
      if (m == NULL) {
		m = method_new();
		methods_set(docs_stats, m, di);
      }
		
      /* decission value */
      fscanf(fdec, "%lf", &value);
				
      if (bitset_in(class_docs, di)) /* value should be positive */ 
		if (value > 0)
		  m->tp++;				
		else
		  m->fn++;		
      else                           /* value should be negative */
		if (value > 0)
		  m->fp++;
		else
		  m->tn++;			
    }	
	
  fclose(fdec);
}
/* -------------------------------------------------------------------------- */
void
methods_evaluate_builtin_in_docs(Methods *docs_stats, Samples ss, BitSet assignment)
{
  double value;

  samples_it_reset(&ss);
  while (!samples_it_end(ss))
    {
      Sample s = samples_it_next(&ss);
      Method m = methods_get(*docs_stats, s->id);
      if (m == NULL) {
		m = method_new();
		methods_set(docs_stats, m, s->id);
      }		
      if (s->y > 0.0)                /* sample should have been assigned */ 
		if (bitset_in(assignment, s->id))
		  m->tp++;				
		else
		  m->fn++;		
      else                           /* sample shouldn't have been assigned */ 
		if (bitset_in(assignment, s->id))
		  m->fp++;
		else
		  m->tn++;			
    }	  
}
/* -------------------------------------------------------------------------- */
void
methods_reset_measures(Methods *m)
{
  guint i;
  Method method;
  for	(i=0; i<(*m)->len; i++) {
    method = methods_get(*m, i);
    method_reset_measures(&method);
  }
}
