/***************************************************************************
 *            model.c
 *
 *  Tue Oct 21 17:09:01 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include "model.h"
#include "tc_utils.h"
#include "tc_types.h"
#include "sample.h"
#include "intstr_table.h"
#include "methods.h"
#include "chrono.h"
#include "class_info.h"
#include "term_info.h"

#define STR_BUFFER_LEN 256
extern gboolean verbose;

/* -------------------------------------------------------------------------- 
   PRIVATE
   -------------------------------------------------------------------------- */
Sample
model_plain_text_process (Model * model, gchar *file, guint minfreq)
{
  Model m = *model;
  Sample s;
  FILE *fdoc;
  StrIntTable tfreq = strint_table_new ();	/* table for frequency counting */
  char strbuff[TC_BUFFER_LENGTH];
  tc_id act_freq;

  if ((fdoc = fopen (file, "r")) == NULL)
    tc_error ("model_build: unable to open plain text file");

  /* 1. Term extraction */
  while (pt_parser_read (m->pt_parser, fdoc, strbuff))
    {		
      /* 2. Frequency counting */
      if ((act_freq = strint_table_get (tfreq, strbuff)) == -1)	/* new term */
	strint_table_set (&tfreq, strbuff, 1);
      else
	strint_table_set (&tfreq, strbuff, act_freq + 1);
    }

  /* 3. Frequency filtering */
  strint_table_filter_value (&tfreq, minfreq);

  /* 4. Traslation into features (term string to integers in a Sample) 
   * The translation table is update too */
  s = sample_new_from_strint_table (tfreq, &m->term_id);

  strint_table_free (&tfreq);
  fclose (fdoc);

  return s;
}

/* -------------------------------------------------------------------------- */
void
__model_add_sample (Model * model, Sample s, tc_id doc_id)
{
  Model m = *model;
  DocInfo *d;
  Feature f;

  sample_it_reset (&s);
  /* We add document to inverted index for every term in it  */
  while (!sample_it_end (s))		
    inverted_index_add (&m->term_doc, sample_it_next (&s)->id, doc_id);

  /* We update doc info vector */
  m->doc_info = g_array_set_size (m->doc_info, doc_id + 1);	/* resize vector */
  d = &g_array_index (m->doc_info, DocInfo, doc_id);
  d->size = 0;
  sample_it_reset(&s);
  while (!sample_it_end(s)) {
	f = sample_it_next(&s);
	d->size += (guint) f->w;
  }
  d->max_tf = (guint) s->max_w;
  d->offset = ftell (m->fdocsv);

  /* Now we store the sample in binary format */
  // sample_serialize (s, m->fdocsv, TRUE); 
  sample_serialize (s, m->fdocsv, FALSE); 
}

/* -------------------------------------------------------------------------- */
Sample
__model_classes_process (Model * model, char *file)
{
  FILE *f;
  Model m = *model;
  gchar strbuff[TC_BUFFER_LENGTH];
  StrIntTable classes = strint_table_new ();
  Sample s;

  if ((f = fopen (file, "r")) == NULL)
    tc_error ("model_build: unable to open classes file");

  while (tc_fscanline (f, strbuff))
    strint_table_set (&classes, strbuff, 0);

  /* we convert list of classes to a sample and update the translation
   * table for classes */
  s = sample_new_from_strint_table (classes, &m->class_id);

  strint_table_free (&classes);
  fclose (f);

  return s;
}

/* -------------------------------------------------------------------------- */
void
__model_add_classes (Model * model, Sample s, tc_id doc_id)
{
  Model m = *model;
  DocInfo *d;

  sample_it_reset (&s);
  while (!sample_it_end (s))
    {
      Feature f = sample_it_next (&s);

      /* We add document to inverted index for every class in it  */
      inverted_index_add (&m->class_doc, f->id, doc_id);
    }

  /* We update doc info vector */
  d = &g_array_index (m->doc_info, DocInfo, doc_id);
  d->cf = s->total;
}

/* -------------------------------------------------------------------------- */
inline void
__model_docs_process (Model * model, IntStrTable ist, guint minfreq)
{
  Model m = *model;
  gchar *filename;
  char *docfile, *classfile;
  tc_id i;

  /* documents processing */
  tc_message ("Processing documents...\n");

  for (i = 0; i <= m->doc_id->max_id; i++)
    {
      Sample s, c;

      if ((i % 10) == 0)
	tc_message_int ("%d\r", i);

      /* file name for that 'i' doc_id */
      filename = intstr_table_get (ist, i);

      /* we get features (sample) from plain text file */
      docfile = g_strconcat (m->ptextdir, PATH_CONNECTOR, filename,  NULL);
      s = model_plain_text_process (&m, docfile, minfreq);
      g_free (docfile);

      /* we get classes for that file */
      classfile = g_strconcat (m->classdir, PATH_CONNECTOR, filename,  NULL);
      c = __model_classes_process (&m, classfile);
      g_free (classfile);

      /*  update model  */
      __model_add_sample (&m, s, i);
      __model_add_classes (&m, c, i);

      sample_free (&s);
      sample_free (&c);
    }
  tc_message_int ("done (%d documents processed)\n", i);
}

/* -------------------------------------------------------------------------- */
void
__model_test_file_matching (Model m)
{
  GDir *ptdir = g_dir_open (m->ptextdir, 0, NULL);
  gchar *entry;
  gchar *classfile;

  if (!ptdir)
    tc_error ("model_build: invalid plain text directory");


  tc_message ("Checking doc/class files matching...");
  while ((entry = (gchar *) g_dir_read_name (ptdir)) != NULL)
    {
      tc_id doc_id;
      classfile = g_strconcat (m->classdir, PATH_CONNECTOR, entry, NULL);
      if (!g_file_test (classfile, G_FILE_TEST_EXISTS))
	tc_error_str ("model_build: mismatched file '%s'\n", entry);
      g_free (classfile);

      /* We create a new identifier for document */
      doc_id = m->doc_id->max_id + 1;
      strint_table_set (&m->doc_id, (char *) entry, doc_id);
    }
  tc_message ("ok\n");

  g_dir_close (ptdir);
}

/* -------------------------------------------------------------------------- */
/* This function stores all the information in disk */
void
__model_save_to_disk (Model * model)
{
  Model m = *model;

  tc_message ("Saving data to disk...");

  strint_table_save(m->term_id, m->filepath_term_ids); 
  strint_table_save(m->class_id, m->filepath_class_ids); 
  strint_table_save(m->doc_id, m->filepath_doc_ids); 

  doc_info_save(m->doc_info, m->filepath_doc_info);
  pt_parser_save(m->pt_parser, m->filepath_pt_parser);

  inverted_index_save(m->term_doc, m->filepath_term_doc_ii, m->filepath_term_doc_idx); 
  inverted_index_save(m->class_doc, m->filepath_class_doc_ii, m->filepath_class_doc_idx); 

  tc_message ("done\n");
}
/* -------------------------------------------------------------------------- */
void __model_set_filepaths(Model *m, gchar *datadir)
{
  (*m)->datadir = g_strdup(datadir);
  (*m)->filepath_samples = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_SAMPLES, NULL);
  (*m)->filepath_samples_ascii = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_SAMPLES_ASCII, NULL);
  (*m)->filepath_term_ids = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_TERM_IDS, NULL);
  (*m)->filepath_class_ids = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_CLASS_IDS, NULL);
  (*m)->filepath_doc_ids = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_DOC_IDS, NULL);
  (*m)->filepath_doc_info = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_DOC_INFO, NULL);
  (*m)->filepath_term_doc_ii = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_TERM_DOC_II, NULL);
  (*m)->filepath_term_doc_idx = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_TERM_DOC_IDX, NULL);
  (*m)->filepath_class_doc_ii = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_CLASS_DOC_II, NULL);
  (*m)->filepath_class_doc_idx = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_CLASS_DOC_IDX, NULL);
  (*m)->filepath_pt_parser = g_strconcat(datadir, PATH_CONNECTOR, MODEL_FILE_PT_PARSER, NULL);
}

/* -------------------------------------------------------------------------- 
   PUBLIC
   -------------------------------------------------------------------------- */
Model
model_new (char *datadir)
{
  Model m = g_new (struct model_struct, 1);

  /* directories  */
  __model_set_filepaths(&m, datadir);

  /* translation tables  */
  m->term_id = NULL;
  m->class_id = NULL;
  m->doc_id = NULL;

  /* information vectors */
  m->doc_info = NULL;

  /* inverted indexes */
  m->term_doc = NULL;
  m->class_doc = NULL;
	
  m->fdocsv = NULL;

  return m;
}

/* -------------------------------------------------------------------------- */
Model
model_new_read (char *datadir)
{
  Model m = model_new (datadir);

  tc_message ("Reading model...");

  m->pt_parser = pt_parser_new_read(m->filepath_pt_parser);
	
  m->term_id = strint_table_new_read(m->filepath_term_ids); 
  m->class_id = strint_table_new_read(m->filepath_class_ids); 
  m->doc_id = strint_table_new_read(m->filepath_doc_ids); 

  m->doc_info = doc_info_new_read(m->filepath_doc_info);
  m->term_doc = inverted_index_new_read(m->filepath_term_doc_ii, m->filepath_term_doc_idx); 
  m->class_doc = inverted_index_new_read(m->filepath_class_doc_ii, m->filepath_class_doc_idx); 
  if ((m->fdocsv = fopen(m->filepath_samples, "rb")) == NULL)
    tc_error("model_new_read: unable to open document vectors file");	
	
  tc_message (" done\n");
  return m;
}

/* -------------------------------------------------------------------------- */
void
model_build (Model * model, char *ptextdir, char *classdir, PTParserParams * pt,
	     guint minfreq)
{
  Model m = *model;
  gchar *docsv;
  IntStrTable ist;

  /* 
   *      Initialization 
   */
  m->term_id = strint_table_new ();
  m->class_id = strint_table_new ();
  m->doc_id = strint_table_new ();
  m->doc_info = g_array_new (FALSE, TRUE, sizeof (DocInfo));
  m->term_doc = inverted_index_new ();
  m->class_doc = inverted_index_new ();

  /* data directory creation when missing */
  if (!g_file_test (m->datadir, G_FILE_TEST_IS_DIR))
    mkdir (m->datadir, 0750);

  /* plain text parser */
  m->pt_parser = pt_parser_new(pt);
	
  /* document vectors (it will be close when freezing the model) */
  docsv =
    g_strconcat (m->datadir, PATH_CONNECTOR, MODEL_FILE_SAMPLES,
		 NULL);

  if ((m->fdocsv = fopen (docsv, "wb")) == NULL)
    tc_error ("__model_doc_process: unable to create document vectors file");
  g_free (docsv);

  m->ptextdir = ptextdir;
  m->classdir = classdir;

  __model_test_file_matching (m);
  ist = intstr_table_new (m->doc_id);
  __model_docs_process (&m, ist, minfreq);
  __model_save_to_disk (model);
  intstr_table_free (&ist);

  if (verbose)
    printf("%d docs\n%d classes\n%d terms\n",m->doc_id->max_id+1, m->class_id->max_id+1, m->term_id->max_id+1);

}
/* -------------------------------------------------------------------------- */
void model_show_classes_strings (Model m, Sample classes)
{
  Feature class;
  IntStrTable class_str = intstr_table_new(m->class_id);
  sample_it_reset(&classes);
  while (!sample_it_end(classes)) {
    class = sample_it_next(&classes);
    printf("%lf %s\n", class->w, intstr_table_get(class_str, class->id));
  }
  intstr_table_free(&class_str);
}
/* -------------------------------------------------------------------------- */
void
model_free (Model *m)
{
  if (*m)
    {
      /* parser */
      if ((*m)->pt_parser)
	pt_parser_free(&(*m)->pt_parser);
      /* translation tables  */
      if ((*m)->term_id) strint_table_free (&(*m)->term_id);
      if ((*m)->class_id) strint_table_free (&(*m)->class_id);
      if ((*m)->doc_id) strint_table_free (&(*m)->doc_id);
      /* information vectors */
      if ((*m)->doc_info) g_array_free ((*m)->doc_info, TRUE);
      /* inverted indexes */
      if ((*m)->term_doc) inverted_index_free (&(*m)->term_doc);
      if ((*m)->class_doc) inverted_index_free (&(*m)->class_doc);
      /* close doc vectors file */
      if ((*m)->fdocsv)
	fclose ((*m)->fdocsv);
      /* file paths */
#define FREE_FILEPATH(filepath) \
		if ((*m)->filepath) \
			g_free((*m)->filepath)
	
      FREE_FILEPATH(datadir);
      FREE_FILEPATH(filepath_samples);
      FREE_FILEPATH(filepath_samples_ascii);
      FREE_FILEPATH(filepath_term_ids);
      FREE_FILEPATH(filepath_class_ids);
      FREE_FILEPATH(filepath_doc_ids);
      FREE_FILEPATH(filepath_doc_info);
      FREE_FILEPATH(filepath_term_doc_ii);
      FREE_FILEPATH(filepath_term_doc_idx);
      FREE_FILEPATH(filepath_class_doc_ii);
      FREE_FILEPATH(filepath_class_doc_idx);
      FREE_FILEPATH(filepath_pt_parser);
      g_free (*m);		
    }
}

	
