#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "pt_parser.h"
#include "tc_utils.h"

/* -------------------------------------------------------------------------- */
#define IS_LETTER(c) (isalnum(c) || c == '+' || c == '-' || c == '\\')

/* -------------------------------------------------------------------------- */
PTParser
pt_parser_new (PTParserParams * params)
{
  PTParser p = g_new(struct pt_parser_struct, 1);
		
  if (params->flags & TC_STEMMER)
    {
      /* stemmer ON, we initialize it */
      p->stemmer = stemmer_new ();

      /* stemming presumes use of lowered case letters */
      params->flags = params->flags | TC_LOWER_CASE;
    }
  else
    p->stemmer = NULL;

  if (params->flags & TC_STOPPER)
    p->stopper = stopper_new (params->stopwdlist);
  else
    p->stopper = NULL;

  p->params = *params;
  p->params.stopwdlist = g_strdup(params->stopwdlist);

  return p;
}

/* -------------------------------------------------------------------------- */
void
pt_parser_free (PTParser * p)
{
  if (*p)
    {
      if ((*p)->params.stopwdlist)
	g_free((*p)->params.stopwdlist);
			
      if ((*p)->stemmer)
	stemmer_free (&(*p)->stemmer);

      if ((*p)->stopper)
	stopper_free (&(*p)->stopper);

      g_free (*p);
    }
}

/* -------------------------------------------------------------------------- */

int
pt_parser_read (PTParser p, FILE * fin, char *strbuff)
{
  int i, strbufflen;
  gboolean valid = FALSE;

  while (!valid) {
    guint l = 0, r, lo;

    if (!tc_fscanstr(fin, strbuff))
      return 0;

    strbufflen = strlen (strbuff);
    if (strbufflen == 0)
      continue;
		
    /* 
     * 1 - Removes heading and trailing punctuation 
     * This is useful e.g. to filter '(abc);' into 'abc' 
     */
	//printf("\n%s --> ", strbuff);
    r = strbufflen - 1;	/* left and right pointers */
    while (!IS_LETTER (strbuff[l]) && l < r)
      l++;
    while (!IS_LETTER (strbuff[r]) && l < r)
      r--;
    lo = l;
    for (i = 0; l <= r; l++, i++)
      strbuff[i] = strbuff[l];
    strbuff[i] = 0;
	strbufflen = strlen(strbuff);
    //printf("%s %d\n", strbuff, strlen(strbuff));
		
    /* 
     * 2 - length filter 
     */
    if (p->params.maxlen > 0 && strbufflen > p->params.maxlen)
      continue;
    if (p->params.minlen > 0 && strbufflen < p->params.minlen)
      continue;

    /* 
     * 3 - lower case 
     */
    if (p->params.flags & TC_LOWER_CASE)
      for (i = 0; i < strbufflen; i++)
	strbuff[i] = (char) tolower ((int) strbuff[i]);

    /* 
     * 4 - stop words removal 
     */
    if (p->stopper)
      if (stopper_is_stop_word (p->stopper, strbuff))
	continue;

    /* 
     * 5 - TBD -- multiwords detection 
     */

    /* 
     * 6 - stemming 
     */
    if (p->stemmer)
      stemmer_stem (p->stemmer, strbuff);

    if (strlen (strbuff) > 0) {
      valid = TRUE;
	  //printf(" %s", strbuff);
	}
  }
  return 1;
}
/* -------------------------------------------------------------------------- */
void pt_parser_save(PTParser p, gchar *filepath)
{
  FILE *fhnd = fopen(filepath, "wb");
  if (!fhnd)
    tc_error("pt_parser_save: unable to create file\n");
  fwrite(&(p->params), sizeof(PTParserParams), 1, fhnd);
  fprintf(fhnd, p->params.stopwdlist);
  //printf("[min %d, max %d, %s]\n", p->params.minlen, p->params.maxlen, p->params.stopwdlist);
  fclose(fhnd);
}
/* -------------------------------------------------------------------------- */
PTParser pt_parser_new_read(gchar *filepath)
{
  PTParserParams params;
  gchar strbuff[255];
  FILE *fhnd = fopen(filepath, "rb");
  if (!fhnd) {
    fprintf(stderr, "%s\n", strerror(errno));
    tc_error("pt_parser_new_read: unable to open file\n");
  }
  fread(&params, sizeof(PTParserParams), 1, fhnd);
  fscanf(fhnd, "%s", strbuff);
  params.stopwdlist = g_strdup(strbuff);
  //printf("[min %d, max %d, %s]\n", params.minlen, params.maxlen, params.stopwdlist);
  fclose(fhnd);
  return pt_parser_new(&params);
}
