/***************************************************************************
 *            sample.c
 *
 *  Tue Oct 21 15:45:46 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "sample.h"
#include "tc_utils.h"

/* -------------------------------------------------------------------------- */
/* Private data structures */
/* -------------------------------------------------------------------------- */

/* This structure is used to pass arguments in a g_hash_table_foreach function
   in the strint_table_to_sample function */
struct __to_sample_struct
{
  Sample s;		/* sample we are creating */
  StrIntTable tt;		/* translation table */
};

/* -------------------------------------------------------------------------- */
void
__sample_new_from_strint_table (gpointer key, gpointer value, gpointer user_data)
{
  tc_id id;
  struct __to_sample_struct *data = (struct __to_sample_struct *) user_data;

  if ((id = strint_table_get (data->tt, (char *) key)) == -1) {	/* new feature */		   
    strint_table_set (&data->tt, (char *) key, data->tt->max_id + 1);
    id = data->tt->max_id;
  }
  sample_add_feature (&data->s, feature_new (id, (double) *(tc_id *) value));
}

/* Creates a new sample using given table and traslation table
   tt -- translation table where terms will be converted to features using int values as weights 
   termtt -- term translation table, to rename terms into integers 
*/
Sample
sample_new_from_strint_table (StrIntTable tt, StrIntTable * termtt)
{
  /* data to be passed to the __strint_table_to_sample function */
  struct __to_sample_struct user_data;
  user_data.tt = *termtt;
  user_data.s = sample_new ();
  g_hash_table_foreach (tt->table, __sample_new_from_strint_table, &user_data);
  sample_add_feature_finished (&user_data.s);

  return user_data.s;
}

/* -------------------------------------------------------------------- */
Sample
sample_new ()
{
  Sample s = g_new (struct sample_struct, 1);
  s->features = NULL;	/* empty list */
  s->modulus = 0.0;
  s->max_id = 0;
  s->max_w = -999999;
  s->min_w = 999999;
  s->total = 0;
  s->it = NULL;		/* iterator pointer */
  return s;
}

/* -------------------------------------------------------------------- */
void
sample_free (Sample * s)
{
  if (*s)
    {
      g_list_foreach ((*s)->features, feature_gfunc_free, NULL);
      g_list_free ((*s)->features);
      g_free (*s);
    }
}

/* -------------------------------------------------------------------- */
/* This function is a formatting of sample_free to be passed in a
   foreach function */
void
sample_gfunc_free (gpointer data, gpointer user_data)
{
  sample_free ((Sample *) & data);
}
/* -------------------------------------------------------------------- */
void
sample_add_feature (Sample * s, Feature f)
{
  if (*s)
    {
      (*s)->features = g_list_prepend ((*s)->features, f);
      if ((*s)->max_id < f->id)
	(*s)->max_id = f->id;
      (*s)->total++;
      (*s)->modulus += f->w * f->w;
      if (f->w > (*s)->max_w)
	(*s)->max_w = f->w;
      if (f->w < (*s)->min_w)
	(*s)->min_w = f->w;
    }
}

/* -------------------------------------------------------------------- */
void
sample_add_feature_finished (Sample * s)
{
  (*s)->modulus = sqrt ((*s)->modulus);
}

/* -------------------------------------------------------------------- */
/* Sets the iterator pointer to the first element in the list */
void
sample_it_reset (Sample * s)
{
  if (!(*s))
    tc_error ("sample_it_reset: Sample not initialized\n");
  (*s)->it = (*s)->features;
}

/* -------------------------------------------------------------------- */
/* Returns the actual term pointed by the internal iterator and moves
   it forward */
Feature
sample_it_next (Sample * s)
{
  Feature t;

  if (!(*s))
    tc_error ("sample_it_next: Sample not initialized\n");

  t = (Feature) (*s)->it->data;
  if ((*s)->it)
    (*s)->it = (*s)->it->next;
  return t;
}

/* -------------------------------------------------------------------- */
/* Returns TRUE if internal iterator has reached last element in the
   list, FALSE otherwise */
gboolean
sample_it_end (Sample s)
{
  if (!s)
    tc_error ("sample_it_end: Sample not initialized\n");
  return (s->it == NULL);
}

/* -------------------------------------------------------------------- */
void
sample_dump (Sample s)
{
  printf ("id: %d\ny: %lf\nmax_id: %d\nmodulus: %lf\n", s->id, s->y, s->max_id, s->modulus);
  sample_it_reset (&s);
  while (!sample_it_end (s))
    feature_dump (sample_it_next (&s));
  printf ("\n");
}

/* -------------------------------------------------------------------- */
void sample_it_remove(Sample *s) {
  if (!(*s))
    tc_error("sample_it_remove: sample has not been initialized");

  if ((*s)->it)
    (*s)->features = g_list_delete_link((*s)->features, (*s)->it->prev);
  else
	(*s)->features = g_list_delete_link((*s)->features, g_list_last((*s)->features));
  (*s)->total--;
}
/* -------------------------------------------------------------------- */
gint
__sample_comp_id (gconstpointer a, gconstpointer b)
{
  return ((gint) ((Feature) a)->id - ((Feature) b)->id);
}

/* Sorts the sample features by id */
void
sample_id_sort (Sample * s)
{
  (*s)->features = g_list_sort ((*s)->features, __sample_comp_id);
}

/* -------------------------------------------------------------------- */
/* Writes the sample in a file 
   When written in non binary format features are sorted by id
*/
void
sample_serialize (Sample s, FILE * fhnd, gboolean binary)
{
  Feature f;

  if (binary)
    {
      fwrite (&s->total, sizeof (guint), 1, fhnd);
      sample_it_reset (&s);
      while (!sample_it_end (s))
	{
	  f = sample_it_next (&s);
	  fwrite (&f->id, sizeof (tc_id), 1, fhnd);
	  fwrite (&f->w, sizeof (double), 1, fhnd);
	}
    }
  else
    {
      /* we sort the vector before */
      sample_id_sort (&s);
      sample_it_reset (&s);
      while (!sample_it_end (s))
	{
	  f = sample_it_next (&s);
	  fprintf (fhnd, " %d:%lf", f->id, f->w);
	}
      fprintf (fhnd, "\n");	/* to mark end of sample */
    }
}

/* -------------------------------------------------------------------- */
/* Reads the sample allocated at the given offset */
Sample
sample_new_deserialize (FILE * fhnd, long offset, gboolean binary)
{
  tc_id id;
  Sample s;
  double w;
  int read;

  if (fseek (fhnd, offset, 0) == -1)
    tc_error ("sample_new_deserialize: unable to find offset\n");

  s = sample_new ();
  if (binary)
    {
      guint total;
      fread (&total, sizeof (guint), 1, fhnd);
      while (total--)
	{
	  fread (&id, sizeof (tc_id), 1, fhnd);
	  fread (&w, sizeof (double), 1, fhnd);
	  sample_add_feature (&s, feature_new (id, w));
	}
    }
  else
    {
      gboolean in_sample = TRUE;
      while (in_sample) {
	if (fgetc (fhnd) == '\n')
	  in_sample = FALSE;
	else {
	  read = fscanf (fhnd, "%d:%lf", &id, &w);
	  if (read != EOF && read == 2)
	    sample_add_feature (&s, feature_new (id, w));
	  else
	    tc_error("samples_new_deserialize: invalid format");
	  //in_sample = FALSE;		    
	}
      }
    }
  sample_add_feature_finished (&s);
  sample_id_sort(&s);
  sample_it_reset(&s);

  return s;
}

/* -------------------------------------------------------------------- */

BitSet sample_to_bitset(Sample s)
{
  BitSet bs = bitset_new(s->max_id);
  Feature f;

  sample_it_reset(&s);
  while (!sample_it_end(s)) {
	f = sample_it_next(&s);
	bitset_insert(&bs, f->id);
  }
	
  return bs;
}

/* -------------------------------------------------------------------- */

inline Sample
sample_new_reweight (Sample s, TermInfoArray term_info, BitSet train_docs, FeatureTransformations ftrans)
{
  TermInfo *tinf;
  Feature f;
  double w = 0.0, norm = 0.0;
  Sample new_s;

  /* 	
     We compute weight and normalization factor 
  */
  sample_it_reset (&s);
  while (!sample_it_end (s)) {	/* for each feature in document */
	f = sample_it_next (&s);		
	tinf = term_info_get (term_info, f->id);
	if (tinf->new_id != -1) { /* if it hasn't been filtered */
	  if (ftrans & TC_WEIGHT_ENTROPY)
		w = (1.0 + log(f->w)) * tinf->entropy;
	  else if (ftrans & TC_WEIGHT_TFIDF)
		w = f->w * tinf->idf;
	  else
		tc_error("sample_new_reweight: no weighting specified");
          if (ftrans & TC_WEIGHT_ICF) 
            w *= tinf->icf;
	  norm += w*w;
	  tinf->tmp_w = w;
	}
  }

  /* 	
     We normalize 
  */
  new_s = sample_new ();
  sample_it_reset (&s);
  while (!sample_it_end (s)) {	/* for each feature in document */
	f = sample_it_next (&s);		
	tinf = term_info_get (term_info, f->id);
	if (tinf->new_id != -1) { /* if it hasn't been filtered */
	  if (ftrans & TC_COSINE_NORMALIZATION && norm != 0.0)
		/* we avoid zero norms, because they exists when all the weights are zero (IDF of zero value) */
	  sample_add_feature (&new_s, feature_new (tinf->new_id, tinf->tmp_w / sqrt(norm)));
	  else
		sample_add_feature (&new_s, feature_new (tinf->new_id, tinf->tmp_w));
	}
  }

  sample_add_feature_finished (&new_s);
  return new_s;
}
