#include <stdlib.h>
#include <errno.h>
#include "samples.h"
#include "tc_utils.h"
#include "tc_types.h"
#include "classifiers/weights.h"
#define MAX_STR_BUFFER 255

extern gboolean verbose;
/* -------------------------------------------------------------------------- */
Samples samples_new() {
  Samples ss = (Samples) g_malloc(sizeof(struct samples_struct));

  if (ss == NULL) 
    tc_error("samples_new: No memory for new samples");

  ss->samples = NULL;    /* empty list */
  ss->it = NULL;
  ss->total = 0;
  ss->r2 = 0.0;
  ss->max_id = 0;
  ss->max_w = -999999;  
  ss->min_w = 999999;
  ss->max_feature_id = 0;
  return ss;	
}
/* -------------------------------------------------------------------------- */
/* This function creates a shallow copy of the list, i.e. list is
   copied, but not elements. The returned samples MUST be freezed with
   samples_shallow_free() */
Samples samples_shallow_clone(Samples ss) {
  Samples new_ss = samples_new();
  *new_ss = *ss;
  new_ss->it = new_ss->samples = g_list_copy(ss->samples);
  if (new_ss->samples == NULL)
    tc_error("samples_shallow_clone: no memory for list copy");
  return new_ss;
}
/* -------------------------------------------------------------------------- */
/* This function freezes samples struct and the list, but not each
   sample. Use this function only for samples created using the
   samples_shallow_clone function. */
void samples_shallow_free(Samples *ss) {
  if (*ss) {
    g_list_free((*ss)->samples);
    g_free(*ss);
  }
}
/* -------------------------------------------------------------------------- */
void samples_free(Samples *ss) {
  if (*ss) {
    g_list_foreach((*ss)->samples, sample_gfunc_free, NULL);
    g_list_free((*ss)->samples);
    g_free(*ss);
  }
}
/* -------------------------------------------------------------------------- */
/* It sets the label of samples and remove those ones that are not
   neither positive nor negative samples */
void samples_assign(Samples *ss, BitSet posdocs, BitSet negdocs) {
  tc_id d;
  Sample s;
  if (!(*ss))
    tc_error("samples_assign: samples has not been initialized");
  samples_it_reset(ss);
  while (!samples_it_end(*ss)) {
    s = samples_it_next(ss);
    if (bitset_in(posdocs, s->id))
      s->y = 1.0;
    else if (bitset_in(negdocs, s->id))
      s->y = -1.0;
    else
      samples_it_remove(ss);
  }
}
/* -------------------------------------------------------------------------- */
Samples
samples_new_read(gchar *samples_path, BitSet docs)
{
  Samples ss = samples_new();
  samples_read(&ss, samples_path, docs);
  return ss;
}
/* -------------------------------------------------------------------------- */
void samples_add(Samples *ss, Sample s) {
  if (!(*ss))
    tc_error("samples_read: samples has not been initialized");
  (*ss)->samples = g_list_prepend ((*ss)->samples, s);
  if ((*ss)->max_feature_id < s->max_id)
    (*ss)->max_feature_id = s->max_id;
  (*ss)->r2 = (s->modulus > (*ss)->r2 ? s->modulus : (*ss)->r2);
  (*ss)->max_id = (s->id > (*ss)->max_id ? s->id : (*ss)->max_id);
  (*ss)->max_w = (s->max_w > (*ss)->max_w ? s->max_w : (*ss)->max_w);
  (*ss)->min_w = (s->min_w < (*ss)->min_w ? s->min_w : (*ss)->min_w);
  (*ss)->total++;
}

/* -------------------------------------------------------------------------- */
/* Reads a file of samples */
int samples_read(Samples *ss, char *file, BitSet ids) {
  Sample s;
  FILE *f = fopen(file, "r");
  size_t offset=0;
  tc_id id;

  if (!f) {
    fprintf(stderr, "%s\n", strerror(errno));
    tc_error_str("samples_read: Unable to open file '%s'", file);
  }

  if (!(*ss))
    tc_error("samples_read: samples has not been initialized");

  if (ids) {
    bitset_it_reset(&ids);             /* we know the id of the document */
    while (!feof(f) && !bitset_it_end(ids)) {
      id = bitset_it_next(&ids);
      s = sample_new_deserialize(f, offset, FALSE);
      s->id = id;
      samples_add(ss, s);
      offset = ftell(f);      
    }
  } else {                             /* we don't set id of the document */
    while (!feof(f)) {
      samples_add(ss, sample_new_deserialize(f, offset, FALSE));
      offset = ftell(f);
    }
  }
  samples_add_finished(ss);
  fclose(f);
  return (*ss)->total;
}
/* -------------------------------------------------------------------------- */
void samples_add_feature(Samples *ss, tc_id id, BitSet docs) {
  Sample s;
  Feature f;
  gboolean already;

  samples_it_reset(ss);
  while (!samples_it_end(*ss)) {
	s = samples_it_next(ss);
	if (bitset_in(docs, s->id)) {
	  already = FALSE;
	  sample_it_reset(&s);
	  while (!sample_it_end(s)) {
		f = sample_it_next(&s);
		if (f->id == id)
		  already = TRUE;
	  }
	  if (!already)
		sample_add_feature(&s, feature_new(id, 1.0));
	}
  }
}
/* -------------------------------------------------------------------------- */
void samples_remove_feature(Samples *ss, tc_id id, BitSet docs) {
  Sample s;
  Feature f;

  samples_it_reset(ss);
  while (!samples_it_end(*ss)) {
	s = samples_it_next(ss);
	if (bitset_in(docs, s->id)) {
	  sample_it_reset(&s);
	  while (!sample_it_end(s)) {
		f = sample_it_next(&s);
		if (f->id == id) {
		  sample_it_remove(&s);
		  break;
		}
	  }
	}
  }
}
/* -------------------------------------------------------------------------- */
/* Creates a collections of samples composed by classes instead of terms */
Samples samples_new_for_class_filter(guint nclasses, InvertedIndex class_doc, BitSet doc_ids) {
  Samples ss = samples_new();
  Sample s;
  tc_id d, i;
  BitSet class_docs;

  bitset_it_reset(&doc_ids);             /* we know the id of the document */
  while (!bitset_it_end(doc_ids)) {
	d = bitset_it_next(&doc_ids);
	s = sample_new();
	s->id = d;
	for (i=0; i<nclasses; i++) {
	  class_docs = inverted_index_get(class_doc, i);
	  if (bitset_in(class_docs, d))
		sample_add_feature(&s, feature_new(i, 1.0));
	  bitset_free(&class_docs);
	}
	samples_add(&ss, s);
  }
  samples_add_finished(&ss);
  return ss;
}
/* -------------------------------------------------------------------------- */
void samples_add_finished(Samples *ss) {
  (*ss)->r2 *= (*ss)->r2;
}
/* -------------------------------------------------------------------------- */
gint __samples_compare(gconstpointer s1, gconstpointer s2, gpointer w) {
  return (gint)(weights_distance_to_sample((Weights)w, (Sample)s2) -
	  weights_distance_to_sample((Weights)w, (Sample)s1));
}
/* -------------------------------------------------------------------------- */
void samples_sort_by_similarity(Samples *ss, Sample s) {
  Weights w = weights_new_from_sample(s);
  (*ss)->it = (*ss)->samples = g_list_sort_with_data((*ss)->samples, __samples_compare, (gpointer)w);  
  weights_free(&w);
}
/* -------------------------------------------------------------------------- */
/* Sets the iterator pointer to the first sample in the list */
void samples_it_reset(Samples *ss) {
  if (!(*ss)) 
    tc_error("samples_it_reset: samples has not been initialized");
  (*ss)->it = (*ss)->samples;
}

/* -------------------------------------------------------------------------- */
/* Removes the node returned in the last call to samples_it_next() */
void samples_it_remove(Samples *ss) {
  if (!(*ss))
    tc_error("samples_it_remove: samples has not been initialized");

  if ((*ss)->it)
    (*ss)->samples = g_list_delete_link((*ss)->samples, (*ss)->it->prev);
  else
	(*ss)->samples = g_list_delete_link((*ss)->samples, g_list_last((*ss)->samples));
  (*ss)->total--;
}
/* -------------------------------------------------------------------------- */
/* Returns the actual sample pointed by the internal iterator and moves
   it forward */
Sample samples_it_next(Samples *ss) {
  Sample s;
	
  if (!(*ss))
    tc_error("samples_it_next: samples has not been initialized");
	
  s = (Sample)(*ss)->it->data;
  if ((*ss)->it)
    (*ss)->it = (*ss)->it->next;
  return s;
}

/* -------------------------------------------------------------------------- */
/* Returns TRUE if internal iterator has reached last element in the
   list, FALSE otherwise */
gboolean samples_it_end(Samples ss) {
  if (!ss)
    tc_error("samples_it_end: samples has not been initialized");
  return (ss->it == NULL);
}
/* -------------------------------------------------------------------------- */
void samples_dump(Samples ss) {
  Sample s;

  if (!ss) 
    tc_error("samples_dump: samples has not been initialized");

  samples_it_reset(&ss);
  while (!samples_it_end(ss)) {
    s = samples_it_next(&ss);
    //printf("%d ", s->id);
    sample_dump(s);
  }
  printf("(%d samples)\n", ss->total);
}
