/***************************************************************************
 *            term_info.c
 *
 *  Thu Nov 27 11:45:54 2003
 *  Copyright  2003  Arturo Montejo Raez
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/

#include <math.h>
#include <stdio.h>
#include "term_info.h"
#include "sample.h"
#include "tc_utils.h"
#include "tm.h"

/* -------------------------------------------------------------------------- 
   PRIVATE
   -------------------------------------------------------------------------- */
inline gint
__comp_ig (gconstpointer a, gconstpointer b)
{
  double dif = ((TermInfo *) b)->ig - ((TermInfo *) a)->ig;

  if (dif == 0.0) return 0;
  if (dif < 0.0) return -1;
  return 1;
}

/* -------------------------------------------------------------------------- */
inline gint
__comp_id (gconstpointer a, gconstpointer b)
{
  gint dif = ((TermInfo *) a)->id - ((TermInfo *) b)->id;

  if (dif == 0.0)	return 0;
  if (dif < 0.0)	return -1;
  return 1;
}

/* -------------------------------------------------------------------------- 
   PUBLIC FUNCTIONS
   -------------------------------------------------------------------------- */
TermInfoArray
term_info_new_compute (InvertedIndex term_doc, InvertedIndex class_doc, BitSet train_docs, guint fig_size)
{
  gchar msg[200];
  guint n, nt, t, i;
  BitSet *mclass_doc, classes;
  guint reduced=0;
  TermInfo *tinf;
  BitSet term_docs, train_t_docs;

  /* We create the array */
  GArray *term_info = g_array_new (FALSE, TRUE, sizeof (TermInfo));
  term_info = g_array_set_size (term_info, term_doc->max_id + 1);

  tc_message ("\tComputing features info...\n");

  /* load inverted index for categories in memory */
  mclass_doc = g_new (BitSet, class_doc->max_id + 1);
  for (i = 0; i <= class_doc->max_id; i++) {
	BitSet class_docs = inverted_index_get (class_doc, i);
	mclass_doc[i] =	bitset_new_intersection (train_docs, class_docs);
	bitset_free (&class_docs);
  }
  tc_message_int ("\t-- %d features", term_doc->max_id + 1);
  tc_message_int (", %d classes", class_doc->max_id + 1);

  classes = bitset_new(class_doc->max_id);
  bitset_fullfill(&classes);

  /* we iterate over each term */
  n = train_docs->total;	/* total number of documents for training */
  tc_message_int (", %d documents\n", n);

  for (t = 0; t < term_info->len; t++) {
	if ((t + 1) % 1000 == 0 || t == term_info->len - 1) {
	  sprintf (msg, "\t-- %d/%d     \r", t + 1, term_info->len);
	  tc_message (msg);
	}
	tinf = &g_array_index (term_info, TermInfo, t);
	tinf->new_id = tinf->id = t;	
	
	term_docs = inverted_index_get (term_doc, t);
	
	train_t_docs = bitset_new_intersection (train_docs, term_docs);
	bitset_free (&term_docs);
	tinf->df = nt = train_t_docs->total;	/* document frequency */
	if (nt > 0) {
	  /* values */
	  if (fig_size > 0) {
		tinf->ig = tm_infgain(mclass_doc, classes, train_t_docs, n, &tinf->cf);
		tinf->icf = log ((gdouble) (class_doc->max_id + 1) / (gdouble) tinf->cf);
                //printf("cf:%lf, icf:%lf\n", tinf->cf, tinf->icf);
	  }
	  tinf->idf = log ((gdouble) n / (gdouble) nt);			
	  tinf->accum_freq = 0;
	  tinf->entropy = 0.0;
	}
	bitset_free (&train_t_docs);
  }

  /* free memory */
  for (i = 0; i <= class_doc->max_id; i++)
    bitset_free (&mclass_doc[i]);
  g_free (mclass_doc);
  bitset_free(&classes);

  tc_message_int ("\n\tdone (%d removed)\n", reduced);
  return term_info;
}
/* -------------------------------------------------------------------------- */
void
term_info_reduce (TermInfoArray* term_info, guint fig_size, guint df_min)
{
  tc_id i, new_id = 1, count=0, reduced=0; /* new_id starts at 1 for
					      SVM and other external
					      classifiers */
  TermInfo *tinf;
	
  tc_message ("\tReducing features... ");

  /* filter by information gain */
  if (fig_size > 0 && fig_size < (*term_info)->len) {
	/* sort by information gain */
	g_array_sort (*term_info, __comp_ig);
	
	/* take just valid ones */
	for (i = 0; i <(*term_info)->len; i++) {
	  tinf = term_info_get(*term_info, i);
	  if (count >= fig_size) {
		tinf->new_id = -1;
		reduced++;
	  }
	  else if (tinf->new_id != -1)
		count++;			
	}
	/* resort by id (to restore order) */
	g_array_sort (*term_info, __comp_id);
	
	tc_message_int(" [%d by ig] ", reduced);
  }
  
  reduced = 0;
  /* filter by document frequency */
  if (df_min > 0) {
	for (i = 0; i < (*term_info)->len; i++) {
	  tinf = term_info_get(*term_info, i);
	  if (tinf->df < df_min && tinf->new_id != -1) {
	    tinf->new_id = -1;
	    reduced++;
	  }
	}
	tc_message_int(" [%d by df] ", reduced);
  }
  
  /* sets new ids */	
  for (i = 0; i < (*term_info)->len; i++) {
	tinf = term_info_get(*term_info, i);
	if (tinf->new_id != -1)
	  tinf->new_id = new_id++;
  }
  
  tc_message_int ("done (reduced to %d features)\n", new_id-1);
}
/* -------------------------------------------------------------------------- */
/* Dumps new_id of terms */

void term_info_dump_new_id(TermInfoArray term_info, IntStrTable idterm) {
  tc_id i, new_id = 1, count=0, reduced=0; /* new_id starts at 1 for
					      SVM and other external
					      classifiers */
  TermInfo *tinf;

  for (i = 0; i < term_info->len; i++) {
    tinf = term_info_get(term_info, i);
    printf("%d %s\n", tinf->new_id, intstr_table_get(idterm, tinf->id));
  }  
}

/* -------------------------------------------------------------------------- */
TermInfo *term_info_get(TermInfoArray term_info, guint index)
{
  return &g_array_index(term_info, TermInfo, index);
}
/* -------------------------------------------------------------------------- */
/* This function computes the entropy of every term */
void
term_info_compute_entropy(FILE * fdocsv, TermInfoArray* term_info, DocInfoArray doc_info, BitSet docs) 
{
  tc_id t;
  TermInfo *tinf;
  Sample s;
  Feature f;
	
  /* 
     First, we compute Fi
     NOTE: tinf->accum_freq and tinf->entropy have been reset in term_info_new_compute
  */
  bitset_it_reset(&docs);
  while (!bitset_it_end(docs))  { /* for each train document */
	s = sample_new_deserialize(fdocsv, doc_info_get(doc_info, bitset_it_next(&docs))->offset, TRUE);
	sample_it_reset(&s);
	while (!sample_it_end(s))  { /* for each feature in document */
	  f = sample_it_next(&s);
	  tinf = &g_array_index(*term_info, TermInfo, f->id);
	  if (tinf->new_id != -1)
	    tinf->accum_freq += (guint) f->w;
	}
	sample_free(&s);
  }
	
  /*
    Then we compute Entropy
  */
  bitset_it_reset(&docs);
  while (!bitset_it_end(docs)) { /* for each train document */
	s = sample_new_deserialize(fdocsv, doc_info_get(doc_info, bitset_it_next(&docs))->offset, TRUE);
	sample_it_reset(&s);
	while (!sample_it_end(s)) { /* for each feature in document */
	  f = sample_it_next(&s);
	  tinf = &g_array_index(*term_info, TermInfo, f->id);
	  if (tinf->new_id != -1)
	    tinf->entropy += f->w/((double)tinf->accum_freq) * tm_plog2p(f->w/((double)tinf->accum_freq));
	}
	sample_free(&s);
  }
	
  for (t=0; t<(*term_info)->len; t++) {
	tinf = &g_array_index(*term_info, TermInfo, t);
	if (tinf->new_id != -1)
	  tinf->entropy = 1.0 + tinf->entropy / log((double)docs->total);
  }		
}
/* -------------------------------------------------------------------------- */
void term_info_free(TermInfoArray *term_info)
{
  g_array_free(*term_info, TRUE);
}
/* -------------------------------------------------------------------------- */
/* Saves to disk */
void
term_info_save (TermInfoArray term_info, gchar *filepath)
{
  FILE *fhnd;
  if (!(fhnd = fopen(filepath, "wb")))
    tc_error("term_info_save: Unable to create file\n"); 
  fwrite(&term_info->len, sizeof(guint), 1, fhnd); 
  fwrite(term_info->data, sizeof(TermInfo), term_info->len, fhnd); 
  fclose(fhnd);
}
/* -------------------------------------------------------------------------- */
/* Reads from disk */
TermInfoArray
term_info_new_read(gchar *filepath)
{	
  TermInfoArray term_info;
  FILE *fhnd;
  guint len; 

  if (!(fhnd = fopen(filepath, "rb"))) 
    tc_error("term_info_new_read: Unable to open info file\n"); 
  term_info = g_array_new(FALSE, TRUE, sizeof(TermInfo)); 

  fread(&len, sizeof(guint), 1, fhnd); 
  term_info = g_array_set_size(term_info, len); 
  fread(term_info->data, sizeof(TermInfo), len, fhnd); 
  fclose(fhnd);
	
  return term_info;	
}
