/***************************************************************************
 *            tm.c
 *
 *  Text Mining miscelaneous functions
 *
 *  arturo.montejo.raez@cern.ch
 ****************************************************************************/

#include <math.h>
#include "tm.h"

/* -------------------------------------------------------------------------- */
/* For computing p log p efficiently */
#define LOG_2 0.693147180559945286
#define TWO_TO_MINUS_FORTY (1.0 / (1024.0 * 1024 * 1024 * 1024))

inline double tm_plog2p (double z)
{
  if (z == 0.0 || z == 1.0)
    return 0.0;
  else if (z > TWO_TO_MINUS_FORTY)
    return (z * log(z) / LOG_2);
  else
    return (-40.0 * z);
}

/* -------------------------------------------------------------------------- */
/* Computes the information gain for a given item (term of class) */
double                                /* information gain value */
tm_infgain(BitSet *mclass_doc,        /* vector of bitsets for classes */
	   BitSet classes,            /* classes in the set */
	   BitSet j_docs,             /* set of documents of item j */
	   guint n,                   /* total documents in the set */
	   guint *cf)                 /* returned class frequency for item */
{
  guint ni, nij, nj = j_docs->total;
  gdouble sum1 = 0.0, sum2 = 0.0, sum3 = 0.0, nj_n = ((gdouble) nj) / ((gdouble) n);
  BitSet i_docs, ij_docs;
  gdouble ni_n, nij_nj, ni_nij_n_nj;
			
  *cf=0;
  /* we iterate over each class */
  bitset_it_reset(&classes);
  while (!bitset_it_end(classes)) {
    tc_id i = bitset_it_next(&classes);
    i_docs = mclass_doc[i];
    if (i_docs == NULL)
      tc_error("tm_infgain: null pointer");
    ni = i_docs->total;
    if (ni == 0)
      continue;	/* this class is not in the training docs */
    ij_docs = bitset_new_intersection (j_docs, i_docs);
    nij = ij_docs->total;
    bitset_free (&ij_docs);
    if (nij) (*cf)++;	/* class frequency for term t */
    
    /* fractions */
    ni_n = ((gdouble) ni) / ((gdouble) n);
    nij_nj = ((gdouble) nij) / ((gdouble) nj);
    ni_nij_n_nj =
      ((gdouble) ni - nij) / ((gdouble) n - nj);
    
    /* summatories accumulations */
    sum1 += ni_n > 0.0 ? tm_plog2p(ni_n) : 0.0;
    sum2 += nij_nj > 0.0 ? tm_plog2p(nij_nj) : 0.0;
    sum3 += ni_nij_n_nj > 0.0 ? tm_plog2p(ni_nij_n_nj) : 0.0;
  }
  //printf("->%.4lf\n", (-sum1 + nj_n * sum2 + (1.0 - nj_n) * sum3));
  //return (-sum1 + nj_n * sum2 + (1.0 - nj_n) * sum3);
  return (1.0 + (nj_n * sum2 + (1.0 - nj_n) * sum3)/-sum1);
}

/* -------------------------------------------------------------------------- */
/* 
   This function computes the imbalance degree of a class
*/
gdouble __abs(gdouble x) {
  return (x<0?-x:x);
}

gdouble tm_imbalance_degree(guint ni, guint n) {
  if (n)
    return __abs(1.0 - 2.0*(double)ni/(double)n);
  return 0.0;
}
